#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "arithm.h"
#include "stats.h"

//-----------------------------------------------------------------------------
GEN
FpE_FpXQE_vert(GEN P, GEN Q, GEN T, GEN p) {
    //
    // Computes the equation of the vertical line through P and returns
    // its evaluation at the point Q.
    //
    if (ell_is_inf(P)) {
        return Rg_to_FpXQ(gen_1, T, p);
    }
    return FPX_FP_ADD(gel(Q, 1), FP_NEG(gel(P, 1), p), p);    
}
//-----------------------------------------------------------------------------
GEN
FpXQE_FpE_vert(GEN Q, GEN P, GEN T, GEN p) {
    //
    // Computes the equation of the vertical line through Q and returns
    // its evaluation at the point P.
    //
    if (ell_is_inf(Q)) {
        return Rg_to_FpXQ(gen_1, T, p);
    }
    return FPX_FP_ADD(FPX_NEG(gel(Q, 1), p), gel(P, 1), p);
}
//-----------------------------------------------------------------------------
GEN
FpE_FpXQE_tangent_update(GEN* R, GEN Q, GEN a4, GEN T, GEN p) {
    //
    // Computes the equation of the line tangent to *R and returns its
    // evaluation at the point Q. Also doubles the point R.
    //
    GEN x1 = gel(*R, 1);
    GEN y1 = gel(*R, 2);

    if ((cmpis(y1, 0) == 0) || (ell_is_inf(*R))) {
        GEN v = FpE_FpXQE_vert(*R, Q, T, p);
        *R = ellinf();
        return v;
    
    } else {
        GEN _2_x1    = MULUI(2, x1);
        GEN _2_y1    = MULUI(2, y1);
        GEN _sqrx1   = SQRI(x1);
            _sqrx1   = REM(_sqrx1, p);
        GEN _3_sqrx1 = MULUI(3, _sqrx1);
        
        GEN slope = FP_DIV(ADDII(_3_sqrx1, a4), _2_y1, p);
        
        GEN x3 = FP_SUB(SQRI(slope), _2_x1, p);
        GEN y3 = FP_NEG(ADDII(y1, MULII(slope, SUBII(x3, x1))), p);
        
        GEN tmp1 = FPX_FP_ADD(gel(Q, 1), FP_NEG(x1, p), p);
        GEN tmp2 = FPX_FP_ADD(FPX_FP_MUL(tmp1, slope, p), y1, p);
        GEN line = FPX_SUB(gel(Q, 2), tmp2, p);
        
        gel(*R, 1) = x3;
        gel(*R, 2) = y3;
        
        return line;
    }
}
//-----------------------------------------------------------------------------
GEN
FpXQE_FpE_tangent_update(GEN* R, GEN P, GEN a4, GEN T, GEN p) {
    //
    // Computes the equation of the line tangent to *R and returns its
    // evaluation at the point P. Also doubles the point R.
    //
    GEN x1 = gel(*R, 1);
    GEN y1 = gel(*R, 2);

    if ((signe(y1) == 0) || (ell_is_inf(*R))) {
        GEN v = FpXQE_FpE_vert(*R, P, T, p);
        *R = ellinf();
        return v;
    
    } else {
        GEN _2_x1    = ZX_MULS(x1, 2);
        GEN _2_y1    = ZX_MULS(y1, 2);
        GEN _sqrx1   = ZX_SQR(x1);
            _sqrx1   = FPX_REM(_sqrx1, T, p);
        GEN _3_sqrx1 = ZX_MULS(_sqrx1, 3);
        
        GEN slope = FPXQ_DIV(ZX_Z_ADD(_3_sqrx1, a4), _2_y1, T, p);
        GEN x3 = FPX_REM(FPX_RED(FPX_SUB(ZX_SQR(slope), _2_x1, p), p), T, p);
        GEN y3 = FPX_REM(FPX_RED(FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p), p), T, p);
        
        GEN tmp1 = FPX_FP_ADD(FPX_NEG(x1, p), gel(P, 1), p);
        GEN tmp2 = FPX_NEG(FPX_ADD(FPXQ_MUL(slope, tmp1, T, p), y1, p), p);
        GEN line = FPX_FP_ADD(tmp2, gel(P, 2), p);
        
        gel(*R, 1) = x3;
        gel(*R, 2) = y3;
        
        return line;
    }
}
//-----------------------------------------------------------------------------
GEN
FpE_FpXQE_chord_update(GEN* R, GEN P, GEN Q, GEN a4, GEN T, GEN p) {
    //
    // Computes the equation of the line through R and P, and returns its 
    // evaluation at the point Q. Also adds P to the point R.
    //
    if (equalii(gel(P, 1), gel(*R, 1))) {
        if (ell_is_inf(*R) || equalii(gel(P, 2), SUBII(p, gel(*R, 2)))) {
            GEN v = FpE_FpXQE_vert(*R, Q, T, p);
            *R    = ellinf();
            return v;

        } else {
            return FpE_FpXQE_tangent_update(R, Q, a4, T, p);
        }
    } else {
        GEN x1 = gel(*R, 1);
        GEN y1 = gel(*R, 2);
        
        GEN x2 = gel(P, 1);
        GEN y2 = gel(P, 2);
        
        GEN slope = FP_DIV(SUBII(y2, y1), SUBII(x2, x1), p);
        
        GEN x3 = FP_SUB(SUBII(gsqr(slope), x1), x2, p);
        GEN y3 = FP_NEG(ADDII(y1, MULII(slope, SUBII(x3, x1))), p);
        
        GEN tmp1 = FPX_FP_MUL(FPX_FP_ADD(gel(Q, 1), FP_NEG(x1, p), p), slope, p);
        GEN tmp2 = FPX_FP_ADD(tmp1, y1, p);
        GEN line = FPX_SUB(gel(Q, 2), tmp2, p);
        
        gel(*R, 1) = x3;
        gel(*R, 2) = y3;
        
        return line;
    }
}
//-----------------------------------------------------------------------------
GEN
FpXQE_FpE_chord_update(GEN* R, GEN Q, GEN P, GEN a4, GEN T, GEN p) {
    //
    // Computes the equation of the line through R and Q, and returns its 
    // evaluation at the point P. Also adds Q to the point R.
    //
    if (_ZX_equal(gel(Q, 1), gel(*R, 1))) {        
        if (ell_is_inf(*R) || _ZX_equal(gel(Q, 2), FpX_neg(gel(*R, 2), p))) {    
            GEN v = FpXQE_FpE_vert(*R, P, T, p);
            *R    = ellinf();
            return v;

        } else {
            return FpXQE_FpE_tangent_update(R, P, a4, T, p);
        }
    } else {
        GEN x1 = gel(*R, 1);
        GEN y1 = gel(*R, 2);
        
        GEN x2 = gel(Q, 1);
        GEN y2 = gel(Q, 2);
        
        GEN slope = FPXQ_DIV(ZX_SUB(y2, y1), ZX_SUB(x2, x1), T, p);
        
        GEN x3 = FPX_REM(FPX_RED(FPX_SUB(ZX_SUB(ZX_SQR(slope), x1), x2, p), p), T, p);
        GEN y3 = FPX_REM(FPX_RED(FPX_NEG(ZX_ADD(y1, ZX_MUL(slope, ZX_SUB(x3, x1))), p), p), T, p);
        
        GEN tmp1 = FPX_FP_ADD(FPX_NEG(x1, p), gel(P, 1), p);
        GEN tmp2 = FPX_NEG(FPX_ADD(FPXQ_MUL(slope, tmp1, T, p), y1, p), p);
        GEN line = FPX_FP_ADD(tmp2, gel(P, 2), p);
        
        line = FPX_REM(line, T, p);
        
        gel(*R, 1) = x3;
        gel(*R, 2) = y3;
        
        return line;
    }
}
//-----------------------------------------------------------------------------
GEN
FpXQE_FpE_miller(GEN Q, GEN P, GEN m, GEN a4, GEN T, GEN p, int elim)
{
    //
    // Returns the Miller function f_{m, Q} evaluated at the point P using
    // the standard Miller algorithm.
    //
    pari_sp ltop = avma;
    
    GEN num   = mkpoln(1, gen_1);
    GEN denom = mkpoln(1, gen_1);
    GEN point = gcopy(Q);
    GEN absm  = absi(m);

    GEN line;
    GEN result;
    
    int  m_is_neg = (cmpis(m, 0) < 0);
    long i        = expi(absm) - 1;
    
    setvarn(num,   varn(gel(Q, 1)));
    setvarn(denom, varn(gel(Q, 1)));
    
    while (i >= 0) {
        num = FPXQ_SQR(num, T, p);
        if (!elim) {
            denom = FPXQ_SQR(denom, T, p);
        }
        
        line = FpXQE_FpE_tangent_update(&point, P, a4, T, p);
        num  = FPXQ_MUL(num, line, T, p);
        if (!elim) {
            GEN v = FpXQE_FpE_vert(point, P, T, p);
            denom = FPXQ_MUL(denom, v, T, p);
        }
        if (bittest(absm, i)) {
            line = FpXQE_FpE_chord_update(&point, Q, P, a4, T, p);
            num  = FPXQ_MUL(num, line, T, p);
            if (!elim) {
                GEN v = FpXQE_FpE_vert(point, P, T, p);
                denom = FPXQ_MUL(denom, v, T, p);
            }
        }
        i--;
    }
    if (!elim) {
        result = FPXQ_DIV(num, denom, T, p);
    } else {
        result = gcopy(num);
    }    
    if (m_is_neg) {
        GEN vmp = FpXQE_FpE_vert(point, P, T, p);
        result  = FPXQ_MUL(result, vmp, T, p);
        result  = FPXQ_INV(result, T, p);
    }
    return gerepileupto(ltop, result);
}
//-----------------------------------------------------------------------------
GEN
FpE_FpXQE_miller(GEN P, GEN Q, GEN m, GEN a4, GEN T, GEN p, int elim) {
    //
    // Returns the Miller function f_{m, P} evaluated at the point Q using
    // the standard Miller algorithm.
    //
    pari_sp ltop = avma;
    
    GEN num   = mkpoln(1, gen_1);
    GEN denom = mkpoln(1, gen_1);
    GEN point = gcopy(P);
    GEN absm  = absi(m);

    GEN line;
    GEN result;
    
    int  m_is_neg = (cmpis(m, 0) < 0);
    long i        = expi(absm) - 1;
    
    setvarn(num,   varn(gel(Q, 1)));
    setvarn(denom, varn(gel(Q, 1)));
    
    while (i >= 0) {
        num = FPXQ_SQR(num, T, p);
        if (!elim) {
            denom = FPXQ_SQR(denom, T, p);
        }
        line = FpE_FpXQE_tangent_update(&point, Q, a4, T, p);
        num  = FPXQ_MUL(num, line, T, p);        
        if (!elim) {
            GEN v = FpE_FpXQE_vert(point, Q, T, p);
            denom = FPXQ_MUL(denom, v, T, p);
        }
        
        if (bittest(absm, i)) {
            line = FpE_FpXQE_chord_update(&point, P, Q, a4, T, p);
            num  = FPXQ_MUL(num, line, T, p);
            if (!elim) {
                GEN v = FpE_FpXQE_vert(point, Q, T, p);
                denom = FPXQ_MUL(denom, v, T, p);
            }
        }
        i--;
    }
    if (!elim) {
        result = FPXQ_DIV(num, denom, T, p);
    } else {
        result = gcopy(num);
    }    
    if (m_is_neg) {
        GEN vmp = FpE_FpXQE_vert(point, Q, T, p);
        result  = FPXQ_MUL(result, vmp, T, p);
        result  = FPXQ_INV(result, T, p);
    }
    return gerepileupto(ltop, result);
}
//-----------------------------------------------------------------------------
GEN
FpE_FpXQE_weil_pairing(GEN P, GEN Q, GEN m, GEN a4, GEN T, GEN p) {
    pari_sp ltop = avma;
    GEN num    = FpE_FpXQE_miller(P, Q, m, a4, T, p, 0);
    GEN denom  = FpXQE_FpE_miller(Q, P, m, a4, T, p, 0);
    GEN result = FPXQ_DIV(num, denom, T, p); 
    if (mpodd(m)) {
        result  = FPX_NEG(result, p);
    }
    return gerepileupto(ltop, result);
}
//-----------------------------------------------------------------------------
GEN
FpXQE_FpE_weil_pairing(GEN Q, GEN P, GEN m, GEN a4, GEN T, GEN p) {
    pari_sp ltop = avma;
    GEN num    = FpXQE_FpE_miller(Q, P, m, a4, T, p, 0);
    GEN denom  = FpE_FpXQE_miller(P, Q, m, a4, T, p, 0);
    GEN result = FPXQ_DIV(num, denom, T, p); 
    if (mpodd(m)) {
        result  = FPX_NEG(result, p);
    }
    return gerepileupto(ltop, result);
}
//-----------------------------------------------------------------------------
