#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "apip_config.h"
#include "funcs.h"
#include "pairings.h"
#include "arithm.h"
#include "commons.h"
#include "final_exp.h"
#include "cyclo_poly.h"

//-----------------------------------------------------------------------------
static GEN  _easy_pow(pairing_data_t* const pdata, GEN unred);
static GEN  _easy_pow_part_2(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_mnt(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_bn(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_scott_barreto_6(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_freeman(pairing_data_t* const pdata, GEN unreduced);
static GEN  _hard_pow_duan_et_al_15_1500(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_5_1500(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_7_1333(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_8_1250(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_9_1333(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_11_1200(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_15_1500(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_16_1250(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_17_1125(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_17_1188(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_18_1333(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_19_1111(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_24_1250(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_25_1300(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_26_1166(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_opti_27_1111(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_unknown(pairing_data_t* const pdata, GEN unreduced);
static GEN  _hard_pow_unknown_interleaving(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_unknown_katoetal(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow_unknown_avanzi(pairing_data_t* const pdata, GEN unred);
static GEN  _hard_pow(pairing_data_t* const pdata, GEN unreduced);
static GEN  _hard_pow_scottetal(pairing_data_t* const pdata, GEN unreduced);
static void _set_r_kato(pairing_data_t* const, unsigned int, unsigned short);
static void _set_l_avanzi(pairing_data_t* const, unsigned int, unsigned int);

MAYBE_UNUSED static int _compute_field_maps(pairing_data_t* const pdata);
//-----------------------------------------------------------------------------
GEN
final_exp_naive_char_p(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN e = pdata->final_exp;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;

    GEN res = FPXQ_POW(unreduced, e, t, p);

    OUT();
    return res;
}
//-----------------------------------------------------------------------------
GEN
final_exp_hard_char_p(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN res = _hard_pow(pdata, unreduced);
    OUT();
    return res;
}
//-----------------------------------------------------------------------------
GEN
final_exp_opti_char_p(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN easy  = _easy_pow(pdata, unreduced);
    GEN final = _hard_pow(pdata, easy);
    OUT();
    return final;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow(pairing_data_t* const pdata, GEN unreduced)
{
    switch (pdata->hardpow_method) {
        case NAIVE: {
            GEN p = pdata->charac;
            GEN t = pdata->irredpol;
            return FPXQ_POW(unreduced, pdata->hard_exp, t, p);
        }
        case INTERLEAVING:
            return _hard_pow_unknown_interleaving(pdata, unreduced);

        case AVANZI_MIHAILESCU:
            return _hard_pow_unknown_avanzi(pdata, unreduced);
            
        case KATO_ET_AL:
            return _hard_pow_unknown_katoetal(pdata, unreduced);
            
        case SCOTT_ET_AL:
            return _hard_pow_scottetal(pdata, unreduced);
        
        default:
            return _hard_pow_unknown_katoetal(pdata, unreduced);
    }
}    
//-----------------------------------------------------------------------------
static GEN
_hard_pow_scottetal(pairing_data_t* const pdata, GEN unreduced) {
    //
    // Family-specific exponentiation -- we need to have the parameter used
    // in the polynomials giving the characteristic, the cardinal and the
    // trace of the Frobenius.
    //
    IN();
    GEN hard;
    if ((pdata->u == NULL) || (pdata->family == UNKNOWN)) {
        hard = _hard_pow_unknown(pdata, unreduced);
        OUT();
        return hard;
    }
    
    switch (pdata->family) {
        //
        // The following functions are based on the idea presented by Scott
        // et al. in "On the final exponentiation for calculating pairings
        // on ordinary elliptic curves".
        //
        case MNT:
            hard = _hard_pow_mnt(pdata, unreduced);
            break;

        case BN:
            hard = _hard_pow_bn(pdata, unreduced);
            break;

        case SCOTT_BARRETO_6:
            hard = _hard_pow_scott_barreto_6(pdata, unreduced);
            break;
        
        case FREEMAN:
            hard = _hard_pow_freeman(pdata, unreduced);
            break;

        case OPTI_5_1500:
            hard = _hard_pow_opti_5_1500(pdata, unreduced);
            break;

        case OPTI_7_1333:
            hard = _hard_pow_opti_7_1333(pdata, unreduced);
            break;
                
        case OPTI_8_1250:
            hard = _hard_pow_opti_8_1250(pdata, unreduced);
            break;
            
        case OPTI_9_1333:
            hard = _hard_pow_opti_9_1333(pdata, unreduced);
            break;
        
        case OPTI_11_1200:
            hard = _hard_pow_opti_11_1200(pdata, unreduced);
            break;

        case DUAN_ET_AL_15_1500:
            hard = _hard_pow_duan_et_al_15_1500(pdata, unreduced);
            break;

        case OPTI_15_1500:
            hard = _hard_pow_opti_15_1500(pdata, unreduced);
            break;
        
        case OPTI_16_1250:
            hard = _hard_pow_opti_16_1250(pdata, unreduced);
            break;
        
        case OPTI_17_1125:
            hard = _hard_pow_opti_17_1125(pdata, unreduced);
            break;
        
        case OPTI_17_1188:
            hard = _hard_pow_opti_17_1188(pdata, unreduced);
            break;
        
        case OPTI_18_1333:
            hard = _hard_pow_opti_18_1333(pdata, unreduced);
            break;

        case OPTI_19_1111:
            hard = _hard_pow_opti_19_1111(pdata, unreduced);
            break;
                
        case OPTI_24_1250:
            hard = _hard_pow_opti_24_1250(pdata, unreduced);
            break;

        case OPTI_25_1300:
            hard = _hard_pow_opti_25_1300(pdata, unreduced);
            break;

        case OPTI_26_1166:
            hard = _hard_pow_opti_26_1166(pdata, unreduced);
            break;
        
        case OPTI_27_1111:
            hard = _hard_pow_opti_27_1111(pdata, unreduced);
            break;
        
        default:
            hard = _hard_pow_unknown(pdata, unreduced);
    }
    OUT();
    return hard;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_scott_barreto_6(pairing_data_t* const pdata, GEN unreduced)
{       
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0, x, t, p);
    GEN mx2 = FPXQ_POW(mx1, x, t, p);
    GEN mx3 = FPXQ_POW(mx2, x, t, p);

    GEN imx3  = FPXQ_INV(mx3, t, p);
    GEN mx1p1 = FPM_FPXQ_MUL(frob_pows[1], mx1, t, p);

    GEN y0 = mx1p1;
    
    GEN y1;
    y1 = FPXQ_MUL(mx2, mx0, t, p);
    y1 = FPXQ_MUL(y1, imx3, t, p);
    
    GEN T0 = FPXQ_SQR(y1, t, p);
    GEN T1 = FPXQ_SQR(T0, t, p);
    GEN T2 = FPXQ_MUL(T1, y0, t, p);
    
    return T2;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_5_1500(pairing_data_t* const pdata, GEN unreduced)
{       
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,   x, t, p);
    GEN mx2  = FPXQ_POW(mx1,   x, t, p);
    GEN mx3  = FPXQ_POW(mx2,   x, t, p);
    GEN mx4  = FPXQ_POW(mx3,   x, t, p);
    GEN mx5  = FPXQ_POW(mx4,   x, t, p);
    GEN mx6  = FPXQ_POW(mx5,   x, t, p);
    GEN mx7  = FPXQ_POW(mx6,   x, t, p);
    GEN mx8  = FPXQ_POW(mx7,   x, t, p);
    GEN mx9  = FPXQ_POW(mx8,   x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);

    GEN imx1  = FPXQ_INV(mx1,   t, p);
    GEN imx3  = FPXQ_INV(mx3,   t, p);
    GEN imx4  = FPXQ_INV(mx4,   t, p);
    GEN imx7  = FPXQ_INV(mx7,   t, p);
    GEN imx9  = FPXQ_INV(mx9,   t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);

    GEN mx8p1   = FPM_FPXQ_MUL(frob_pows[1], mx8,   t, p);
    GEN mx5p1   = FPM_FPXQ_MUL(frob_pows[1], mx5,   t, p);
    GEN mx11p1  = FPM_FPXQ_MUL(frob_pows[1], mx11,  t, p);
    GEN mx2p1   = FPM_FPXQ_MUL(frob_pows[1], mx2,   t, p);
    GEN mx9p1   = FPM_FPXQ_MUL(frob_pows[1], mx9,   t, p);
    GEN mx0p1   = FPM_FPXQ_MUL(frob_pows[1], mx0,   t, p);
    GEN imx7p1  = FPM_FPXQ_MUL(frob_pows[1], imx7,  t, p);
    GEN imx4p1  = FPM_FPXQ_MUL(frob_pows[1], imx4,  t, p);
    GEN imx10p1 = FPM_FPXQ_MUL(frob_pows[1], imx10, t, p);
    GEN imx1p1  = FPM_FPXQ_MUL(frob_pows[1], imx1,  t, p);

    GEN mx10p2 = FPM_FPXQ_MUL(frob_pows[2], mx10, t, p);
    GEN mx8p2  = FPM_FPXQ_MUL(frob_pows[2], mx8,  t, p);
    GEN mx6p2  = FPM_FPXQ_MUL(frob_pows[2], mx6,  t, p);
    GEN mx4p2  = FPM_FPXQ_MUL(frob_pows[2], mx4,  t, p);
    GEN mx2p2  = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx0p2  = FPM_FPXQ_MUL(frob_pows[2], mx0,  t, p);
    GEN imx9p2 = FPM_FPXQ_MUL(frob_pows[2], imx9, t, p);
    GEN imx7p2 = FPM_FPXQ_MUL(frob_pows[2], imx7, t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx1p2 = FPM_FPXQ_MUL(frob_pows[2], imx1, t, p);

    GEN mx4p3  = FPM_FPXQ_MUL(frob_pows[3], mx4,  t, p);
    GEN mx2p3  = FPM_FPXQ_MUL(frob_pows[3], mx2,  t, p);
    GEN mx0p3  = FPM_FPXQ_MUL(frob_pows[3], mx0,  t, p);
    GEN imx3p3 = FPM_FPXQ_MUL(frob_pows[3], imx3, t, p);
    GEN imx1p3 = FPM_FPXQ_MUL(frob_pows[3], imx1, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx8, mx11,   t, p);
    y0 = FPXQ_MUL(y0, mx9,     t, p);
    y0 = FPXQ_MUL(y0, imx3,    t, p);
    y0 = FPXQ_MUL(y0, imx10,   t, p);
    y0 = FPXQ_MUL(y0, mx8p1,   t, p);
    y0 = FPXQ_MUL(y0, mx11p1,  t, p);
    y0 = FPXQ_MUL(y0, mx9p1,   t, p);
    y0 = FPXQ_MUL(y0, mx0p1,   t, p);
    y0 = FPXQ_MUL(y0, imx7p1,  t, p);
    y0 = FPXQ_MUL(y0, imx10p1, t, p);
    y0 = FPXQ_MUL(y0, mx10p2,  t, p);
    y0 = FPXQ_MUL(y0, mx6p2,   t, p);
    y0 = FPXQ_MUL(y0, mx0p2,   t, p);
    y0 = FPXQ_MUL(y0, mx4p3,   t, p);
    y0 = FPXQ_MUL(y0, mx0p3,   t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx5, mx6,   t, p);
    y1 = FPXQ_MUL(y1, imx7,   t, p);
    y1 = FPXQ_MUL(y1, imx4p1, t, p);
    y1 = FPXQ_MUL(y1, imx1p1, t, p);
    y1 = FPXQ_MUL(y1, imx9p2, t, p);
    y1 = FPXQ_MUL(y1, imx7p2, t, p);
    y1 = FPXQ_MUL(y1, imx1p2, t, p);
    y1 = FPXQ_MUL(y1, imx3p3, t, p);
    y1 = FPXQ_MUL(y1, imx1p3, t, p);

    GEN y2 = FPXQ_MUL(mx8p2, mx2p3, t, p);

    GEN y3 = FPXQ_MUL(mx0, mx4p2, t, p);

    GEN y4;
    y4 = FPXQ_MUL(mx2, imx1,  t, p);
    y4 = FPXQ_MUL(y4, mx5p1,  t, p);
    y4 = FPXQ_MUL(y4, mx2p1,  t, p);
    y4 = FPXQ_MUL(y4, imx3p2, t, p);

    GEN y5 = mx2p2;

    GEN T0 = FPXQ_MUL(y5, y4, t, p);
    GEN T1 = FPXQ_MUL(y5, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y3, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_MUL(T2, y2, t, p);
    GEN T5 = FPXQ_MUL(T3, T2, t, p);
    GEN T6 = FPXQ_MUL(T4, y1, t, p);
    GEN T7 = FPXQ_MUL(T5, T4, t, p);
    GEN T8 = FPXQ_SQR(T6, t, p);
    GEN T9 = FPXQ_MUL(T8, T7, t, p);
    
    return T9;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_7_1333(pairing_data_t* const pdata, GEN unreduced)
{       
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;

    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0, x, t, p);
    GEN mx2 = FPXQ_POW(mx1, x, t, p);
    GEN mx3 = FPXQ_POW(mx2, x, t, p);
    GEN mx4 = FPXQ_POW(mx3, x, t, p);
    GEN mx5 = FPXQ_POW(mx4, x, t, p);
    GEN mx6 = FPXQ_POW(mx5, x, t, p);
    GEN mx7 = FPXQ_POW(mx6, x, t, p);

    GEN imx2 = FPXQ_INV(mx2, t, p);
    GEN imx3 = FPXQ_INV(mx3, t, p);
    GEN imx6 = FPXQ_INV(mx6, t, p);
    GEN imx7 = FPXQ_INV(mx7, t, p);

    GEN mx5p1  = FPM_FPXQ_MUL(frob_pows[1], mx5,  t, p);
    GEN mx4p1  = FPM_FPXQ_MUL(frob_pows[1], mx4,  t, p);
    GEN mx1p1  = FPM_FPXQ_MUL(frob_pows[1], mx1,  t, p);
    GEN mx0p1  = FPM_FPXQ_MUL(frob_pows[1], mx0,  t, p);
    GEN imx6p1 = FPM_FPXQ_MUL(frob_pows[1], imx6, t, p);
    GEN imx3p1 = FPM_FPXQ_MUL(frob_pows[1], imx3, t, p);
    GEN imx7p1 = FPM_FPXQ_MUL(frob_pows[1], imx7, t, p);

    GEN mx5p2  = FPM_FPXQ_MUL(frob_pows[2], mx5,  t, p);
    GEN mx4p2  = FPM_FPXQ_MUL(frob_pows[2], mx4,  t, p);
    GEN mx1p2  = FPM_FPXQ_MUL(frob_pows[2], mx1,  t, p);
    GEN mx0p2  = FPM_FPXQ_MUL(frob_pows[2], mx0,  t, p);
    GEN imx6p2 = FPM_FPXQ_MUL(frob_pows[2], imx6, t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx2p2 = FPM_FPXQ_MUL(frob_pows[2], imx2, t, p);
    GEN imx7p2 = FPM_FPXQ_MUL(frob_pows[2], imx7, t, p);

    GEN mx6p3  = FPM_FPXQ_MUL(frob_pows[3], mx6,  t, p);
    GEN mx5p3  = FPM_FPXQ_MUL(frob_pows[3], mx5,  t, p);
    GEN mx4p3  = FPM_FPXQ_MUL(frob_pows[3], mx4,  t, p);
    GEN mx2p3  = FPM_FPXQ_MUL(frob_pows[3], mx2,  t, p);
    GEN mx1p3  = FPM_FPXQ_MUL(frob_pows[3], mx1,  t, p);
    GEN mx0p3  = FPM_FPXQ_MUL(frob_pows[3], mx0,  t, p);
    GEN imx3p3 = FPM_FPXQ_MUL(frob_pows[3], imx3, t, p);

    GEN mx6p4 = FPM_FPXQ_MUL(frob_pows[4], mx6, t, p);
    GEN mx5p4 = FPM_FPXQ_MUL(frob_pows[4], mx5, t, p);
    GEN mx4p4 = FPM_FPXQ_MUL(frob_pows[4], mx4, t, p);
    GEN mx2p4 = FPM_FPXQ_MUL(frob_pows[4], mx2, t, p);
    GEN mx1p4 = FPM_FPXQ_MUL(frob_pows[4], mx1, t, p);
    GEN mx0p4 = FPM_FPXQ_MUL(frob_pows[4], mx0, t, p);

    GEN mx2p5 = FPM_FPXQ_MUL(frob_pows[5], mx2, t, p);
    GEN mx1p5 = FPM_FPXQ_MUL(frob_pows[5], mx1, t, p);
    GEN mx0p5 = FPM_FPXQ_MUL(frob_pows[5], mx0, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx5, mx3,   t, p);
    y0 = FPXQ_MUL(y0, mx5p1,  t, p);
    y0 = FPXQ_MUL(y0, imx6p1, t, p);
    y0 = FPXQ_MUL(y0, imx7p1, t, p);
    y0 = FPXQ_MUL(y0, mx5p2,  t, p);
    y0 = FPXQ_MUL(y0, mx4p2,  t, p);
    y0 = FPXQ_MUL(y0, mx0p2,  t, p);
    y0 = FPXQ_MUL(y0, imx6p2, t, p);
    y0 = FPXQ_MUL(y0, imx2p2, t, p);
    y0 = FPXQ_MUL(y0, imx7p2, t, p);
    y0 = FPXQ_MUL(y0, mx6p3,  t, p);
    y0 = FPXQ_MUL(y0, mx4p3,  t, p);
    y0 = FPXQ_MUL(y0, mx0p3,  t, p);
    y0 = FPXQ_MUL(y0, imx3p3, t, p);
    y0 = FPXQ_MUL(y0, mx6p4,  t, p);
    y0 = FPXQ_MUL(y0, mx4p4,  t, p);
    y0 = FPXQ_MUL(y0, mx0p4,  t, p);
    y0 = FPXQ_MUL(y0, mx2p5,  t, p);
    y0 = FPXQ_MUL(y0, mx0p5,  t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx4, mx4p1, t, p);
    y1 = FPXQ_MUL(y1, mx5p3,  t, p);
    y1 = FPXQ_MUL(y1, mx5p4,  t, p);
    y1 = FPXQ_MUL(y1, mx1p5,  t, p);

    GEN y2 = FPXQ_MUL(imx3p1, mx2p3, t, p);

    GEN y3 = FPXQ_MUL(mx1, mx0, t, p);

    GEN y4;
    y4 = FPXQ_MUL(mx1p1, mx0p1, t, p);
    y4 = FPXQ_MUL(y4, mx1p2,    t, p);
    y4 = FPXQ_MUL(y4, imx3p2,   t, p);
    y4 = FPXQ_MUL(y4, mx1p3,    t, p);
    y4 = FPXQ_MUL(y4, mx2p4,    t, p);

    GEN y5 = mx1p4;
    
    GEN T0 = FPXQ_MUL(y5, y4, t, p);
    GEN T1 = FPXQ_MUL(y5, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y3, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_MUL(T2, y2, t, p);
    GEN T5 = FPXQ_MUL(T3, T2, t, p);
    GEN T6 = FPXQ_MUL(T4, y1, t, p);
    GEN T7 = FPXQ_MUL(T5, T4, t, p);
    GEN T8 = FPXQ_SQR(T6, t, p);
    GEN T9 = FPXQ_MUL(T8, T7, t, p);
    
    return T9;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_8_1250(pairing_data_t* const pdata, GEN unreduced)
{    
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0, x, t, p);
    GEN mx2 = FPXQ_POW(mx1, x, t, p);
    GEN mx3 = FPXQ_POW(mx2, x, t, p);
    GEN mx4 = FPXQ_POW(mx3, x, t, p);
    GEN mx5 = FPXQ_POW(mx4, x, t, p);
    GEN mx6 = FPXQ_POW(mx5, x, t, p);
    GEN mx7 = FPXQ_POW(mx6, x, t, p);
    GEN mx8 = FPXQ_POW(mx7, x, t, p);

    GEN imx1 = FPXQ_INV(mx1, t, p);
    GEN imx3 = FPXQ_INV(mx3, t, p);
    GEN imx6 = FPXQ_INV(mx6, t, p);
    GEN imx7 = FPXQ_INV(mx7, t, p);
    GEN imx8 = FPXQ_INV(mx8, t, p);

    GEN imx8p1 = FPM_FPXQ_MUL(frob_pows[1], imx8, t, p);
    GEN imx6p1 = FPM_FPXQ_MUL(frob_pows[1], imx6, t, p);
    GEN imx3p1 = FPM_FPXQ_MUL(frob_pows[1], imx3, t, p);
    GEN imx7p1 = FPM_FPXQ_MUL(frob_pows[1], imx7, t, p);

    GEN mx7p2  = FPM_FPXQ_MUL(frob_pows[2], mx7,  t, p);
    GEN mx6p2  = FPM_FPXQ_MUL(frob_pows[2], mx6,  t, p);
    GEN mx5p2  = FPM_FPXQ_MUL(frob_pows[2], mx5,  t, p);
    GEN mx2p2  = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx1p2 = FPM_FPXQ_MUL(frob_pows[2], imx1, t, p);

    GEN mx2p3 = FPM_FPXQ_MUL(frob_pows[3], mx2, t, p);
    GEN mx1p3 = FPM_FPXQ_MUL(frob_pows[3], mx1, t, p);
    GEN mx0p3 = FPM_FPXQ_MUL(frob_pows[3], mx0, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx4, mx5,   t, p);
    y0 = FPXQ_MUL(y0, mx3,    t, p);
    y0 = FPXQ_MUL(y0, imx8p1, t, p);
    y0 = FPXQ_MUL(y0, imx6p1, t, p);
    y0 = FPXQ_MUL(y0, imx7p1, t, p);
    y0 = FPXQ_MUL(y0, mx7p2,  t, p);
    y0 = FPXQ_MUL(y0, mx6p2,  t, p);
    y0 = FPXQ_MUL(y0, mx5p2,  t, p);
    y0 = FPXQ_MUL(y0, imx3p2, t, p);
    y0 = FPXQ_MUL(y0, imx1p2, t, p);
    y0 = FPXQ_MUL(y0, mx2p3,  t, p);
    y0 = FPXQ_MUL(y0, mx1p3,  t, p);
    y0 = FPXQ_MUL(y0, mx0p3,  t, p);

    GEN y1 = mx2p2;

    GEN y2 = FPXQ_MUL(mx0, imx3p1, t, p);
        
    GEN T0 = FPXQ_MUL(y2, y1, t, p);
    GEN T1 = FPXQ_MUL(y2, y0, t, p);
    GEN T2 = FPXQ_SQR(T0, t, p);
    GEN T3 = FPXQ_MUL(T2, T1, t, p);
    
    return T3;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_9_1333(pairing_data_t* const pdata, GEN unreduced)
{    
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0,  x, t, p);
    GEN mx2 = FPXQ_POW(mx1,  x, t, p);
    GEN mx3 = FPXQ_POW(mx2,  x, t, p);
    GEN mx4 = FPXQ_POW(mx3,  x, t, p);
    GEN mx5 = FPXQ_POW(mx4,  x, t, p);
    GEN mx6 = FPXQ_POW(mx5,  x, t, p);
    GEN mx7 = FPXQ_POW(mx6,  x, t, p);

    GEN imx1 = FPXQ_INV(mx1, t, p);
    GEN imx2 = FPXQ_INV(mx2, t, p);
    GEN imx3 = FPXQ_INV(mx3, t, p);
    GEN imx4 = FPXQ_INV(mx4, t, p);
    GEN imx5 = FPXQ_INV(mx5, t, p);
    GEN imx6 = FPXQ_INV(mx6, t, p);
    GEN imx7 = FPXQ_INV(mx7, t, p);

    GEN mx5p1  = FPM_FPXQ_MUL(frob_pows[1], mx5,  t, p);
    GEN mx2p1  = FPM_FPXQ_MUL(frob_pows[1], mx2,  t, p);
    GEN imx6p1 = FPM_FPXQ_MUL(frob_pows[1], imx6, t, p);
    GEN imx4p1 = FPM_FPXQ_MUL(frob_pows[1], imx4, t, p);

    GEN mx4p2  = FPM_FPXQ_MUL(frob_pows[2], mx4,  t, p);
    GEN mx2p2  = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx1p2  = FPM_FPXQ_MUL(frob_pows[2], mx1,  t, p);
    GEN mx0p2  = FPM_FPXQ_MUL(frob_pows[2], mx0,  t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx5p2 = FPM_FPXQ_MUL(frob_pows[2], imx5, t, p);

    GEN mx6p3  = FPM_FPXQ_MUL(frob_pows[3], mx6,  t, p);
    GEN mx3p3  = FPM_FPXQ_MUL(frob_pows[3], mx3,  t, p);
    GEN imx7p3 = FPM_FPXQ_MUL(frob_pows[3], imx7, t, p);
    GEN imx5p3 = FPM_FPXQ_MUL(frob_pows[3], imx5, t, p);

    GEN mx5p4  = FPM_FPXQ_MUL(frob_pows[4], mx5,  t, p);
    GEN mx3p4  = FPM_FPXQ_MUL(frob_pows[4], mx3,  t, p);
    GEN mx2p4  = FPM_FPXQ_MUL(frob_pows[4], mx2,  t, p);
    GEN mx1p4  = FPM_FPXQ_MUL(frob_pows[4], mx1,  t, p);
    GEN imx6p4 = FPM_FPXQ_MUL(frob_pows[4], imx6, t, p);
    GEN imx4p4 = FPM_FPXQ_MUL(frob_pows[4], imx4, t, p);

    GEN mx2p5  = FPM_FPXQ_MUL(frob_pows[5], mx2,  t, p);
    GEN mx0p5  = FPM_FPXQ_MUL(frob_pows[5], mx0,  t, p);
    GEN imx1p5 = FPM_FPXQ_MUL(frob_pows[5], imx1, t, p);

    //
    // Coef 1
    //
    GEN y0;
    y0 = FPXQ_MUL(mx3, imx4,   t, p);
    y0 = FPXQ_MUL(y0,  imx2,   t, p);
    y0 = FPXQ_MUL(y0,  mx5p1,  t, p);
    y0 = FPXQ_MUL(y0,  imx6p1, t, p);
    y0 = FPXQ_MUL(y0,  imx4p1, t, p);
    y0 = FPXQ_MUL(y0,  mx4p2,  t, p);
    y0 = FPXQ_MUL(y0,  mx2p2,  t, p);
    y0 = FPXQ_MUL(y0,  mx0p2,  t, p);
    y0 = FPXQ_MUL(y0,  imx3p2, t, p);
    y0 = FPXQ_MUL(y0,  imx5p2, t, p);
    y0 = FPXQ_MUL(y0,  mx6p3,  t, p);
    y0 = FPXQ_MUL(y0,  imx7p3, t, p);
    y0 = FPXQ_MUL(y0,  imx5p3, t, p);
    y0 = FPXQ_MUL(y0,  mx5p4,  t, p);
    y0 = FPXQ_MUL(y0,  mx3p4,  t, p);
    y0 = FPXQ_MUL(y0,  mx1p4,  t, p);
    y0 = FPXQ_MUL(y0,  imx6p4, t, p);
    y0 = FPXQ_MUL(y0,  imx4p4, t, p);
    y0 = FPXQ_MUL(y0,  mx2p5,  t, p);
    y0 = FPXQ_MUL(y0,  mx0p5,  t, p);
    y0 = FPXQ_MUL(y0,  imx1p5, t, p);

    //
    // Coef 2
    //
    GEN y1 = FPXQ_MUL(mx1p2, mx2p4, t, p);

    //
    // Coef 3
    //
    GEN y2;
    y2 = FPXQ_MUL(mx0, mx2p1, t, p);
    y2 = FPXQ_MUL(y2,  mx3p3, t, p);

    GEN T0 = FPXQ_MUL(y2, y1, t, p);
    GEN T1 = FPXQ_MUL(y2, y0, t, p);
    GEN T2 = FPXQ_SQR(T0, t, p);
    GEN T3 = FPXQ_MUL(T2, T1, t, p);
    
    return T3;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_11_1200(pairing_data_t* const pdata, GEN unreduced)
{    
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);
    GEN mx12 = FPXQ_POW(mx11, x, t, p);
    GEN mx13 = FPXQ_POW(mx12, x, t, p);
    GEN mx14 = FPXQ_POW(mx13, x, t, p);
    GEN mx15 = FPXQ_POW(mx14, x, t, p);
    GEN mx16 = FPXQ_POW(mx15, x, t, p);
    GEN mx17 = FPXQ_POW(mx16, x, t, p);
    GEN mx18 = FPXQ_POW(mx17, x, t, p);
    GEN mx19 = FPXQ_POW(mx18, x, t, p);
    GEN mx20 = FPXQ_POW(mx19, x, t, p);
    GEN mx21 = FPXQ_POW(mx20, x, t, p);
    GEN mx22 = FPXQ_POW(mx21, x, t, p);
    GEN mx23 = FPXQ_POW(mx22, x, t, p);

    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);
    GEN imx13 = FPXQ_INV(mx13, t, p);
    GEN imx15 = FPXQ_INV(mx15, t, p);
    GEN imx16 = FPXQ_INV(mx16, t, p);
    GEN imx17 = FPXQ_INV(mx17, t, p);
    GEN imx18 = FPXQ_INV(mx18, t, p);
    GEN imx19 = FPXQ_INV(mx19, t, p);
    GEN imx20 = FPXQ_INV(mx20, t, p);
    GEN imx21 = FPXQ_INV(mx21, t, p);
    GEN imx22 = FPXQ_INV(mx22, t, p);

    GEN mx14p1  = FPM_FPXQ_MUL(frob_pows[1], mx14,  t, p);
    GEN mx11p1  = FPM_FPXQ_MUL(frob_pows[1], mx11,  t, p);
    GEN mx9p1   = FPM_FPXQ_MUL(frob_pows[1], mx9,   t, p);
    GEN mx23p1  = FPM_FPXQ_MUL(frob_pows[1], mx23,  t, p);
    GEN mx2p1   = FPM_FPXQ_MUL(frob_pows[1], mx2,   t, p);
    GEN mx21p1  = FPM_FPXQ_MUL(frob_pows[1], mx21,  t, p);
    GEN mx0p1   = FPM_FPXQ_MUL(frob_pows[1], mx0,   t, p);
    GEN imx13p1 = FPM_FPXQ_MUL(frob_pows[1], imx13, t, p);
    GEN imx10p1 = FPM_FPXQ_MUL(frob_pows[1], imx10, t, p);
    GEN imx3p1  = FPM_FPXQ_MUL(frob_pows[1], imx3,  t, p);
    GEN imx22p1 = FPM_FPXQ_MUL(frob_pows[1], imx22, t, p);
    GEN imx1p1  = FPM_FPXQ_MUL(frob_pows[1], imx1,  t, p);

    GEN mx20p2  = FPM_FPXQ_MUL(frob_pows[2], mx20,  t, p);
    GEN mx14p2  = FPM_FPXQ_MUL(frob_pows[2], mx14,  t, p);
    GEN mx10p2  = FPM_FPXQ_MUL(frob_pows[2], mx10,  t, p);
    GEN mx22p2  = FPM_FPXQ_MUL(frob_pows[2], mx22,  t, p);
    GEN mx12p2  = FPM_FPXQ_MUL(frob_pows[2], mx12,  t, p);
    GEN mx2p2   = FPM_FPXQ_MUL(frob_pows[2], mx2,   t, p);
    GEN mx8p2   = FPM_FPXQ_MUL(frob_pows[2], mx8,   t, p);
    GEN mx0p2   = FPM_FPXQ_MUL(frob_pows[2], mx0,   t, p);
    GEN imx19p2 = FPM_FPXQ_MUL(frob_pows[2], imx19, t, p);
    GEN imx13p2 = FPM_FPXQ_MUL(frob_pows[2], imx13, t, p);
    GEN imx9p2  = FPM_FPXQ_MUL(frob_pows[2], imx9,  t, p);
    GEN imx3p2  = FPM_FPXQ_MUL(frob_pows[2], imx3,  t, p);
    GEN imx21p2 = FPM_FPXQ_MUL(frob_pows[2], imx21, t, p);
    GEN imx1p2  = FPM_FPXQ_MUL(frob_pows[2], imx1,  t, p);

    GEN mx19p3  = FPM_FPXQ_MUL(frob_pows[3], mx19,  t, p);
    GEN mx14p3  = FPM_FPXQ_MUL(frob_pows[3], mx14,  t, p);
    GEN mx10p3  = FPM_FPXQ_MUL(frob_pows[3], mx10,  t, p);
    GEN mx12p3  = FPM_FPXQ_MUL(frob_pows[3], mx12,  t, p);
    GEN mx2p3   = FPM_FPXQ_MUL(frob_pows[3], mx2,   t, p);
    GEN mx0p3   = FPM_FPXQ_MUL(frob_pows[3], mx0,   t, p);
    GEN imx20p3 = FPM_FPXQ_MUL(frob_pows[3], imx20, t, p);
    GEN imx13p3 = FPM_FPXQ_MUL(frob_pows[3], imx13, t, p);
    GEN imx9p3  = FPM_FPXQ_MUL(frob_pows[3], imx9,  t, p);
    GEN imx3p3  = FPM_FPXQ_MUL(frob_pows[3], imx3,  t, p);
    GEN imx8p3  = FPM_FPXQ_MUL(frob_pows[3], imx8,  t, p);
    GEN imx18p3 = FPM_FPXQ_MUL(frob_pows[3], imx18, t, p);
    GEN imx1p3  = FPM_FPXQ_MUL(frob_pows[3], imx1,  t, p);

    GEN mx19p4  = FPM_FPXQ_MUL(frob_pows[4], mx19,  t, p);
    GEN mx14p4  = FPM_FPXQ_MUL(frob_pows[4], mx14,  t, p);
    GEN mx9p4   = FPM_FPXQ_MUL(frob_pows[4], mx9,   t, p);
    GEN mx7p4   = FPM_FPXQ_MUL(frob_pows[4], mx7,   t, p);
    GEN mx12p4  = FPM_FPXQ_MUL(frob_pows[4], mx12,  t, p);
    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,   t, p);
    GEN mx0p4   = FPM_FPXQ_MUL(frob_pows[4], mx0,   t, p);
    GEN imx20p4 = FPM_FPXQ_MUL(frob_pows[4], imx20, t, p);
    GEN imx13p4 = FPM_FPXQ_MUL(frob_pows[4], imx13, t, p);
    GEN imx6p4  = FPM_FPXQ_MUL(frob_pows[4], imx6,  t, p);
    GEN imx3p4  = FPM_FPXQ_MUL(frob_pows[4], imx3,  t, p);
    GEN imx8p4  = FPM_FPXQ_MUL(frob_pows[4], imx8,  t, p);
    GEN imx18p4 = FPM_FPXQ_MUL(frob_pows[4], imx18, t, p);
    GEN imx1p4  = FPM_FPXQ_MUL(frob_pows[4], imx1,  t, p);

    GEN mx16p5  = FPM_FPXQ_MUL(frob_pows[5], mx16,  t, p);
    GEN mx14p5  = FPM_FPXQ_MUL(frob_pows[5], mx14,  t, p);
    GEN mx6p5   = FPM_FPXQ_MUL(frob_pows[5], mx6,   t, p);
    GEN mx12p5  = FPM_FPXQ_MUL(frob_pows[5], mx12,  t, p);
    GEN mx2p5   = FPM_FPXQ_MUL(frob_pows[5], mx2,   t, p);
    GEN mx18p5  = FPM_FPXQ_MUL(frob_pows[5], mx18,  t, p);
    GEN mx0p5   = FPM_FPXQ_MUL(frob_pows[5], mx0,   t, p);
    GEN imx19p5 = FPM_FPXQ_MUL(frob_pows[5], imx19, t, p);
    GEN imx17p5 = FPM_FPXQ_MUL(frob_pows[5], imx17, t, p);
    GEN imx13p5 = FPM_FPXQ_MUL(frob_pows[5], imx13, t, p);
    GEN imx7p5  = FPM_FPXQ_MUL(frob_pows[5], imx7,  t, p);
    GEN imx3p5  = FPM_FPXQ_MUL(frob_pows[5], imx3,  t, p);
    GEN imx5p5  = FPM_FPXQ_MUL(frob_pows[5], imx5,  t, p);
    GEN imx1p5  = FPM_FPXQ_MUL(frob_pows[5], imx1,  t, p);

    GEN mx17p6  = FPM_FPXQ_MUL(frob_pows[6], mx17,  t, p);
    GEN mx14p6  = FPM_FPXQ_MUL(frob_pows[6], mx14,  t, p);
    GEN mx6p6   = FPM_FPXQ_MUL(frob_pows[6], mx6,   t, p);
    GEN mx12p6  = FPM_FPXQ_MUL(frob_pows[6], mx12,  t, p);
    GEN mx2p6   = FPM_FPXQ_MUL(frob_pows[6], mx2,   t, p);
    GEN mx5p6   = FPM_FPXQ_MUL(frob_pows[6], mx5,   t, p);
    GEN mx15p6  = FPM_FPXQ_MUL(frob_pows[6], mx15,  t, p);
    GEN mx0p6   = FPM_FPXQ_MUL(frob_pows[6], mx0,   t, p);
    GEN imx16p6 = FPM_FPXQ_MUL(frob_pows[6], imx16, t, p);
    GEN imx13p6 = FPM_FPXQ_MUL(frob_pows[6], imx13, t, p);
    GEN imx7p6  = FPM_FPXQ_MUL(frob_pows[6], imx7,  t, p);
    GEN imx3p6  = FPM_FPXQ_MUL(frob_pows[6], imx3,  t, p);
    GEN imx1p6  = FPM_FPXQ_MUL(frob_pows[6], imx1,  t, p);

    GEN mx17p7  = FPM_FPXQ_MUL(frob_pows[7], mx17,  t, p);
    GEN mx14p7  = FPM_FPXQ_MUL(frob_pows[7], mx14,  t, p);
    GEN mx12p7  = FPM_FPXQ_MUL(frob_pows[7], mx12,  t, p);
    GEN mx2p7   = FPM_FPXQ_MUL(frob_pows[7], mx2,   t, p);
    GEN mx5p7   = FPM_FPXQ_MUL(frob_pows[7], mx5,   t, p);
    GEN mx15p7  = FPM_FPXQ_MUL(frob_pows[7], mx15,  t, p);
    GEN mx0p7   = FPM_FPXQ_MUL(frob_pows[7], mx0,   t, p);
    GEN imx16p7 = FPM_FPXQ_MUL(frob_pows[7], imx16, t, p);
    GEN imx13p7 = FPM_FPXQ_MUL(frob_pows[7], imx13, t, p);
    GEN imx6p7  = FPM_FPXQ_MUL(frob_pows[7], imx6,  t, p);
    GEN imx4p7  = FPM_FPXQ_MUL(frob_pows[7], imx4,  t, p);
    GEN imx1p7  = FPM_FPXQ_MUL(frob_pows[7], imx1,  t, p);

    GEN mx16p8  = FPM_FPXQ_MUL(frob_pows[8], mx16,  t, p);
    GEN mx14p8  = FPM_FPXQ_MUL(frob_pows[8], mx14,  t, p);
    GEN mx4p8   = FPM_FPXQ_MUL(frob_pows[8], mx4,   t, p);
    GEN mx12p8  = FPM_FPXQ_MUL(frob_pows[8], mx12,  t, p);
    GEN mx2p8   = FPM_FPXQ_MUL(frob_pows[8], mx2,   t, p);
    GEN mx0p8   = FPM_FPXQ_MUL(frob_pows[8], mx0,   t, p);
    GEN imx13p8 = FPM_FPXQ_MUL(frob_pows[8], imx13, t, p);
    GEN imx3p8  = FPM_FPXQ_MUL(frob_pows[8], imx3,  t, p);
    GEN imx15p8 = FPM_FPXQ_MUL(frob_pows[8], imx15, t, p);
    GEN imx1p8  = FPM_FPXQ_MUL(frob_pows[8], imx1,  t, p);

    GEN mx4p9  = FPM_FPXQ_MUL(frob_pows[9], mx4,  t, p);
    GEN mx2p9  = FPM_FPXQ_MUL(frob_pows[9], mx2,  t, p);
    GEN mx0p9  = FPM_FPXQ_MUL(frob_pows[9], mx0,  t, p);
    GEN imx3p9 = FPM_FPXQ_MUL(frob_pows[9], imx3, t, p);
    GEN imx1p9 = FPM_FPXQ_MUL(frob_pows[9], imx1, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx14, mx23,  t, p);
    y0 = FPXQ_MUL(y0, mx21,    t, p);
    y0 = FPXQ_MUL(y0, imx3,    t, p);
    y0 = FPXQ_MUL(y0, imx22,   t, p);
    y0 = FPXQ_MUL(y0, mx14p1,  t, p);
    y0 = FPXQ_MUL(y0, mx9p1,   t, p);
    y0 = FPXQ_MUL(y0, mx23p1,  t, p);
    y0 = FPXQ_MUL(y0, mx21p1,  t, p);
    y0 = FPXQ_MUL(y0, mx0p1,   t, p);
    y0 = FPXQ_MUL(y0, imx13p1, t, p);
    y0 = FPXQ_MUL(y0, imx3p1,  t, p);
    y0 = FPXQ_MUL(y0, imx22p1, t, p);
    y0 = FPXQ_MUL(y0, mx14p2,  t, p);
    y0 = FPXQ_MUL(y0, mx22p2,  t, p);
    y0 = FPXQ_MUL(y0, mx12p2,  t, p);
    y0 = FPXQ_MUL(y0, mx8p2,   t, p);
    y0 = FPXQ_MUL(y0, mx0p2,   t, p);
    y0 = FPXQ_MUL(y0, imx19p2, t, p);
    y0 = FPXQ_MUL(y0, imx13p2, t, p);
    y0 = FPXQ_MUL(y0, imx3p2,  t, p);
    y0 = FPXQ_MUL(y0, mx19p3,  t, p);
    y0 = FPXQ_MUL(y0, mx14p3,  t, p);
    y0 = FPXQ_MUL(y0, mx10p3,  t, p);
    y0 = FPXQ_MUL(y0, mx12p3,  t, p);
    y0 = FPXQ_MUL(y0, mx0p3,   t, p);
    y0 = FPXQ_MUL(y0, imx20p3, t, p);
    y0 = FPXQ_MUL(y0, imx13p3, t, p);
    y0 = FPXQ_MUL(y0, imx9p3,  t, p);
    y0 = FPXQ_MUL(y0, imx3p3,  t, p);
    y0 = FPXQ_MUL(y0, imx18p3, t, p);
    y0 = FPXQ_MUL(y0, mx19p4,  t, p);
    y0 = FPXQ_MUL(y0, mx14p4,  t, p);
    y0 = FPXQ_MUL(y0, mx9p4,   t, p);
    y0 = FPXQ_MUL(y0, mx12p4,  t, p);
    y0 = FPXQ_MUL(y0, mx0p4,   t, p);
    y0 = FPXQ_MUL(y0, imx20p4, t, p);
    y0 = FPXQ_MUL(y0, imx13p4, t, p);
    y0 = FPXQ_MUL(y0, imx6p4,  t, p);
    y0 = FPXQ_MUL(y0, imx3p4,  t, p);
    y0 = FPXQ_MUL(y0, imx18p4, t, p);
    y0 = FPXQ_MUL(y0, mx16p5,  t, p);
    y0 = FPXQ_MUL(y0, mx14p5,  t, p);
    y0 = FPXQ_MUL(y0, mx12p5,  t, p);
    y0 = FPXQ_MUL(y0, mx0p5,   t, p);
    y0 = FPXQ_MUL(y0, imx19p5, t, p);
    y0 = FPXQ_MUL(y0, imx13p5, t, p);
    y0 = FPXQ_MUL(y0, imx3p5,  t, p);
    y0 = FPXQ_MUL(y0, imx5p5,  t, p);
    y0 = FPXQ_MUL(y0, mx17p6,  t, p);
    y0 = FPXQ_MUL(y0, mx14p6,  t, p);
    y0 = FPXQ_MUL(y0, mx6p6,   t, p);
    y0 = FPXQ_MUL(y0, mx12p6,  t, p);
    y0 = FPXQ_MUL(y0, mx15p6,  t, p);
    y0 = FPXQ_MUL(y0, mx0p6,   t, p);
    y0 = FPXQ_MUL(y0, imx16p6, t, p);
    y0 = FPXQ_MUL(y0, imx13p6, t, p);
    y0 = FPXQ_MUL(y0, imx7p6,  t, p);
    y0 = FPXQ_MUL(y0, imx3p6,  t, p);
    y0 = FPXQ_MUL(y0, mx17p7,  t, p);
    y0 = FPXQ_MUL(y0, mx14p7,  t, p);
    y0 = FPXQ_MUL(y0, mx12p7,  t, p);
    y0 = FPXQ_MUL(y0, mx15p7,  t, p);
    y0 = FPXQ_MUL(y0, mx0p7,   t, p);
    y0 = FPXQ_MUL(y0, imx16p7, t, p);
    y0 = FPXQ_MUL(y0, imx13p7, t, p);
    y0 = FPXQ_MUL(y0, imx6p7,  t, p);
    y0 = FPXQ_MUL(y0, mx16p8,  t, p);
    y0 = FPXQ_MUL(y0, mx12p8,  t, p);
    y0 = FPXQ_MUL(y0, mx0p8,   t, p);
    y0 = FPXQ_MUL(y0, mx4p9,   t, p);
    y0 = FPXQ_MUL(y0, mx0p9,   t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx11, mx12,  t, p);
    y1 = FPXQ_MUL(y1, imx13,   t, p);
    y1 = FPXQ_MUL(y1, imx10p1, t, p);
    y1 = FPXQ_MUL(y1, imx1p1,  t, p);
    y1 = FPXQ_MUL(y1, mx20p2,  t, p);
    y1 = FPXQ_MUL(y1, imx21p2, t, p);
    y1 = FPXQ_MUL(y1, imx1p2,  t, p);
    y1 = FPXQ_MUL(y1, imx8p3,  t, p);
    y1 = FPXQ_MUL(y1, imx1p3,  t, p);
    y1 = FPXQ_MUL(y1, mx7p4,   t, p);
    y1 = FPXQ_MUL(y1, imx1p4,  t, p);
    y1 = FPXQ_MUL(y1, mx18p5,  t, p);
    y1 = FPXQ_MUL(y1, imx17p5, t, p);
    y1 = FPXQ_MUL(y1, imx1p5,  t, p);
    y1 = FPXQ_MUL(y1, mx5p6,   t, p);
    y1 = FPXQ_MUL(y1, imx1p6,  t, p);
    y1 = FPXQ_MUL(y1, imx4p7,  t, p);
    y1 = FPXQ_MUL(y1, imx1p7,  t, p);
    y1 = FPXQ_MUL(y1, imx13p8, t, p);
    y1 = FPXQ_MUL(y1, imx15p8, t, p);
    y1 = FPXQ_MUL(y1, imx1p8,  t, p);
    y1 = FPXQ_MUL(y1, imx3p9,  t, p);
    y1 = FPXQ_MUL(y1, imx1p9,  t, p);

    GEN y2 = FPXQ_MUL(mx14p8, mx2p9, t, p);

    GEN y3;
    y3 = FPXQ_MUL(mx0, mx10p2, t, p);
    y3 = FPXQ_MUL(y3, imx9p2,  t, p);
    y3 = FPXQ_MUL(y3, mx6p5,   t, p);
    y3 = FPXQ_MUL(y3, imx7p5,  t, p);
    y3 = FPXQ_MUL(y3, mx4p8,   t, p);

    GEN y4;
    y4 = FPXQ_MUL(mx2, imx1,  t, p);
    y4 = FPXQ_MUL(y4, mx11p1, t, p);
    y4 = FPXQ_MUL(y4, mx2p1,  t, p);
    y4 = FPXQ_MUL(y4, mx2p2,  t, p);
    y4 = FPXQ_MUL(y4, mx2p3,  t, p);
    y4 = FPXQ_MUL(y4, mx2p4,  t, p);
    y4 = FPXQ_MUL(y4, imx8p4, t, p);
    y4 = FPXQ_MUL(y4, mx2p5,  t, p);
    y4 = FPXQ_MUL(y4, mx2p6,  t, p);
    y4 = FPXQ_MUL(y4, mx2p7,  t, p);
    y4 = FPXQ_MUL(y4, mx5p7,  t, p);
    y4 = FPXQ_MUL(y4, imx3p8, t, p);

    GEN y5 = mx2p8;
    
    GEN T0 = FPXQ_MUL(y5, y4, t, p);
    GEN T1 = FPXQ_MUL(y5, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y3, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_MUL(T2, y2, t, p);
    GEN T5 = FPXQ_MUL(T3, T2, t, p);
    GEN T6 = FPXQ_MUL(T4, y1, t, p);
    GEN T7 = FPXQ_MUL(T5, T4, t, p);
    GEN T8 = FPXQ_SQR(T6, t, p);
    GEN T9 = FPXQ_MUL(T8, T7, t, p);
    
    return T9;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_15_1500(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;

    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);

    GEN imx0  = FPXQ_INV(mx0,  t, p);
    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx2  = FPXQ_INV(mx2,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);
    GEN imx11 = FPXQ_INV(mx11, t, p);

    GEN mx4p1   = FPM_FPXQ_MUL(frob_pows[1], mx4,   t, p);
    GEN mx11p1  = FPM_FPXQ_MUL(frob_pows[1], mx11,  t, p);
    GEN mx6p1   = FPM_FPXQ_MUL(frob_pows[1], mx6,   t, p);
    GEN mx9p1   = FPM_FPXQ_MUL(frob_pows[1], mx9,   t, p);
    GEN mx1p1   = FPM_FPXQ_MUL(frob_pows[1], mx1,   t, p);
    GEN imx7p1  = FPM_FPXQ_MUL(frob_pows[1], imx7,  t, p);
    GEN imx5p1  = FPM_FPXQ_MUL(frob_pows[1], imx5,  t, p);
    GEN imx10p1 = FPM_FPXQ_MUL(frob_pows[1], imx10, t, p);
    GEN imx2p1  = FPM_FPXQ_MUL(frob_pows[1], imx2,  t, p);
    GEN imx0p1  = FPM_FPXQ_MUL(frob_pows[1], imx0,  t, p);

    GEN mx8p2   = FPM_FPXQ_MUL(frob_pows[2], mx8,   t, p);
    GEN mx5p2   = FPM_FPXQ_MUL(frob_pows[2], mx5,   t, p);
    GEN mx3p2   = FPM_FPXQ_MUL(frob_pows[2], mx3,   t, p);
    GEN mx10p2  = FPM_FPXQ_MUL(frob_pows[2], mx10,  t, p);
    GEN imx4p2  = FPM_FPXQ_MUL(frob_pows[2], imx4,  t, p);
    GEN imx11p2 = FPM_FPXQ_MUL(frob_pows[2], imx11, t, p);
    GEN imx9p2  = FPM_FPXQ_MUL(frob_pows[2], imx9,  t, p);
    GEN imx1p2  = FPM_FPXQ_MUL(frob_pows[2], imx1,  t, p);
    GEN imx0p2  = FPM_FPXQ_MUL(frob_pows[2], imx0,  t, p);

    GEN mx7p3   = FPM_FPXQ_MUL(frob_pows[3], mx7,   t, p);
    GEN mx5p3   = FPM_FPXQ_MUL(frob_pows[3], mx5,   t, p);
    GEN mx3p3   = FPM_FPXQ_MUL(frob_pows[3], mx3,   t, p);
    GEN mx11p3  = FPM_FPXQ_MUL(frob_pows[3], mx11,  t, p);
    GEN mx9p3   = FPM_FPXQ_MUL(frob_pows[3], mx9,   t, p);
    GEN mx0p3   = FPM_FPXQ_MUL(frob_pows[3], mx0,   t, p);
    GEN imx8p3  = FPM_FPXQ_MUL(frob_pows[3], imx8,  t, p);
    GEN imx4p3  = FPM_FPXQ_MUL(frob_pows[3], imx4,  t, p);
    GEN imx10p3 = FPM_FPXQ_MUL(frob_pows[3], imx10, t, p);
    GEN imx6p3  = FPM_FPXQ_MUL(frob_pows[3], imx6,  t, p);
    GEN imx1p3  = FPM_FPXQ_MUL(frob_pows[3], imx1,  t, p);

    GEN mx4p4   = FPM_FPXQ_MUL(frob_pows[4], mx4,   t, p);
    GEN mx11p4  = FPM_FPXQ_MUL(frob_pows[4], mx11,  t, p);
    GEN mx6p4   = FPM_FPXQ_MUL(frob_pows[4], mx6,   t, p);
    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,   t, p);
    GEN mx9p4   = FPM_FPXQ_MUL(frob_pows[4], mx9,   t, p);
    GEN mx1p4   = FPM_FPXQ_MUL(frob_pows[4], mx1,   t, p);
    GEN imx8p4  = FPM_FPXQ_MUL(frob_pows[4], imx8,  t, p);
    GEN imx5p4  = FPM_FPXQ_MUL(frob_pows[4], imx5,  t, p);
    GEN imx3p4  = FPM_FPXQ_MUL(frob_pows[4], imx3,  t, p);
    GEN imx10p4 = FPM_FPXQ_MUL(frob_pows[4], imx10, t, p);
    GEN imx0p4  = FPM_FPXQ_MUL(frob_pows[4], imx0,  t, p);

    GEN mx8p5   = FPM_FPXQ_MUL(frob_pows[5], mx8,   t, p);
    GEN mx5p5   = FPM_FPXQ_MUL(frob_pows[5], mx5,   t, p);
    GEN mx3p5   = FPM_FPXQ_MUL(frob_pows[5], mx3,   t, p);
    GEN mx10p5  = FPM_FPXQ_MUL(frob_pows[5], mx10,  t, p);
    GEN mx6p5   = FPM_FPXQ_MUL(frob_pows[5], mx6,   t, p);
    GEN imx7p5  = FPM_FPXQ_MUL(frob_pows[5], imx7,  t, p);
    GEN imx4p5  = FPM_FPXQ_MUL(frob_pows[5], imx4,  t, p);
    GEN imx11p5 = FPM_FPXQ_MUL(frob_pows[5], imx11, t, p);
    GEN imx2p5  = FPM_FPXQ_MUL(frob_pows[5], imx2,  t, p);
    GEN imx9p5  = FPM_FPXQ_MUL(frob_pows[5], imx9,  t, p);
    GEN imx1p5  = FPM_FPXQ_MUL(frob_pows[5], imx1,  t, p);

    GEN mx9p6   = FPM_FPXQ_MUL(frob_pows[6], mx9,   t, p);
    GEN mx7p6   = FPM_FPXQ_MUL(frob_pows[6], mx7,   t, p);
    GEN mx3p6   = FPM_FPXQ_MUL(frob_pows[6], mx3,   t, p);
    GEN mx1p6   = FPM_FPXQ_MUL(frob_pows[6], mx1,   t, p);
    GEN mx5p6   = FPM_FPXQ_MUL(frob_pows[6], mx5,   t, p);
    GEN imx8p6  = FPM_FPXQ_MUL(frob_pows[6], imx8,  t, p);
    GEN imx6p6  = FPM_FPXQ_MUL(frob_pows[6], imx6,  t, p);
    GEN imx2p6  = FPM_FPXQ_MUL(frob_pows[6], imx2,  t, p);
    GEN imx10p6 = FPM_FPXQ_MUL(frob_pows[6], imx10, t, p);
    GEN imx0p6  = FPM_FPXQ_MUL(frob_pows[6], imx0,  t, p);

    GEN mx4p7  = FPM_FPXQ_MUL(frob_pows[7], mx4,  t, p);
    GEN mx2p7  = FPM_FPXQ_MUL(frob_pows[7], mx2,  t, p);
    GEN mx0p7  = FPM_FPXQ_MUL(frob_pows[7], mx0,  t, p);
    GEN imx3p7 = FPM_FPXQ_MUL(frob_pows[7], imx3, t, p);
    GEN imx1p7 = FPM_FPXQ_MUL(frob_pows[7], imx1, t, p);

    GEN y0;
    y0 = FPXQ_MUL(imx8, imx4,  t, p);
    y0 = FPXQ_MUL(y0, mx11p1,  t, p);
    y0 = FPXQ_MUL(y0, mx3p2,   t, p);
    y0 = FPXQ_MUL(y0, mx11p3,  t, p);
    y0 = FPXQ_MUL(y0, mx11p4,  t, p);
    y0 = FPXQ_MUL(y0, mx1p4,   t, p);
    y0 = FPXQ_MUL(y0, mx6p5,   t, p);
    y0 = FPXQ_MUL(y0, imx11p5, t, p);
    y0 = FPXQ_MUL(y0, imx2p5,  t, p);
    y0 = FPXQ_MUL(y0, imx1p5,  t, p);
    y0 = FPXQ_MUL(y0, mx5p6,   t, p);
    y0 = FPXQ_MUL(y0, imx6p6,  t, p);
    y0 = FPXQ_MUL(y0, imx10p6, t, p);
    y0 = FPXQ_MUL(y0, imx0p6,  t, p);
    y0 = FPXQ_MUL(y0, mx4p7,   t, p);
    y0 = FPXQ_MUL(y0, mx0p7,   t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx7, mx5,    t, p);
    y1 = FPXQ_MUL(y1, mx9p1,   t, p);
    y1 = FPXQ_MUL(y1, imx10p1, t, p);
    y1 = FPXQ_MUL(y1, imx2p1,  t, p);
    y1 = FPXQ_MUL(y1, imx0p1,  t, p);
    y1 = FPXQ_MUL(y1, imx11p2, t, p);
    y1 = FPXQ_MUL(y1, imx1p2,  t, p);
    y1 = FPXQ_MUL(y1, mx5p3,   t, p);
    y1 = FPXQ_MUL(y1, mx3p3,   t, p);
    y1 = FPXQ_MUL(y1, imx10p3, t, p);
    y1 = FPXQ_MUL(y1, imx1p3,  t, p);
    y1 = FPXQ_MUL(y1, mx6p4,   t, p);
    y1 = FPXQ_MUL(y1, imx8p4,  t, p);
    y1 = FPXQ_MUL(y1, imx10p4, t, p);
    y1 = FPXQ_MUL(y1, imx0p4,  t, p);
    y1 = FPXQ_MUL(y1, mx5p5,   t, p);
    y1 = FPXQ_MUL(y1, mx9p6,   t, p);
    y1 = FPXQ_MUL(y1, mx7p6,   t, p);
    y1 = FPXQ_MUL(y1, mx3p6,   t, p);
    y1 = FPXQ_MUL(y1, imx2p6,  t, p);
    y1 = FPXQ_MUL(y1, imx3p7,  t, p);
    y1 = FPXQ_MUL(y1, imx1p7,  t, p);

    GEN y2;
    y2 = FPXQ_MUL(mx2, mx0,   t, p);
    y2 = FPXQ_MUL(y2, imx6,   t, p);
    y2 = FPXQ_MUL(y2, imx1,   t, p);
    y2 = FPXQ_MUL(y2, imx7p1, t, p);
    y2 = FPXQ_MUL(y2, mx8p2,  t, p);
    y2 = FPXQ_MUL(y2, imx4p2, t, p);
    y2 = FPXQ_MUL(y2, mx7p3,  t, p);
    y2 = FPXQ_MUL(y2, mx9p3,  t, p);
    y2 = FPXQ_MUL(y2, imx8p3, t, p);
    y2 = FPXQ_MUL(y2, imx4p3, t, p);
    y2 = FPXQ_MUL(y2, mx2p4,  t, p);
    y2 = FPXQ_MUL(y2, mx9p4,  t, p);
    y2 = FPXQ_MUL(y2, mx3p5,  t, p);
    y2 = FPXQ_MUL(y2, mx10p5, t, p);
    y2 = FPXQ_MUL(y2, imx7p5, t, p);
    y2 = FPXQ_MUL(y2, mx1p6,  t, p);
    y2 = FPXQ_MUL(y2, imx8p6, t, p);
    y2 = FPXQ_MUL(y2, mx2p7,  t, p);

    GEN y3;
    y3 = FPXQ_MUL(mx4p1, mx6p1, t, p);
    y3 = FPXQ_MUL(y3, mx10p2,   t, p);
    y3 = FPXQ_MUL(y3, imx0p2,   t, p);
    y3 = FPXQ_MUL(y3, imx6p3,   t, p);
    y3 = FPXQ_MUL(y3, imx4p5,   t, p);

    GEN y4;
    y4 = FPXQ_MUL(mx1p1, mx5p2, t, p);
    y4 = FPXQ_MUL(y4, imx9p2,   t, p);
    y4 = FPXQ_MUL(y4, mx8p5,    t, p);
    y4 = FPXQ_MUL(y4, imx9p5,   t, p);

    GEN y5;
    y5 = FPXQ_MUL(imx5p1, mx0p3, t, p);
    y5 = FPXQ_MUL(y5, mx4p4,     t, p);
    y5 = FPXQ_MUL(y5, imx5p4,    t, p);
    y5 = FPXQ_MUL(y5, imx3p4,    t, p);
    
    GEN T0 = FPXQ_MUL(y5, y4, t, p);
    GEN T1 = FPXQ_MUL(y5, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y3, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_MUL(T2, y2, t, p);
    GEN T5 = FPXQ_MUL(T3, T2, t, p);
    GEN T6 = FPXQ_MUL(T4, y1, t, p);
    GEN T7 = FPXQ_MUL(T5, T4, t, p);
    GEN T8 = FPXQ_SQR(T6, t, p);
    GEN T9 = FPXQ_MUL(T8, T7, t, p);
    
    return T9;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_16_1250(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0,   x, t, p);
    GEN mx2 = FPXQ_POW(mx1,   x, t, p);
    GEN mx3 = FPXQ_POW(mx2,   x, t, p);
    GEN mx4 = FPXQ_POW(mx3,   x, t, p);
    GEN mx5 = FPXQ_POW(mx4,   x, t, p);
    GEN mx6 = FPXQ_POW(mx5,   x, t, p);
    GEN mx7 = FPXQ_POW(mx6,   x, t, p);
    GEN mx8 = FPXQ_POW(mx7,   x, t, p);
    GEN mx9 = FPXQ_POW(mx8,   x, t, p);

    GEN imx0 = FPXQ_INV(mx0,  t, p);
    GEN imx1 = FPXQ_INV(mx1,  t, p);
    GEN imx2 = FPXQ_INV(mx2,  t, p);
    GEN imx3 = FPXQ_INV(mx3,  t, p);
    GEN imx4 = FPXQ_INV(mx4,  t, p);
    GEN imx5 = FPXQ_INV(mx5,  t, p);
    GEN imx6 = FPXQ_INV(mx6,  t, p);
    GEN imx7 = FPXQ_INV(mx7,  t, p);
    GEN imx8 = FPXQ_INV(mx8,  t, p);
    GEN imx9 = FPXQ_INV(mx9,  t, p);

    GEN mx8p1 = FPM_FPXQ_MUL(frob_pows[1], mx8,  t, p);
    GEN mx4p1 = FPM_FPXQ_MUL(frob_pows[1], mx4,  t, p);
    GEN mx6p1 = FPM_FPXQ_MUL(frob_pows[1], mx6,  t, p);
    GEN mx2p1 = FPM_FPXQ_MUL(frob_pows[1], mx2,  t, p);
    GEN mx3p1 = FPM_FPXQ_MUL(frob_pows[1], mx3,  t, p);
    GEN mx7p1 = FPM_FPXQ_MUL(frob_pows[1], mx7,  t, p);

    GEN mx6p2 = FPM_FPXQ_MUL(frob_pows[2], mx6,  t, p);
    GEN mx5p2 = FPM_FPXQ_MUL(frob_pows[2], mx5,  t, p);
    GEN mx2p2 = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx3p2 = FPM_FPXQ_MUL(frob_pows[2], mx3,  t, p);
    GEN mx7p2 = FPM_FPXQ_MUL(frob_pows[2], mx7,  t, p);
    GEN mx1p2 = FPM_FPXQ_MUL(frob_pows[2], mx1,  t, p);

    GEN imx4p3 = FPM_FPXQ_MUL(frob_pows[3], imx4,  t, p);
    GEN imx6p3 = FPM_FPXQ_MUL(frob_pows[3], imx6,  t, p);
    GEN imx5p3 = FPM_FPXQ_MUL(frob_pows[3], imx5,  t, p);
    GEN imx2p3 = FPM_FPXQ_MUL(frob_pows[3], imx2,  t, p);
    GEN imx1p3 = FPM_FPXQ_MUL(frob_pows[3], imx1,  t, p);
    GEN imx0p3 = FPM_FPXQ_MUL(frob_pows[3], imx0,  t, p);

    GEN mx5p4 = FPM_FPXQ_MUL(frob_pows[4], mx5,  t, p);
    GEN mx3p4 = FPM_FPXQ_MUL(frob_pows[4], mx3,  t, p);
    GEN mx0p4 = FPM_FPXQ_MUL(frob_pows[4], mx0,  t, p);
    GEN imx8p4 = FPM_FPXQ_MUL(frob_pows[4], imx8,  t, p);
    GEN imx4p4 = FPM_FPXQ_MUL(frob_pows[4], imx4,  t, p);
    GEN imx9p4 = FPM_FPXQ_MUL(frob_pows[4], imx9,  t, p);
    GEN imx7p4 = FPM_FPXQ_MUL(frob_pows[4], imx7,  t, p);

    GEN imx8p5 = FPM_FPXQ_MUL(frob_pows[5], imx8,  t, p);
    GEN imx4p5 = FPM_FPXQ_MUL(frob_pows[5], imx4,  t, p);
    GEN imx6p5 = FPM_FPXQ_MUL(frob_pows[5], imx6,  t, p);
    GEN imx2p5 = FPM_FPXQ_MUL(frob_pows[5], imx2,  t, p);
    GEN imx3p5 = FPM_FPXQ_MUL(frob_pows[5], imx3,  t, p);
    GEN imx7p5 = FPM_FPXQ_MUL(frob_pows[5], imx7,  t, p);

    GEN mx7p6 = FPM_FPXQ_MUL(frob_pows[6], mx7,  t, p);
    GEN mx6p6 = FPM_FPXQ_MUL(frob_pows[6], mx6,  t, p);
    GEN mx5p6 = FPM_FPXQ_MUL(frob_pows[6], mx5,  t, p);
    GEN mx3p6 = FPM_FPXQ_MUL(frob_pows[6], mx3,  t, p);
    GEN mx2p6 = FPM_FPXQ_MUL(frob_pows[6], mx2,  t, p);
    GEN mx1p6 = FPM_FPXQ_MUL(frob_pows[6], mx1,  t, p);

    GEN mx2p7 = FPM_FPXQ_MUL(frob_pows[7], mx2,  t, p);
    GEN mx1p7 = FPM_FPXQ_MUL(frob_pows[7], mx1,  t, p);
    GEN mx0p7 = FPM_FPXQ_MUL(frob_pows[7], mx0,  t, p);

    GEN y0 = imx9p4;
    GEN y1 = imx8p4;
    GEN y2 = imx7p4;
    GEN y3 = imx9;
    GEN y4 = mx8p1;
    GEN y5 = imx8p5;
    GEN y6 = imx8;
    GEN y7 = mx7p2;
    GEN y8 = mx5p4;
    GEN y9 = mx7p1;
    GEN y10 = imx7p5;
    GEN y11 = FPXQ_MUL(mx6p2, mx7p6, t, p);
    GEN y12 = imx4p4;
    GEN y13 = imx7;
    GEN y14 = mx6p1;
    GEN y15 = FPXQ_MUL(imx6p5, mx6p6, t, p);
    GEN y16 = FPXQ_MUL(mx5p2, imx6p3, t, p);
    GEN y17 = mx3p4;
    GEN y18 = mx4p1;
    GEN y19 = FPXQ_MUL(imx5p3, mx5p6, t, p);
    GEN y20 = imx5;
    GEN y21 = imx4p5;
    GEN y22 = imx4p3;
    GEN y23 = mx2p7;
    GEN y24 = mx3p2;
    GEN y25 = mx3p6;
    GEN y26 = mx2p1;
    GEN y27 = imx4;
    GEN y28 = mx3p1;
    GEN y29 = mx0;
    GEN y30 = imx3;
    GEN y31 = mx1p7;
    GEN y32 = imx3p5;
    GEN y33 = imx2p5;
    GEN y34 = imx2p3;
    GEN y35 = mx2p2;
    GEN y36 = mx0p7;
    GEN y37 = mx0p4;
    GEN y38 = mx1p2;
    GEN y39 = mx2p6;
    GEN y40 = mx1p6;
    GEN y41 = imx1p3;
    GEN y42 = imx0p3;
    
    GEN T0 = FPXQ_MUL(y42, y31, t, p);
    GEN T1 = FPXQ_MUL(y42, y41, t, p);
    GEN T2 = FPXQ_MUL(y42, y19, t, p);
    GEN T3 = FPXQ_MUL(T1, y35, t, p);
    GEN T4 = FPXQ_SQR(T1, t, p);
    GEN T5 = FPXQ_MUL(T4, y39, t, p);
    GEN T6 = FPXQ_MUL(y40, y23, t, p);
    GEN T7 = FPXQ_MUL(T5, y38, t, p);
    GEN T8 = FPXQ_MUL(T5, y15, t, p);
    GEN T9 = FPXQ_MUL(T7, y34, t, p);
    GEN T10 = FPXQ_MUL(T7, T0, t, p);
    GEN T11 = FPXQ_MUL(y37, y33, t, p);
    GEN T12 = FPXQ_MUL(y40, y36, t, p);
    GEN T13 = FPXQ_MUL(y36, y16, t, p);
    GEN T14 = FPXQ_MUL(T12, T3, t, p);
    GEN T15 = FPXQ_MUL(T12, y24, t, p);
    GEN T16 = FPXQ_MUL(T14, T9, t, p);
    GEN T17 = FPXQ_MUL(T16, T11, t, p);
    GEN T18 = FPXQ_MUL(T16, y14, t, p);
    GEN T19 = FPXQ_MUL(T17, y17, t, p);
    GEN T20 = FPXQ_SQR(T17, t, p);
    GEN T21 = FPXQ_MUL(T20, y30, t, p);
    GEN T22 = FPXQ_MUL(T21, y32, t, p);
    GEN T23 = FPXQ_MUL(T10, y37, t, p);
    GEN T24 = FPXQ_MUL(y37, y8, t, p);
    GEN T25 = FPXQ_SQR(T23, t, p);
    GEN T26 = FPXQ_MUL(T25, T6, t, p);
    GEN T27 = FPXQ_MUL(T22, T15, t, p);
    GEN T28 = FPXQ_MUL(y29, y26, t, p);
    GEN T29 = FPXQ_MUL(y28, y12, t, p);
    GEN T30 = FPXQ_MUL(y28, y27, t, p);
    GEN T31 = FPXQ_MUL(T29, y28, t, p);
    GEN T32 = FPXQ_MUL(T30, y32, t, p);
    GEN T33 = FPXQ_MUL(T30, y29, t, p);
    GEN T34 = FPXQ_MUL(T28, y25, t, p);
    GEN T35 = FPXQ_MUL(T28, T18, t, p);
    GEN T36 = FPXQ_MUL(T34, T27, t, p);
    GEN T37 = FPXQ_MUL(T35, T34, t, p);
    GEN T38 = FPXQ_MUL(T36, T32, t, p);
    GEN T39 = FPXQ_MUL(T36, y11, t, p);
    GEN T40 = FPXQ_MUL(T38, T26, t, p);
    GEN T41 = FPXQ_MUL(T38, y7, t, p);
    GEN T42 = FPXQ_MUL(T40, y22, t, p);
    GEN T43 = FPXQ_MUL(T40, T2, t, p);
    GEN T44 = FPXQ_MUL(T42, y21, t, p);
    GEN T45 = FPXQ_MUL(T42, y10, t, p);
    GEN T46 = FPXQ_MUL(T44, T22, t, p);
    GEN T47 = FPXQ_MUL(T44, T19, t, p);
    GEN T48 = FPXQ_SQR(T46, t, p);
    GEN T49 = FPXQ_MUL(T48, y18, t, p);
    GEN T50 = FPXQ_MUL(T14, y20, t, p);
    GEN T51 = FPXQ_MUL(T14, y6, t, p);
    GEN T52 = FPXQ_MUL(T50, T33, t, p);
    GEN T53 = FPXQ_MUL(T52, T43, t, p);
    GEN T54 = FPXQ_MUL(T52, T51, t, p);
    GEN T55 = FPXQ_MUL(T53, T49, t, p);
    GEN T56 = FPXQ_MUL(T53, y9, t, p);
    GEN T57 = FPXQ_MUL(T55, T47, t, p);
    GEN T58 = FPXQ_MUL(T55, T37, t, p);
    GEN T59 = FPXQ_MUL(T57, T13, t, p);
    GEN T60 = FPXQ_MUL(T57, y5, t, p);
    GEN T61 = FPXQ_MUL(T59, T8, t, p);
    GEN T62 = FPXQ_MUL(T59, T41, t, p);
    GEN T63 = FPXQ_MUL(T61, T58, t, p);
    GEN T64 = FPXQ_MUL(T62, T61, t, p);
    GEN T65 = FPXQ_MUL(T63, y13, t, p);
    GEN T66 = FPXQ_MUL(T63, T60, t, p);
    GEN T67 = FPXQ_MUL(T65, T31, t, p);
    GEN T68 = FPXQ_MUL(T67, T39, t, p);
    GEN T69 = FPXQ_MUL(T67, y1, t, p);
    GEN T70 = FPXQ_MUL(T68, T45, t, p);
    GEN T71 = FPXQ_MUL(T68, y2, t, p);
    GEN T72 = FPXQ_MUL(T70, T56, t, p);
    GEN T73 = FPXQ_MUL(T71, T70, t, p);
    GEN T74 = FPXQ_MUL(T72, T24, t, p);
    GEN T75 = FPXQ_MUL(T72, T65, t, p);
    GEN T76 = FPXQ_MUL(T74, T64, t, p);
    GEN T77 = FPXQ_MUL(T74, T69, t, p);
    GEN T78 = FPXQ_MUL(T76, y4, t, p);
    GEN T79 = FPXQ_MUL(T76, T73, t, p);
    GEN T80 = FPXQ_MUL(T66, T54, t, p);
    GEN T81 = FPXQ_MUL(T54, y0, t, p);
    GEN T82 = FPXQ_SQR(T80, t, p);
    GEN T83 = FPXQ_MUL(T82, T79, t, p);
    GEN T84 = FPXQ_MUL(T78, y3, t, p);
    GEN T85 = FPXQ_MUL(T78, T77, t, p);
    GEN T86 = FPXQ_MUL(T84, T83, t, p);
    GEN T87 = FPXQ_MUL(T84, T75, t, p);
    GEN T88 = FPXQ_MUL(T86, T50, t, p);
    GEN T89 = FPXQ_MUL(T86, T85, t, p);
    GEN T90 = FPXQ_MUL(T89, T88, t, p);
    GEN T91 = FPXQ_MUL(T88, T81, t, p);
    GEN T92 = FPXQ_SQR(T90, t, p);
    GEN T93 = FPXQ_MUL(T92, T91, t, p);
    GEN T94 = FPXQ_SQR(T93, t, p);
    GEN T95 = FPXQ_MUL(T94, T87, t, p);

    return T95;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_17_1125(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,   x, t, p);
    GEN mx2  = FPXQ_POW(mx1,   x, t, p);
    GEN mx3  = FPXQ_POW(mx2,   x, t, p);
    GEN mx4  = FPXQ_POW(mx3,   x, t, p);
    GEN mx5  = FPXQ_POW(mx4,   x, t, p);
    GEN mx6  = FPXQ_POW(mx5,   x, t, p);
    GEN mx7  = FPXQ_POW(mx6,   x, t, p);
    GEN mx8  = FPXQ_POW(mx7,   x, t, p);
    GEN mx9  = FPXQ_POW(mx8,   x, t, p);
    GEN mx10 = FPXQ_POW(mx9,   x, t, p);
    GEN mx11 = FPXQ_POW(mx10,   x, t, p);
    GEN mx12 = FPXQ_POW(mx11,   x, t, p);
    GEN mx13 = FPXQ_POW(mx12,   x, t, p);
    GEN mx14 = FPXQ_POW(mx13,   x, t, p);
    GEN mx15 = FPXQ_POW(mx14,   x, t, p);
    GEN mx16 = FPXQ_POW(mx15,   x, t, p);
    GEN mx17 = FPXQ_POW(mx16,   x, t, p);
    GEN mx18 = FPXQ_POW(mx17,   x, t, p);
    GEN mx19 = FPXQ_POW(mx18,   x, t, p);
    GEN mx20 = FPXQ_POW(mx19,   x, t, p);
    GEN mx21 = FPXQ_POW(mx20,   x, t, p);
    GEN mx22 = FPXQ_POW(mx21,   x, t, p);
    GEN mx23 = FPXQ_POW(mx22,   x, t, p);
    GEN mx24 = FPXQ_POW(mx23,   x, t, p);
    GEN mx25 = FPXQ_POW(mx24,   x, t, p);
    GEN mx26 = FPXQ_POW(mx25,   x, t, p);
    GEN mx27 = FPXQ_POW(mx26,   x, t, p);
    GEN mx28 = FPXQ_POW(mx27,   x, t, p);
    GEN mx29 = FPXQ_POW(mx28,   x, t, p);
    GEN mx30 = FPXQ_POW(mx29,   x, t, p);
    GEN mx31 = FPXQ_POW(mx30,   x, t, p);
    GEN mx32 = FPXQ_POW(mx31,   x, t, p);
    GEN mx33 = FPXQ_POW(mx32,   x, t, p);
    GEN mx34 = FPXQ_POW(mx33,   x, t, p);
    GEN mx35 = FPXQ_POW(mx34,   x, t, p);

    
    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10,  t, p);
    GEN imx11 = FPXQ_INV(mx11,  t, p);
    GEN imx12 = FPXQ_INV(mx12,  t, p);
    GEN imx13 = FPXQ_INV(mx13,  t, p);
    GEN imx14 = FPXQ_INV(mx14,  t, p);
    GEN imx15 = FPXQ_INV(mx15,  t, p);
    GEN imx16 = FPXQ_INV(mx16,  t, p);
    GEN imx19 = FPXQ_INV(mx19,  t, p);
    GEN imx21 = FPXQ_INV(mx21,  t, p);
    GEN imx22 = FPXQ_INV(mx22,  t, p);
    GEN imx23 = FPXQ_INV(mx23,  t, p);
    GEN imx24 = FPXQ_INV(mx24,  t, p);
    GEN imx25 = FPXQ_INV(mx25,  t, p);
    GEN imx26 = FPXQ_INV(mx26,  t, p);
    GEN imx27 = FPXQ_INV(mx27,  t, p);
    GEN imx28 = FPXQ_INV(mx28,  t, p);
    GEN imx29 = FPXQ_INV(mx29,  t, p);
    GEN imx30 = FPXQ_INV(mx30,  t, p);
    GEN imx31 = FPXQ_INV(mx31,  t, p);
    GEN imx32 = FPXQ_INV(mx32,  t, p);
    GEN imx33 = FPXQ_INV(mx33,  t, p);
    GEN imx34 = FPXQ_INV(mx34,  t, p);

    GEN mx20p1  = FPM_FPXQ_MUL(frob_pows[1], mx20,  t, p);
    GEN mx17p1  = FPM_FPXQ_MUL(frob_pows[1], mx17,  t, p);
    GEN mx15p1  = FPM_FPXQ_MUL(frob_pows[1], mx15,  t, p);
    GEN mx35p1  = FPM_FPXQ_MUL(frob_pows[1], mx35,  t, p);
    GEN mx2p1   = FPM_FPXQ_MUL(frob_pows[1], mx2,  t, p);
    GEN mx33p1  = FPM_FPXQ_MUL(frob_pows[1], mx33,  t, p);
    GEN mx0p1   = FPM_FPXQ_MUL(frob_pows[1], mx0,  t, p);
    GEN imx19p1 = FPM_FPXQ_MUL(frob_pows[1], imx19,  t, p);
    GEN imx16p1 = FPM_FPXQ_MUL(frob_pows[1], imx16,  t, p);
    GEN imx3p1  = FPM_FPXQ_MUL(frob_pows[1], imx3,  t, p);
    GEN imx34p1 = FPM_FPXQ_MUL(frob_pows[1], imx34,  t, p);
    GEN imx1p1  = FPM_FPXQ_MUL(frob_pows[1], imx1,  t, p);

    GEN mx32p2  = FPM_FPXQ_MUL(frob_pows[2], mx32,  t, p);
    GEN mx20p2  = FPM_FPXQ_MUL(frob_pows[2], mx20,  t, p);
    GEN mx16p2  = FPM_FPXQ_MUL(frob_pows[2], mx16,  t, p);
    GEN mx34p2  = FPM_FPXQ_MUL(frob_pows[2], mx34,  t, p);
    GEN mx18p2  = FPM_FPXQ_MUL(frob_pows[2], mx18,  t, p);
    GEN mx2p2   = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx14p2  = FPM_FPXQ_MUL(frob_pows[2], mx14,  t, p);
    GEN mx0p2   = FPM_FPXQ_MUL(frob_pows[2], mx0,  t, p);
    GEN imx31p2 = FPM_FPXQ_MUL(frob_pows[2], imx31,  t, p);
    GEN imx19p2 = FPM_FPXQ_MUL(frob_pows[2], imx19,  t, p);
    GEN imx15p2 = FPM_FPXQ_MUL(frob_pows[2], imx15,  t, p);
    GEN imx3p2  = FPM_FPXQ_MUL(frob_pows[2], imx3,  t, p);
    GEN imx33p2 = FPM_FPXQ_MUL(frob_pows[2], imx33,  t, p);
    GEN imx1p2  = FPM_FPXQ_MUL(frob_pows[2], imx1,  t, p);

    GEN mx31p3  = FPM_FPXQ_MUL(frob_pows[3], mx31,  t, p);
    GEN mx20p3  = FPM_FPXQ_MUL(frob_pows[3], mx20,  t, p);
    GEN mx16p3  = FPM_FPXQ_MUL(frob_pows[3], mx16,  t, p);
    GEN mx18p3  = FPM_FPXQ_MUL(frob_pows[3], mx18,  t, p);
    GEN mx2p3   = FPM_FPXQ_MUL(frob_pows[3], mx2,  t, p);
    GEN mx0p3   = FPM_FPXQ_MUL(frob_pows[3], mx0,  t, p);
    GEN imx32p3 = FPM_FPXQ_MUL(frob_pows[3], imx32,  t, p);
    GEN imx19p3 = FPM_FPXQ_MUL(frob_pows[3], imx19,  t, p);
    GEN imx15p3 = FPM_FPXQ_MUL(frob_pows[3], imx15,  t, p);
    GEN imx3p3  = FPM_FPXQ_MUL(frob_pows[3], imx3,  t, p);
    GEN imx14p3 = FPM_FPXQ_MUL(frob_pows[3], imx14,  t, p);
    GEN imx30p3 = FPM_FPXQ_MUL(frob_pows[3], imx30,  t, p);
    GEN imx1p3  = FPM_FPXQ_MUL(frob_pows[3], imx1,  t, p);

    GEN mx31p4  = FPM_FPXQ_MUL(frob_pows[4], mx31,  t, p);
    GEN mx20p4  = FPM_FPXQ_MUL(frob_pows[4], mx20,  t, p);
    GEN mx15p4  = FPM_FPXQ_MUL(frob_pows[4], mx15,  t, p);
    GEN mx13p4  = FPM_FPXQ_MUL(frob_pows[4], mx13,  t, p);
    GEN mx18p4  = FPM_FPXQ_MUL(frob_pows[4], mx18,  t, p);
    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,  t, p);
    GEN mx0p4   = FPM_FPXQ_MUL(frob_pows[4], mx0,  t, p);
    GEN imx32p4 = FPM_FPXQ_MUL(frob_pows[4], imx32,  t, p);
    GEN imx19p4 = FPM_FPXQ_MUL(frob_pows[4], imx19,  t, p);
    GEN imx12p4 = FPM_FPXQ_MUL(frob_pows[4], imx12,  t, p);
    GEN imx3p4  = FPM_FPXQ_MUL(frob_pows[4], imx3,  t, p);
    GEN imx14p4 = FPM_FPXQ_MUL(frob_pows[4], imx14,  t, p);
    GEN imx30p4 = FPM_FPXQ_MUL(frob_pows[4], imx30,  t, p);
    GEN imx1p4  = FPM_FPXQ_MUL(frob_pows[4], imx1,  t, p);

    GEN mx28p5  = FPM_FPXQ_MUL(frob_pows[5], mx28,  t, p);
    GEN mx20p5  = FPM_FPXQ_MUL(frob_pows[5], mx20,  t, p);
    GEN mx12p5  = FPM_FPXQ_MUL(frob_pows[5], mx12,  t, p);
    GEN mx18p5  = FPM_FPXQ_MUL(frob_pows[5], mx18,  t, p);
    GEN mx2p5   = FPM_FPXQ_MUL(frob_pows[5], mx2,  t, p);
    GEN mx30p5  = FPM_FPXQ_MUL(frob_pows[5], mx30,  t, p);
    GEN mx0p5   = FPM_FPXQ_MUL(frob_pows[5], mx0,  t, p);
    GEN imx31p5 = FPM_FPXQ_MUL(frob_pows[5], imx31,  t, p);
    GEN imx29p5 = FPM_FPXQ_MUL(frob_pows[5], imx29,  t, p);
    GEN imx19p5 = FPM_FPXQ_MUL(frob_pows[5], imx19,  t, p);
    GEN imx13p5 = FPM_FPXQ_MUL(frob_pows[5], imx13,  t, p);
    GEN imx3p5  = FPM_FPXQ_MUL(frob_pows[5], imx3,  t, p);
    GEN imx11p5 = FPM_FPXQ_MUL(frob_pows[5], imx11,  t, p);
    GEN imx1p5  = FPM_FPXQ_MUL(frob_pows[5], imx1,  t, p);

    GEN mx29p6  = FPM_FPXQ_MUL(frob_pows[6], mx29,  t, p);
    GEN mx20p6  = FPM_FPXQ_MUL(frob_pows[6], mx20,  t, p);
    GEN mx12p6  = FPM_FPXQ_MUL(frob_pows[6], mx12,  t, p);
    GEN mx18p6  = FPM_FPXQ_MUL(frob_pows[6], mx18,  t, p);
    GEN mx2p6   = FPM_FPXQ_MUL(frob_pows[6], mx2,  t, p);
    GEN mx11p6  = FPM_FPXQ_MUL(frob_pows[6], mx11,  t, p);
    GEN mx27p6  = FPM_FPXQ_MUL(frob_pows[6], mx27,  t, p);
    GEN mx0p6   = FPM_FPXQ_MUL(frob_pows[6], mx0,  t, p);
    GEN imx28p6 = FPM_FPXQ_MUL(frob_pows[6], imx28,  t, p);
    GEN imx19p6 = FPM_FPXQ_MUL(frob_pows[6], imx19,  t, p);
    GEN imx13p6 = FPM_FPXQ_MUL(frob_pows[6], imx13,  t, p);
    GEN imx3p6  = FPM_FPXQ_MUL(frob_pows[6], imx3,  t, p);
    GEN imx1p6  = FPM_FPXQ_MUL(frob_pows[6], imx1,  t, p);

    GEN mx29p7  = FPM_FPXQ_MUL(frob_pows[7], mx29,  t, p);
    GEN mx20p7  = FPM_FPXQ_MUL(frob_pows[7], mx20,  t, p);
    GEN mx9p7   = FPM_FPXQ_MUL(frob_pows[7], mx9,  t, p);
    GEN mx18p7  = FPM_FPXQ_MUL(frob_pows[7], mx18,  t, p);
    GEN mx2p7   = FPM_FPXQ_MUL(frob_pows[7], mx2,  t, p);
    GEN mx11p7  = FPM_FPXQ_MUL(frob_pows[7], mx11,  t, p);
    GEN mx27p7  = FPM_FPXQ_MUL(frob_pows[7], mx27,  t, p);
    GEN mx0p7   = FPM_FPXQ_MUL(frob_pows[7], mx0,  t, p);
    GEN imx28p7 = FPM_FPXQ_MUL(frob_pows[7], imx28,  t, p);
    GEN imx19p7 = FPM_FPXQ_MUL(frob_pows[7], imx19,  t, p);
    GEN imx12p7 = FPM_FPXQ_MUL(frob_pows[7], imx12,  t, p);
    GEN imx10p7 = FPM_FPXQ_MUL(frob_pows[7], imx10,  t, p);
    GEN imx3p7  = FPM_FPXQ_MUL(frob_pows[7], imx3,  t, p);
    GEN imx1p7  = FPM_FPXQ_MUL(frob_pows[7], imx1,  t, p);

    GEN mx28p8  = FPM_FPXQ_MUL(frob_pows[8], mx28,  t, p);
    GEN mx26p8  = FPM_FPXQ_MUL(frob_pows[8], mx26,  t, p);
    GEN mx20p8  = FPM_FPXQ_MUL(frob_pows[8], mx20,  t, p);
    GEN mx10p8  = FPM_FPXQ_MUL(frob_pows[8], mx10,  t, p);
    GEN mx18p8  = FPM_FPXQ_MUL(frob_pows[8], mx18,  t, p);
    GEN mx2p8   = FPM_FPXQ_MUL(frob_pows[8], mx2,  t, p);
    GEN mx8p8   = FPM_FPXQ_MUL(frob_pows[8], mx8,  t, p);
    GEN mx0p8   = FPM_FPXQ_MUL(frob_pows[8], mx0,  t, p);
    GEN imx25p8 = FPM_FPXQ_MUL(frob_pows[8], imx25,  t, p);
    GEN imx19p8 = FPM_FPXQ_MUL(frob_pows[8], imx19,  t, p);
    GEN imx9p8  = FPM_FPXQ_MUL(frob_pows[8], imx9,  t, p);
    GEN imx3p8  = FPM_FPXQ_MUL(frob_pows[8], imx3,  t, p);
    GEN imx27p8 = FPM_FPXQ_MUL(frob_pows[8], imx27,  t, p);
    GEN imx1p8  = FPM_FPXQ_MUL(frob_pows[8], imx1,  t, p);

    GEN mx25p9  = FPM_FPXQ_MUL(frob_pows[9], mx25,  t, p);
    GEN mx20p9  = FPM_FPXQ_MUL(frob_pows[9], mx20,  t, p);
    GEN mx10p9  = FPM_FPXQ_MUL(frob_pows[9], mx10,  t, p);
    GEN mx18p9  = FPM_FPXQ_MUL(frob_pows[9], mx18,  t, p);
    GEN mx2p9   = FPM_FPXQ_MUL(frob_pows[9], mx2,  t, p);
    GEN mx0p9   = FPM_FPXQ_MUL(frob_pows[9], mx0,  t, p);
    GEN imx26p9 = FPM_FPXQ_MUL(frob_pows[9], imx26,  t, p);
    GEN imx19p9 = FPM_FPXQ_MUL(frob_pows[9], imx19,  t, p);
    GEN imx9p9  = FPM_FPXQ_MUL(frob_pows[9], imx9,  t, p);
    GEN imx3p9  = FPM_FPXQ_MUL(frob_pows[9], imx3,  t, p);
    GEN imx8p9  = FPM_FPXQ_MUL(frob_pows[9], imx8,  t, p);
    GEN imx24p9 = FPM_FPXQ_MUL(frob_pows[9], imx24,  t, p);
    GEN imx1p9  = FPM_FPXQ_MUL(frob_pows[9], imx1,  t, p);

    GEN mx25p10  = FPM_FPXQ_MUL(frob_pows[10], mx25,  t, p);
    GEN mx20p10  = FPM_FPXQ_MUL(frob_pows[10], mx20,  t, p);
    GEN mx9p10   = FPM_FPXQ_MUL(frob_pows[10], mx9,  t, p);
    GEN mx7p10   = FPM_FPXQ_MUL(frob_pows[10], mx7,  t, p);
    GEN mx18p10  = FPM_FPXQ_MUL(frob_pows[10], mx18,  t, p);
    GEN mx2p10   = FPM_FPXQ_MUL(frob_pows[10], mx2,  t, p);
    GEN mx0p10   = FPM_FPXQ_MUL(frob_pows[10], mx0,  t, p);
    GEN imx26p10 = FPM_FPXQ_MUL(frob_pows[10], imx26,  t, p);
    GEN imx19p10 = FPM_FPXQ_MUL(frob_pows[10], imx19,  t, p);
    GEN imx6p10  = FPM_FPXQ_MUL(frob_pows[10], imx6,  t, p);
    GEN imx3p10  = FPM_FPXQ_MUL(frob_pows[10], imx3,  t, p);
    GEN imx8p10  = FPM_FPXQ_MUL(frob_pows[10], imx8,  t, p);
    GEN imx24p10 = FPM_FPXQ_MUL(frob_pows[10], imx24,  t, p);
    GEN imx1p10  = FPM_FPXQ_MUL(frob_pows[10], imx1,  t, p);

    GEN mx22p11  = FPM_FPXQ_MUL(frob_pows[11], mx22,  t, p);
    GEN mx20p11  = FPM_FPXQ_MUL(frob_pows[11], mx20,  t, p);
    GEN mx6p11   = FPM_FPXQ_MUL(frob_pows[11], mx6,  t, p);
    GEN mx18p11  = FPM_FPXQ_MUL(frob_pows[11], mx18,  t, p);
    GEN mx2p11   = FPM_FPXQ_MUL(frob_pows[11], mx2,  t, p);
    GEN mx24p11  = FPM_FPXQ_MUL(frob_pows[11], mx24,  t, p);
    GEN mx0p11   = FPM_FPXQ_MUL(frob_pows[11], mx0,  t, p);
    GEN imx25p11 = FPM_FPXQ_MUL(frob_pows[11], imx25,  t, p);
    GEN imx23p11 = FPM_FPXQ_MUL(frob_pows[11], imx23,  t, p);
    GEN imx19p11 = FPM_FPXQ_MUL(frob_pows[11], imx19,  t, p);
    GEN imx7p11  = FPM_FPXQ_MUL(frob_pows[11], imx7,  t, p);
    GEN imx3p11  = FPM_FPXQ_MUL(frob_pows[11], imx3,  t, p);
    GEN imx5p11  = FPM_FPXQ_MUL(frob_pows[11], imx5,  t, p);
    GEN imx1p11  = FPM_FPXQ_MUL(frob_pows[11], imx1,  t, p);

    GEN mx23p12  = FPM_FPXQ_MUL(frob_pows[12], mx23,  t, p);
    GEN mx20p12  = FPM_FPXQ_MUL(frob_pows[12], mx20,  t, p);
    GEN mx6p12   = FPM_FPXQ_MUL(frob_pows[12], mx6,  t, p);
    GEN mx18p12  = FPM_FPXQ_MUL(frob_pows[12], mx18,  t, p);
    GEN mx2p12   = FPM_FPXQ_MUL(frob_pows[12], mx2,  t, p);
    GEN mx5p12   = FPM_FPXQ_MUL(frob_pows[12], mx5,  t, p);
    GEN mx21p12  = FPM_FPXQ_MUL(frob_pows[12], mx21,  t, p);
    GEN mx0p12   = FPM_FPXQ_MUL(frob_pows[12], mx0,  t, p);
    GEN imx22p12 = FPM_FPXQ_MUL(frob_pows[12], imx22,  t, p);
    GEN imx19p12 = FPM_FPXQ_MUL(frob_pows[12], imx19,  t, p);
    GEN imx7p12  = FPM_FPXQ_MUL(frob_pows[12], imx7,  t, p);
    GEN imx3p12  = FPM_FPXQ_MUL(frob_pows[12], imx3,  t, p);
    GEN imx1p12  = FPM_FPXQ_MUL(frob_pows[12], imx1,  t, p);

    GEN mx23p13  = FPM_FPXQ_MUL(frob_pows[13], mx23,  t, p);
    GEN mx20p13  = FPM_FPXQ_MUL(frob_pows[13], mx20,  t, p);
    GEN mx18p13  = FPM_FPXQ_MUL(frob_pows[13], mx18,  t, p);
    GEN mx2p13   = FPM_FPXQ_MUL(frob_pows[13], mx2,  t, p);
    GEN mx5p13   = FPM_FPXQ_MUL(frob_pows[13], mx5,  t, p);
    GEN mx21p13  = FPM_FPXQ_MUL(frob_pows[13], mx21,  t, p);
    GEN mx0p13   = FPM_FPXQ_MUL(frob_pows[13], mx0,  t, p);
    GEN imx22p13 = FPM_FPXQ_MUL(frob_pows[13], imx22,  t, p);
    GEN imx19p13 = FPM_FPXQ_MUL(frob_pows[13], imx19,  t, p);
    GEN imx6p13  = FPM_FPXQ_MUL(frob_pows[13], imx6,  t, p);
    GEN imx4p13  = FPM_FPXQ_MUL(frob_pows[13], imx4,  t, p);
    GEN imx1p13  = FPM_FPXQ_MUL(frob_pows[13], imx1,  t, p);

    GEN mx22p14  = FPM_FPXQ_MUL(frob_pows[14], mx22,  t, p);
    GEN mx20p14  = FPM_FPXQ_MUL(frob_pows[14], mx20,  t, p);
    GEN mx4p14   = FPM_FPXQ_MUL(frob_pows[14], mx4,  t, p);
    GEN mx18p14  = FPM_FPXQ_MUL(frob_pows[14], mx18,  t, p);
    GEN mx2p14   = FPM_FPXQ_MUL(frob_pows[14], mx2,  t, p);
    GEN mx0p14   = FPM_FPXQ_MUL(frob_pows[14], mx0,  t, p);
    GEN imx19p14 = FPM_FPXQ_MUL(frob_pows[14], imx19,  t, p);
    GEN imx3p14  = FPM_FPXQ_MUL(frob_pows[14], imx3,  t, p);
    GEN imx21p14 = FPM_FPXQ_MUL(frob_pows[14], imx21,  t, p);
    GEN imx1p14  = FPM_FPXQ_MUL(frob_pows[14], imx1,  t, p);

    GEN mx4p15   = FPM_FPXQ_MUL(frob_pows[15], mx4,  t, p);
    GEN mx2p15   = FPM_FPXQ_MUL(frob_pows[15], mx2,  t, p);
    GEN mx0p15   = FPM_FPXQ_MUL(frob_pows[15], mx0,  t, p);
    GEN imx3p15  = FPM_FPXQ_MUL(frob_pows[15], imx3,  t, p);
    GEN imx1p15  = FPM_FPXQ_MUL(frob_pows[15], imx1,  t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx20, mx35, t, p);
    y0 = FPXQ_MUL(y0, mx33, t, p);
    y0 = FPXQ_MUL(y0, imx3, t, p);
    y0 = FPXQ_MUL(y0, imx34, t, p);
    y0 = FPXQ_MUL(y0, mx20p1, t, p);
    y0 = FPXQ_MUL(y0, mx15p1, t, p);
    y0 = FPXQ_MUL(y0, mx35p1, t, p);
    y0 = FPXQ_MUL(y0, mx33p1, t, p);
    y0 = FPXQ_MUL(y0, mx0p1, t, p);
    y0 = FPXQ_MUL(y0, imx19p1, t, p);
    y0 = FPXQ_MUL(y0, imx3p1, t, p);
    y0 = FPXQ_MUL(y0, imx34p1, t, p);
    y0 = FPXQ_MUL(y0, mx20p2, t, p);
    y0 = FPXQ_MUL(y0, mx34p2, t, p);
    y0 = FPXQ_MUL(y0, mx18p2, t, p);
    y0 = FPXQ_MUL(y0, mx14p2, t, p);
    y0 = FPXQ_MUL(y0, mx0p2, t, p);
    y0 = FPXQ_MUL(y0, imx31p2, t, p);
    y0 = FPXQ_MUL(y0, imx19p2, t, p);
    y0 = FPXQ_MUL(y0, imx3p2, t, p);
    y0 = FPXQ_MUL(y0, mx31p3, t, p);
    y0 = FPXQ_MUL(y0, mx20p3, t, p);
    y0 = FPXQ_MUL(y0, mx16p3, t, p);
    y0 = FPXQ_MUL(y0, mx18p3, t, p);
    y0 = FPXQ_MUL(y0, mx0p3, t, p);
    y0 = FPXQ_MUL(y0, imx32p3, t, p);
    y0 = FPXQ_MUL(y0, imx19p3, t, p);
    y0 = FPXQ_MUL(y0, imx15p3, t, p);
    y0 = FPXQ_MUL(y0, imx3p3, t, p);
    y0 = FPXQ_MUL(y0, imx30p3, t, p);
    y0 = FPXQ_MUL(y0, mx31p4, t, p);
    y0 = FPXQ_MUL(y0, mx20p4, t, p);
    y0 = FPXQ_MUL(y0, mx15p4, t, p);
    y0 = FPXQ_MUL(y0, mx18p4, t, p);
    y0 = FPXQ_MUL(y0, mx0p4, t, p);
    y0 = FPXQ_MUL(y0, imx32p4, t, p);
    y0 = FPXQ_MUL(y0, imx19p4, t, p);
    y0 = FPXQ_MUL(y0, imx12p4, t, p);
    y0 = FPXQ_MUL(y0, imx3p4, t, p);
    y0 = FPXQ_MUL(y0, imx30p4, t, p);
    y0 = FPXQ_MUL(y0, mx28p5, t, p);
    y0 = FPXQ_MUL(y0, mx20p5, t, p);
    y0 = FPXQ_MUL(y0, mx18p5, t, p);
    y0 = FPXQ_MUL(y0, mx0p5, t, p);
    y0 = FPXQ_MUL(y0, imx31p5, t, p);
    y0 = FPXQ_MUL(y0, imx19p5, t, p);
    y0 = FPXQ_MUL(y0, imx3p5, t, p);
    y0 = FPXQ_MUL(y0, imx11p5, t, p);
    y0 = FPXQ_MUL(y0, mx29p6, t, p);
    y0 = FPXQ_MUL(y0, mx20p6, t, p);
    y0 = FPXQ_MUL(y0, mx12p6, t, p);
    y0 = FPXQ_MUL(y0, mx18p6, t, p);
    y0 = FPXQ_MUL(y0, mx27p6, t, p);
    y0 = FPXQ_MUL(y0, mx0p6, t, p);
    y0 = FPXQ_MUL(y0, imx28p6, t, p);
    y0 = FPXQ_MUL(y0, imx19p6, t, p);
    y0 = FPXQ_MUL(y0, imx13p6, t, p);
    y0 = FPXQ_MUL(y0, imx3p6, t, p);
    y0 = FPXQ_MUL(y0, mx29p7, t, p);
    y0 = FPXQ_MUL(y0, mx20p7, t, p);
    y0 = FPXQ_MUL(y0, mx9p7, t, p);
    y0 = FPXQ_MUL(y0, mx18p7, t, p);
    y0 = FPXQ_MUL(y0, mx27p7, t, p);
    y0 = FPXQ_MUL(y0, mx0p7, t, p);
    y0 = FPXQ_MUL(y0, imx28p7, t, p);
    y0 = FPXQ_MUL(y0, imx19p7, t, p);
    y0 = FPXQ_MUL(y0, imx12p7, t, p);
    y0 = FPXQ_MUL(y0, imx3p7, t, p);
    y0 = FPXQ_MUL(y0, mx28p8, t, p);
    y0 = FPXQ_MUL(y0, mx20p8, t, p);
    y0 = FPXQ_MUL(y0, mx18p8, t, p);
    y0 = FPXQ_MUL(y0, mx8p8, t, p);
    y0 = FPXQ_MUL(y0, mx0p8, t, p);
    y0 = FPXQ_MUL(y0, imx25p8, t, p);
    y0 = FPXQ_MUL(y0, imx19p8, t, p);
    y0 = FPXQ_MUL(y0, imx3p8, t, p);
    y0 = FPXQ_MUL(y0, mx25p9, t, p);
    y0 = FPXQ_MUL(y0, mx20p9, t, p);
    y0 = FPXQ_MUL(y0, mx10p9, t, p);
    y0 = FPXQ_MUL(y0, mx18p9, t, p);
    y0 = FPXQ_MUL(y0, mx0p9, t, p);
    y0 = FPXQ_MUL(y0, imx26p9, t, p);
    y0 = FPXQ_MUL(y0, imx19p9, t, p);
    y0 = FPXQ_MUL(y0, imx9p9, t, p);
    y0 = FPXQ_MUL(y0, imx3p9, t, p);
    y0 = FPXQ_MUL(y0, imx24p9, t, p);
    y0 = FPXQ_MUL(y0, mx25p10, t, p);
    y0 = FPXQ_MUL(y0, mx20p10, t, p);
    y0 = FPXQ_MUL(y0, mx9p10, t, p);
    y0 = FPXQ_MUL(y0, mx18p10, t, p);
    y0 = FPXQ_MUL(y0, mx0p10, t, p);
    y0 = FPXQ_MUL(y0, imx26p10, t, p);
    y0 = FPXQ_MUL(y0, imx19p10, t, p);
    y0 = FPXQ_MUL(y0, imx6p10, t, p);
    y0 = FPXQ_MUL(y0, imx3p10, t, p);
    y0 = FPXQ_MUL(y0, imx24p10, t, p);
    y0 = FPXQ_MUL(y0, mx22p11, t, p);
    y0 = FPXQ_MUL(y0, mx20p11, t, p);
    y0 = FPXQ_MUL(y0, mx18p11, t, p);
    y0 = FPXQ_MUL(y0, mx0p11, t, p);
    y0 = FPXQ_MUL(y0, imx25p11, t, p);
    y0 = FPXQ_MUL(y0, imx19p11, t, p);
    y0 = FPXQ_MUL(y0, imx3p11, t, p);
    y0 = FPXQ_MUL(y0, imx5p11, t, p);
    y0 = FPXQ_MUL(y0, mx23p12, t, p);
    y0 = FPXQ_MUL(y0, mx20p12, t, p);
    y0 = FPXQ_MUL(y0, mx6p12, t, p);
    y0 = FPXQ_MUL(y0, mx18p12, t, p);
    y0 = FPXQ_MUL(y0, mx21p12, t, p);
    y0 = FPXQ_MUL(y0, mx0p12, t, p);
    y0 = FPXQ_MUL(y0, imx22p12, t, p);
    y0 = FPXQ_MUL(y0, imx19p12, t, p);
    y0 = FPXQ_MUL(y0, imx7p12, t, p);
    y0 = FPXQ_MUL(y0, imx3p12, t, p);
    y0 = FPXQ_MUL(y0, mx23p13, t, p);
    y0 = FPXQ_MUL(y0, mx20p13, t, p);
    y0 = FPXQ_MUL(y0, mx18p13, t, p);
    y0 = FPXQ_MUL(y0, mx21p13, t, p);
    y0 = FPXQ_MUL(y0, mx0p13, t, p);
    y0 = FPXQ_MUL(y0, imx22p13, t, p);
    y0 = FPXQ_MUL(y0, imx19p13, t, p);
    y0 = FPXQ_MUL(y0, imx6p13, t, p);
    y0 = FPXQ_MUL(y0, mx22p14, t, p);
    y0 = FPXQ_MUL(y0, mx18p14, t, p);
    y0 = FPXQ_MUL(y0, mx0p14, t, p);
    y0 = FPXQ_MUL(y0, mx4p15, t, p);
    y0 = FPXQ_MUL(y0, mx0p15, t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx17, mx18, t, p);
    y1 = FPXQ_MUL(y1, imx19, t, p);
    y1 = FPXQ_MUL(y1, imx16p1, t, p);
    y1 = FPXQ_MUL(y1, imx1p1, t, p);
    y1 = FPXQ_MUL(y1, mx32p2, t, p);
    y1 = FPXQ_MUL(y1, imx33p2, t, p);
    y1 = FPXQ_MUL(y1, imx1p2, t, p);
    y1 = FPXQ_MUL(y1, imx14p3, t, p);
    y1 = FPXQ_MUL(y1, imx1p3, t, p);
    y1 = FPXQ_MUL(y1, mx13p4, t, p);
    y1 = FPXQ_MUL(y1, imx1p4, t, p);
    y1 = FPXQ_MUL(y1, mx30p5, t, p);
    y1 = FPXQ_MUL(y1, imx29p5, t, p);
    y1 = FPXQ_MUL(y1, imx1p5, t, p);
    y1 = FPXQ_MUL(y1, mx11p6, t, p);
    y1 = FPXQ_MUL(y1, imx1p6, t, p);
    y1 = FPXQ_MUL(y1, imx10p7, t, p);
    y1 = FPXQ_MUL(y1, imx1p7, t, p);
    y1 = FPXQ_MUL(y1, mx26p8, t, p);
    y1 = FPXQ_MUL(y1, imx27p8, t, p);
    y1 = FPXQ_MUL(y1, imx1p8, t, p);
    y1 = FPXQ_MUL(y1, imx8p9, t, p);
    y1 = FPXQ_MUL(y1, imx1p9, t, p);
    y1 = FPXQ_MUL(y1, mx7p10, t, p);
    y1 = FPXQ_MUL(y1, imx1p10, t, p);
    y1 = FPXQ_MUL(y1, mx24p11, t, p);
    y1 = FPXQ_MUL(y1, imx23p11, t, p);
    y1 = FPXQ_MUL(y1, imx1p11, t, p);
    y1 = FPXQ_MUL(y1, mx5p12, t, p);
    y1 = FPXQ_MUL(y1, imx1p12, t, p);
    y1 = FPXQ_MUL(y1, imx4p13, t, p);
    y1 = FPXQ_MUL(y1, imx1p13, t, p);
    y1 = FPXQ_MUL(y1, imx19p14, t, p);
    y1 = FPXQ_MUL(y1, imx21p14, t, p);
    y1 = FPXQ_MUL(y1, imx1p14, t, p);
    y1 = FPXQ_MUL(y1, imx3p15, t, p);
    y1 = FPXQ_MUL(y1, imx1p15, t, p);

    GEN y2 = FPXQ_MUL(mx20p14, mx2p15, t, p);

    GEN y3;
    y3 = FPXQ_MUL(mx0, mx16p2, t, p);
    y3 = FPXQ_MUL(y3, imx15p2, t, p);
    y3 = FPXQ_MUL(y3, mx12p5, t, p);
    y3 = FPXQ_MUL(y3, imx13p5, t, p);
    y3 = FPXQ_MUL(y3, mx10p8, t, p);
    y3 = FPXQ_MUL(y3, imx9p8, t, p);
    y3 = FPXQ_MUL(y3, mx6p11, t, p);
    y3 = FPXQ_MUL(y3, imx7p11, t, p);
    y3 = FPXQ_MUL(y3, mx4p14, t, p);

    GEN y4;
    y4 = FPXQ_MUL(mx2, imx1, t, p);
    y4 = FPXQ_MUL(y4, mx17p1, t, p);
    y4 = FPXQ_MUL(y4, mx2p1, t, p);
    y4 = FPXQ_MUL(y4, mx2p2, t, p);
    y4 = FPXQ_MUL(y4, mx2p3, t, p);
    y4 = FPXQ_MUL(y4, mx2p4, t, p);
    y4 = FPXQ_MUL(y4, imx14p4, t, p);
    y4 = FPXQ_MUL(y4, mx2p5, t, p);
    y4 = FPXQ_MUL(y4, mx2p6, t, p);
    y4 = FPXQ_MUL(y4, mx2p7, t, p);
    y4 = FPXQ_MUL(y4, mx11p7, t, p);
    y4 = FPXQ_MUL(y4, mx2p8, t, p);
    y4 = FPXQ_MUL(y4, mx2p9, t, p);
    y4 = FPXQ_MUL(y4, mx2p10, t, p);
    y4 = FPXQ_MUL(y4, imx8p10, t, p);
    y4 = FPXQ_MUL(y4, mx2p11, t, p);
    y4 = FPXQ_MUL(y4, mx2p12, t, p);
    y4 = FPXQ_MUL(y4, mx2p13, t, p);
    y4 = FPXQ_MUL(y4, mx5p13, t, p);
    y4 = FPXQ_MUL(y4, imx3p14, t, p);

    GEN y5 = mx2p14;
    
    GEN T0 = FPXQ_MUL(y5, y4, t, p);
    GEN T1 = FPXQ_MUL(y5, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y3, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_MUL(T2, y2, t, p);
    GEN T5 = FPXQ_MUL(T3, T2, t, p);
    GEN T6 = FPXQ_MUL(T4, y1, t, p);
    GEN T7 = FPXQ_MUL(T5, T4, t, p);
    GEN T8 = FPXQ_SQR(T6, t, p);
    GEN T9 = FPXQ_MUL(T8, T7, t, p);
    
    return T9;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_17_1188(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    MSG("Step 100\n");
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    MSG("Step 200\n");
    /*
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,   x, t, p);
    GEN mx2  = FPXQ_POW(mx1,   x, t, p);
    GEN mx3  = FPXQ_POW(mx2,   x, t, p);
    GEN mx4  = FPXQ_POW(mx3,   x, t, p);
    GEN mx5  = FPXQ_POW(mx4,   x, t, p);
    GEN mx6  = FPXQ_POW(mx5,   x, t, p);
    GEN mx7  = FPXQ_POW(mx6,   x, t, p);
    GEN mx8  = FPXQ_POW(mx7,   x, t, p);
    GEN mx9  = FPXQ_POW(mx8,   x, t, p);
    GEN mx10 = FPXQ_POW(mx9,   x, t, p);
    GEN mx11 = FPXQ_POW(mx10,   x, t, p);
    GEN mx12 = FPXQ_POW(mx11,   x, t, p);
    GEN mx13 = FPXQ_POW(mx12,   x, t, p);
    GEN mx14 = FPXQ_POW(mx13,   x, t, p);
    GEN mx15 = FPXQ_POW(mx14,   x, t, p);
    GEN mx16 = FPXQ_POW(mx15,   x, t, p);
    GEN mx17 = FPXQ_POW(mx16,   x, t, p);
    GEN mx18 = FPXQ_POW(mx17,   x, t, p);
    GEN mx19 = FPXQ_POW(mx18,   x, t, p);
    GEN mx20 = FPXQ_POW(mx19,   x, t, p);
    GEN mx21 = FPXQ_POW(mx20,   x, t, p);
    GEN mx22 = FPXQ_POW(mx21,   x, t, p);
    GEN mx23 = FPXQ_POW(mx22,   x, t, p);
    GEN mx24 = FPXQ_POW(mx23,   x, t, p);
    GEN mx25 = FPXQ_POW(mx24,   x, t, p);
    GEN mx26 = FPXQ_POW(mx25,   x, t, p);
    GEN mx27 = FPXQ_POW(mx26,   x, t, p);
    GEN mx28 = FPXQ_POW(mx27,   x, t, p);
    GEN mx29 = FPXQ_POW(mx28,   x, t, p);
    GEN mx30 = FPXQ_POW(mx29,   x, t, p);
    GEN mx31 = FPXQ_POW(mx30,   x, t, p);
    GEN mx32 = FPXQ_POW(mx31,   x, t, p);
    GEN mx33 = FPXQ_POW(mx32,   x, t, p);
    GEN mx34 = FPXQ_POW(mx33,   x, t, p);
    GEN mx35 = FPXQ_POW(mx34,   x, t, p);
    GEN mx36 = FPXQ_POW(mx35,   x, t, p);

    MSG("Step 300\n");

    GEN imx6  = FPXQ_INV(mx6,  t, p);    
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx10 = FPXQ_INV(mx10,  t, p);
    GEN imx12 = FPXQ_INV(mx12,  t, p);
    GEN imx14 = FPXQ_INV(mx14,  t, p);
    GEN imx16 = FPXQ_INV(mx16,  t, p);
    GEN imx18 = FPXQ_INV(mx18,  t, p);
    GEN imx20 = FPXQ_INV(mx20,  t, p);
    GEN imx22 = FPXQ_INV(mx22,  t, p);
    GEN imx24 = FPXQ_INV(mx24,  t, p);
    GEN imx26 = FPXQ_INV(mx26,  t, p);
    GEN imx28 = FPXQ_INV(mx28,  t, p);
    GEN imx30 = FPXQ_INV(mx30,  t, p);
    GEN imx32 = FPXQ_INV(mx32,  t, p);
    GEN imx34 = FPXQ_INV(mx34,  t, p);
    GEN imx36 = FPXQ_INV(mx36,  t, p);

    MSG("Step 400\n");

    GEN mx2p1  = FPM_FPXQ_MUL(frob_pows[1], mx2,  t, p);
    GEN mx4p1  = FPM_FPXQ_MUL(frob_pows[1], mx4,  t, p);
    GEN mx30p1 = FPM_FPXQ_MUL(frob_pows[1], mx30, t, p);
    GEN mx32p1 = FPM_FPXQ_MUL(frob_pows[1], mx32, t, p);
    GEN mx34p1 = FPM_FPXQ_MUL(frob_pows[1], mx34, t, p);
    GEN mx0p1  = FPM_FPXQ_MUL(frob_pows[1], mx0,  t, p);

    GEN mx2p2   = FPM_FPXQ_MUL(frob_pows[2], mx2,   t, p);
    GEN mx4p2   = FPM_FPXQ_MUL(frob_pows[2], mx4,   t, p);
    GEN mx0p2   = FPM_FPXQ_MUL(frob_pows[2], mx0,   t, p);
    GEN imx28p2 = FPM_FPXQ_MUL(frob_pows[2], imx28, t, p);
    GEN imx30p2 = FPM_FPXQ_MUL(frob_pows[2], imx30, t, p);
    GEN imx32p2 = FPM_FPXQ_MUL(frob_pows[2], imx32, t, p);

    GEN mx2p3  = FPM_FPXQ_MUL(frob_pows[3], mx2,  t, p);
    GEN mx4p3  = FPM_FPXQ_MUL(frob_pows[3], mx4,  t, p);
    GEN mx26p3 = FPM_FPXQ_MUL(frob_pows[3], mx26, t, p);
    GEN mx28p3 = FPM_FPXQ_MUL(frob_pows[3], mx28, t, p);
    GEN mx30p3 = FPM_FPXQ_MUL(frob_pows[3], mx30, t, p);
    GEN mx0p3  = FPM_FPXQ_MUL(frob_pows[3], mx0,  t, p);

    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,   t, p);
    GEN mx4p4   = FPM_FPXQ_MUL(frob_pows[4], mx4,   t, p);
    GEN mx0p4   = FPM_FPXQ_MUL(frob_pows[4], mx0,   t, p);
    GEN imx24p4 = FPM_FPXQ_MUL(frob_pows[4], imx24, t, p);
    GEN imx26p4 = FPM_FPXQ_MUL(frob_pows[4], imx26, t, p);
    GEN imx28p4 = FPM_FPXQ_MUL(frob_pows[4], imx28, t, p);

    GEN mx2p5  = FPM_FPXQ_MUL(frob_pows[5], mx2,  t, p);
    GEN mx4p5  = FPM_FPXQ_MUL(frob_pows[5], mx4,  t, p);
    GEN mx22p5 = FPM_FPXQ_MUL(frob_pows[5], mx22, t, p);
    GEN mx24p5 = FPM_FPXQ_MUL(frob_pows[5], mx24, t, p);
    GEN mx26p5 = FPM_FPXQ_MUL(frob_pows[5], mx26, t, p);
    GEN mx0p5  = FPM_FPXQ_MUL(frob_pows[5], mx0,  t, p);

    GEN mx2p6   = FPM_FPXQ_MUL(frob_pows[6], mx2,   t, p);
    GEN mx4p6   = FPM_FPXQ_MUL(frob_pows[6], mx4,   t, p);
    GEN mx0p6   = FPM_FPXQ_MUL(frob_pows[6], mx0,   t, p);
    GEN imx20p6 = FPM_FPXQ_MUL(frob_pows[6], imx20, t, p);
    GEN imx22p6 = FPM_FPXQ_MUL(frob_pows[6], imx22, t, p);
    GEN imx24p6 = FPM_FPXQ_MUL(frob_pows[6], imx24, t, p);

    GEN mx2p7  = FPM_FPXQ_MUL(frob_pows[7], mx2,  t, p);
    GEN mx4p7  = FPM_FPXQ_MUL(frob_pows[7], mx4,  t, p);
    GEN mx18p7 = FPM_FPXQ_MUL(frob_pows[7], mx18, t, p);
    GEN mx20p7 = FPM_FPXQ_MUL(frob_pows[7], mx20, t, p);
    GEN mx22p7 = FPM_FPXQ_MUL(frob_pows[7], mx22, t, p);
    GEN mx0p7  = FPM_FPXQ_MUL(frob_pows[7], mx0,  t, p);

    MSG("Step 440\n");

    GEN mx2p8   = FPM_FPXQ_MUL(frob_pows[8], mx2,   t, p);
    GEN mx4p8   = FPM_FPXQ_MUL(frob_pows[8], mx4,   t, p);
    GEN mx0p8   = FPM_FPXQ_MUL(frob_pows[8], mx0,   t, p);
    GEN imx16p8 = FPM_FPXQ_MUL(frob_pows[8], imx16, t, p);
    GEN imx18p8 = FPM_FPXQ_MUL(frob_pows[8], imx18, t, p);
    GEN imx20p8 = FPM_FPXQ_MUL(frob_pows[8], imx20, t, p);

    MSG("Step 450\n");

    GEN mx2p9  = FPM_FPXQ_MUL(frob_pows[9], mx2,  t, p);
    GEN mx4p9  = FPM_FPXQ_MUL(frob_pows[9], mx4,  t, p);
    GEN mx14p9 = FPM_FPXQ_MUL(frob_pows[9], mx14, t, p);
    GEN mx16p9 = FPM_FPXQ_MUL(frob_pows[9], mx16, t, p);
    GEN mx18p9 = FPM_FPXQ_MUL(frob_pows[9], mx18, t, p);
    GEN mx0p9  = FPM_FPXQ_MUL(frob_pows[9], mx0,  t, p);

    MSG("Step 460\n");

    GEN mx2p10   = FPM_FPXQ_MUL(frob_pows[10], mx2,   t, p);
    GEN mx4p10   = FPM_FPXQ_MUL(frob_pows[10], mx4,   t, p);
    GEN mx0p10   = FPM_FPXQ_MUL(frob_pows[10], mx0,   t, p);
    GEN imx12p10 = FPM_FPXQ_MUL(frob_pows[10], imx12, t, p);
    GEN imx14p10 = FPM_FPXQ_MUL(frob_pows[10], imx14, t, p);
    GEN imx16p10 = FPM_FPXQ_MUL(frob_pows[10], imx16, t, p);

    MSG("Step 465\n");

    GEN mx2p11  = FPM_FPXQ_MUL(frob_pows[11], mx2,  t, p);
    GEN mx4p11  = FPM_FPXQ_MUL(frob_pows[11], mx4,  t, p);
    GEN mx10p11 = FPM_FPXQ_MUL(frob_pows[11], mx10, t, p);
    GEN mx12p11 = FPM_FPXQ_MUL(frob_pows[11], mx12, t, p);
    GEN mx14p11 = FPM_FPXQ_MUL(frob_pows[11], mx14, t, p);
    GEN mx0p11  = FPM_FPXQ_MUL(frob_pows[11], mx0,  t, p);

    MSG("Step 470\n");

    GEN mx2p12   = FPM_FPXQ_MUL(frob_pows[12], mx2,   t, p);
    GEN mx4p12   = FPM_FPXQ_MUL(frob_pows[12], mx4,   t, p);
    GEN mx0p12   = FPM_FPXQ_MUL(frob_pows[12], mx0,   t, p);
    GEN imx8p12  = FPM_FPXQ_MUL(frob_pows[12], imx8,  t, p);
    GEN imx10p12 = FPM_FPXQ_MUL(frob_pows[12], imx10, t, p);
    GEN imx12p12 = FPM_FPXQ_MUL(frob_pows[12], imx12, t, p);

    MSG("Step 475\n");

    GEN mx2p13  = FPM_FPXQ_MUL(frob_pows[13], mx2,  t, p);
    GEN mx4p13  = FPM_FPXQ_MUL(frob_pows[13], mx4,  t, p);
    GEN mx6p13  = FPM_FPXQ_MUL(frob_pows[13], mx6,  t, p);
    GEN mx8p13  = FPM_FPXQ_MUL(frob_pows[13], mx8,  t, p);
    GEN mx10p13 = FPM_FPXQ_MUL(frob_pows[13], mx10, t, p);
    GEN mx0p13  = FPM_FPXQ_MUL(frob_pows[13], mx0,  t, p);

    MSG("Step 480\n");

    GEN mx2p14  = FPM_FPXQ_MUL(frob_pows[14], mx2,  t, p);
    GEN mx0p14  = FPM_FPXQ_MUL(frob_pows[14], mx0,  t, p);
    GEN imx8p14 = FPM_FPXQ_MUL(frob_pows[14], imx8, t, p);
    GEN imx6p14 = FPM_FPXQ_MUL(frob_pows[14], imx6, t, p);

    MSG("Step 485\n");

    GEN mx6p15 = FPM_FPXQ_MUL(frob_pows[15], mx6,  t, p);
    GEN mx4p15 = FPM_FPXQ_MUL(frob_pows[15], mx4,  t, p);
    GEN mx2p15 = FPM_FPXQ_MUL(frob_pows[15], mx2,  t, p);
    GEN mx0p15 = FPM_FPXQ_MUL(frob_pows[15], mx0,  t, p);

    MSG("Step 600\n");

    GEN y0;
    y0 = FPXQ_MUL(mx4, imx32, t, p);
    y0 = FPXQ_MUL(y0, imx36, t, p);
    y0 = FPXQ_MUL(y0, mx4p1, t, p);
    y0 = FPXQ_MUL(y0, mx30p1, t, p);
    y0 = FPXQ_MUL(y0, mx34p1, t, p);
    y0 = FPXQ_MUL(y0, mx0p1, t, p);
    y0 = FPXQ_MUL(y0, mx4p2, t, p);
    y0 = FPXQ_MUL(y0, mx0p2, t, p);
    y0 = FPXQ_MUL(y0, imx28p2, t, p);
    y0 = FPXQ_MUL(y0, imx32p2, t, p);
    y0 = FPXQ_MUL(y0, mx4p3, t, p);
    y0 = FPXQ_MUL(y0, mx26p3, t, p);
    y0 = FPXQ_MUL(y0, mx30p3, t, p);
    y0 = FPXQ_MUL(y0, mx0p3, t, p);
    y0 = FPXQ_MUL(y0, mx4p4, t, p);
    y0 = FPXQ_MUL(y0, mx0p4, t, p);
    y0 = FPXQ_MUL(y0, imx24p4, t, p);
    y0 = FPXQ_MUL(y0, imx28p4, t, p);
    y0 = FPXQ_MUL(y0, mx4p5, t, p);
    y0 = FPXQ_MUL(y0, mx22p5, t, p);
    y0 = FPXQ_MUL(y0, mx26p5, t, p);
    y0 = FPXQ_MUL(y0, mx0p5, t, p);
    y0 = FPXQ_MUL(y0, mx4p6, t, p);
    y0 = FPXQ_MUL(y0, mx0p6, t, p);
    y0 = FPXQ_MUL(y0, imx20p6, t, p);
    y0 = FPXQ_MUL(y0, imx24p6, t, p);
    y0 = FPXQ_MUL(y0, mx4p7, t, p);
    y0 = FPXQ_MUL(y0, mx18p7, t, p);
    y0 = FPXQ_MUL(y0, mx22p7, t, p);
    y0 = FPXQ_MUL(y0, mx0p7, t, p);
    y0 = FPXQ_MUL(y0, mx4p8, t, p);
    y0 = FPXQ_MUL(y0, mx0p8, t, p);
    y0 = FPXQ_MUL(y0, imx16p8, t, p);
    y0 = FPXQ_MUL(y0, imx20p8, t, p);
    y0 = FPXQ_MUL(y0, mx4p9, t, p);
    y0 = FPXQ_MUL(y0, mx14p9, t, p);
    y0 = FPXQ_MUL(y0, mx18p9, t, p);
    y0 = FPXQ_MUL(y0, mx0p9, t, p);
    y0 = FPXQ_MUL(y0, mx4p10, t, p);
    y0 = FPXQ_MUL(y0, mx0p10, t, p);
    y0 = FPXQ_MUL(y0, imx12p10, t, p);
    y0 = FPXQ_MUL(y0, imx16p10, t, p);
    y0 = FPXQ_MUL(y0, mx4p11, t, p);
    y0 = FPXQ_MUL(y0, mx10p11, t, p);
    y0 = FPXQ_MUL(y0, mx14p11, t, p);
    y0 = FPXQ_MUL(y0, mx0p11, t, p);
    y0 = FPXQ_MUL(y0, mx4p12, t, p);
    y0 = FPXQ_MUL(y0, mx0p12, t, p);
    y0 = FPXQ_MUL(y0, imx8p12, t, p);
    y0 = FPXQ_MUL(y0, imx12p12, t, p);
    y0 = FPXQ_MUL(y0, mx4p13, t, p);
    y0 = FPXQ_MUL(y0, mx6p13, t, p);
    y0 = FPXQ_MUL(y0, mx10p13, t, p);
    y0 = FPXQ_MUL(y0, mx0p13, t, p);
    y0 = FPXQ_MUL(y0, mx0p14, t, p);
    y0 = FPXQ_MUL(y0, imx8p14, t, p);
    y0 = FPXQ_MUL(y0, mx6p15, t, p);
    y0 = FPXQ_MUL(y0, mx0p15, t, p);

    MSG("Step 700\n");

    GEN y1;
    y1 = FPXQ_MUL(mx2, imx34, t, p);
    y1 = FPXQ_MUL(y1, mx2p1, t, p);
    y1 = FPXQ_MUL(y1, mx32p1, t, p);
    y1 = FPXQ_MUL(y1, mx2p2, t, p);
    y1 = FPXQ_MUL(y1, imx30p2, t, p);
    y1 = FPXQ_MUL(y1, mx2p3, t, p);
    y1 = FPXQ_MUL(y1, mx28p3, t, p);
    y1 = FPXQ_MUL(y1, mx2p4, t, p);
    y1 = FPXQ_MUL(y1, imx26p4, t, p);
    y1 = FPXQ_MUL(y1, mx2p5, t, p);
    y1 = FPXQ_MUL(y1, mx24p5, t, p);
    y1 = FPXQ_MUL(y1, mx2p6, t, p);
    y1 = FPXQ_MUL(y1, imx22p6, t, p);
    y1 = FPXQ_MUL(y1, mx2p7, t, p);
    y1 = FPXQ_MUL(y1, mx20p7, t, p);
    y1 = FPXQ_MUL(y1, mx2p8, t, p);
    y1 = FPXQ_MUL(y1, imx18p8, t, p);
    y1 = FPXQ_MUL(y1, mx2p9, t, p);
    y1 = FPXQ_MUL(y1, mx16p9, t, p);
    y1 = FPXQ_MUL(y1, mx2p10, t, p);
    y1 = FPXQ_MUL(y1, imx14p10, t, p);
    y1 = FPXQ_MUL(y1, mx2p11, t, p);
    y1 = FPXQ_MUL(y1, mx12p11, t, p);
    y1 = FPXQ_MUL(y1, mx2p12, t, p);
    y1 = FPXQ_MUL(y1, imx10p12, t, p);
    y1 = FPXQ_MUL(y1, mx2p13, t, p);
    y1 = FPXQ_MUL(y1, mx8p13, t, p);
    y1 = FPXQ_MUL(y1, mx2p14, t, p);
    y1 = FPXQ_MUL(y1, imx6p14, t, p);
    
    MSG("Step 800\n");
    
    GEN y2 = FPXQ_MUL(mx4p15, mx2p15, t, p);
    GEN y3 = mx0;
    */
    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0,   x, t, p);
    GEN mx2 = FPXQ_POW(mx1,   x, t, p);
    GEN mx3 = FPXQ_POW(mx2,   x, t, p);
    GEN mx4 = FPXQ_POW(mx3,   x, t, p);
    GEN mx5 = FPXQ_POW(mx4,   x, t, p);
    GEN mx6 = FPXQ_POW(mx5,   x, t, p);
    GEN mx7 = FPXQ_POW(mx6,   x, t, p);
    GEN mx8 = FPXQ_POW(mx7,   x, t, p);
    GEN mx9 = FPXQ_POW(mx8,   x, t, p);
    GEN mx10 = FPXQ_POW(mx9,   x, t, p);
    GEN mx11 = FPXQ_POW(mx10,   x, t, p);
    GEN mx12 = FPXQ_POW(mx11,   x, t, p);
    GEN mx13 = FPXQ_POW(mx12,   x, t, p);
    GEN mx14 = FPXQ_POW(mx13,   x, t, p);
    GEN mx15 = FPXQ_POW(mx14,   x, t, p);
    GEN mx16 = FPXQ_POW(mx15,   x, t, p);
    GEN mx17 = FPXQ_POW(mx16,   x, t, p);
    GEN mx18 = FPXQ_POW(mx17,   x, t, p);
    GEN mx19 = FPXQ_POW(mx18,   x, t, p);
    GEN mx20 = FPXQ_POW(mx19,   x, t, p);
    GEN mx21 = FPXQ_POW(mx20,   x, t, p);
    GEN mx22 = FPXQ_POW(mx21,   x, t, p);
    GEN mx23 = FPXQ_POW(mx22,   x, t, p);
    GEN mx24 = FPXQ_POW(mx23,   x, t, p);
    GEN mx25 = FPXQ_POW(mx24,   x, t, p);
    GEN mx26 = FPXQ_POW(mx25,   x, t, p);
    GEN mx27 = FPXQ_POW(mx26,   x, t, p);
    GEN mx28 = FPXQ_POW(mx27,   x, t, p);
    GEN mx29 = FPXQ_POW(mx28,   x, t, p);
    GEN mx30 = FPXQ_POW(mx29,   x, t, p);
    GEN mx31 = FPXQ_POW(mx30,   x, t, p);
    GEN mx32 = FPXQ_POW(mx31,   x, t, p);
    GEN mx33 = FPXQ_POW(mx32,   x, t, p);
    GEN mx34 = FPXQ_POW(mx33,   x, t, p);
    GEN mx35 = FPXQ_POW(mx34,   x, t, p);
    GEN mx36 = FPXQ_POW(mx35,   x, t, p);

    GEN imx6 = FPXQ_INV(mx6,  t, p);
    GEN imx8 = FPXQ_INV(mx8,  t, p);
    GEN imx10 = FPXQ_INV(mx10,  t, p);
    GEN imx12 = FPXQ_INV(mx12,  t, p);
    GEN imx14 = FPXQ_INV(mx14,  t, p);
    GEN imx16 = FPXQ_INV(mx16,  t, p);
    GEN imx18 = FPXQ_INV(mx18,  t, p);
    GEN imx20 = FPXQ_INV(mx20,  t, p);
    GEN imx22 = FPXQ_INV(mx22,  t, p);
    GEN imx24 = FPXQ_INV(mx24,  t, p);
    GEN imx26 = FPXQ_INV(mx26,  t, p);
    GEN imx28 = FPXQ_INV(mx28,  t, p);
    GEN imx30 = FPXQ_INV(mx30,  t, p);
    GEN imx32 = FPXQ_INV(mx32,  t, p);
    GEN imx34 = FPXQ_INV(mx34,  t, p);

    GEN imx36 = FPXQ_INV(mx36,  t, p);

    GEN mx2p1 = FPM_FPXQ_MUL(frob_pows[1], mx2,  t, p);
    GEN mx4p1 = FPM_FPXQ_MUL(frob_pows[1], mx4,  t, p);
    GEN mx30p1 = FPM_FPXQ_MUL(frob_pows[1], mx30,  t, p);
    GEN mx32p1 = FPM_FPXQ_MUL(frob_pows[1], mx32,  t, p);
    GEN mx34p1 = FPM_FPXQ_MUL(frob_pows[1], mx34,  t, p);
    GEN mx0p1 = FPM_FPXQ_MUL(frob_pows[1], mx0,  t, p);

    GEN mx2p2 = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx4p2 = FPM_FPXQ_MUL(frob_pows[2], mx4,  t, p);
    GEN mx0p2 = FPM_FPXQ_MUL(frob_pows[2], mx0,  t, p);
    GEN imx28p2 = FPM_FPXQ_MUL(frob_pows[2], imx28,  t, p);
    GEN imx30p2 = FPM_FPXQ_MUL(frob_pows[2], imx30,  t, p);
    GEN imx32p2 = FPM_FPXQ_MUL(frob_pows[2], imx32,  t, p);

    GEN mx2p3 = FPM_FPXQ_MUL(frob_pows[3], mx2,  t, p);
    GEN mx4p3 = FPM_FPXQ_MUL(frob_pows[3], mx4,  t, p);
    GEN mx26p3 = FPM_FPXQ_MUL(frob_pows[3], mx26,  t, p);
    GEN mx28p3 = FPM_FPXQ_MUL(frob_pows[3], mx28,  t, p);
    GEN mx30p3 = FPM_FPXQ_MUL(frob_pows[3], mx30,  t, p);
    GEN mx0p3 = FPM_FPXQ_MUL(frob_pows[3], mx0,  t, p);

    GEN mx2p4 = FPM_FPXQ_MUL(frob_pows[4], mx2,  t, p);
    GEN mx4p4 = FPM_FPXQ_MUL(frob_pows[4], mx4,  t, p);
    GEN mx0p4 = FPM_FPXQ_MUL(frob_pows[4], mx0,  t, p);
    GEN imx24p4 = FPM_FPXQ_MUL(frob_pows[4], imx24,  t, p);
    GEN imx26p4 = FPM_FPXQ_MUL(frob_pows[4], imx26,  t, p);
    GEN imx28p4 = FPM_FPXQ_MUL(frob_pows[4], imx28,  t, p);

    GEN mx2p5 = FPM_FPXQ_MUL(frob_pows[5], mx2,  t, p);
    GEN mx4p5 = FPM_FPXQ_MUL(frob_pows[5], mx4,  t, p);
    GEN mx22p5 = FPM_FPXQ_MUL(frob_pows[5], mx22,  t, p);
    GEN mx24p5 = FPM_FPXQ_MUL(frob_pows[5], mx24,  t, p);
    GEN mx26p5 = FPM_FPXQ_MUL(frob_pows[5], mx26,  t, p);
    GEN mx0p5 = FPM_FPXQ_MUL(frob_pows[5], mx0,  t, p);

    GEN mx2p6 = FPM_FPXQ_MUL(frob_pows[6], mx2,  t, p);
    GEN mx4p6 = FPM_FPXQ_MUL(frob_pows[6], mx4,  t, p);
    GEN mx0p6 = FPM_FPXQ_MUL(frob_pows[6], mx0,  t, p);
    GEN imx20p6 = FPM_FPXQ_MUL(frob_pows[6], imx20,  t, p);
    GEN imx22p6 = FPM_FPXQ_MUL(frob_pows[6], imx22,  t, p);
    GEN imx24p6 = FPM_FPXQ_MUL(frob_pows[6], imx24,  t, p);

    GEN mx2p7 = FPM_FPXQ_MUL(frob_pows[7], mx2,  t, p);
    GEN mx4p7 = FPM_FPXQ_MUL(frob_pows[7], mx4,  t, p);
    GEN mx18p7 = FPM_FPXQ_MUL(frob_pows[7], mx18,  t, p);
    GEN mx20p7 = FPM_FPXQ_MUL(frob_pows[7], mx20,  t, p);
    GEN mx22p7 = FPM_FPXQ_MUL(frob_pows[7], mx22,  t, p);
    GEN mx0p7 = FPM_FPXQ_MUL(frob_pows[7], mx0,  t, p);

    GEN mx2p8 = FPM_FPXQ_MUL(frob_pows[8], mx2,  t, p);
    GEN mx4p8 = FPM_FPXQ_MUL(frob_pows[8], mx4,  t, p);
    GEN mx0p8 = FPM_FPXQ_MUL(frob_pows[8], mx0,  t, p);
    GEN imx16p8 = FPM_FPXQ_MUL(frob_pows[8], imx16,  t, p);
    GEN imx18p8 = FPM_FPXQ_MUL(frob_pows[8], imx18,  t, p);
    GEN imx20p8 = FPM_FPXQ_MUL(frob_pows[8], imx20,  t, p);

    GEN mx2p9 = FPM_FPXQ_MUL(frob_pows[9], mx2,  t, p);
    GEN mx4p9 = FPM_FPXQ_MUL(frob_pows[9], mx4,  t, p);
    GEN mx14p9 = FPM_FPXQ_MUL(frob_pows[9], mx14,  t, p);
    GEN mx16p9 = FPM_FPXQ_MUL(frob_pows[9], mx16,  t, p);
    GEN mx18p9 = FPM_FPXQ_MUL(frob_pows[9], mx18,  t, p);
    GEN mx0p9 = FPM_FPXQ_MUL(frob_pows[9], mx0,  t, p);

    GEN mx2p10 = FPM_FPXQ_MUL(frob_pows[10], mx2,  t, p);
    GEN mx4p10 = FPM_FPXQ_MUL(frob_pows[10], mx4,  t, p);
    GEN mx0p10 = FPM_FPXQ_MUL(frob_pows[10], mx0,  t, p);
    GEN imx12p10 = FPM_FPXQ_MUL(frob_pows[10], imx12,  t, p);
    GEN imx14p10 = FPM_FPXQ_MUL(frob_pows[10], imx14,  t, p);
    GEN imx16p10 = FPM_FPXQ_MUL(frob_pows[10], imx16,  t, p);

    GEN mx2p11 = FPM_FPXQ_MUL(frob_pows[11], mx2,  t, p);
    GEN mx4p11 = FPM_FPXQ_MUL(frob_pows[11], mx4,  t, p);
    GEN mx10p11 = FPM_FPXQ_MUL(frob_pows[11], mx10,  t, p);
    GEN mx12p11 = FPM_FPXQ_MUL(frob_pows[11], mx12,  t, p);
    GEN mx14p11 = FPM_FPXQ_MUL(frob_pows[11], mx14,  t, p);
    GEN mx0p11 = FPM_FPXQ_MUL(frob_pows[11], mx0,  t, p);

    GEN mx2p12 = FPM_FPXQ_MUL(frob_pows[12], mx2,  t, p);
    GEN mx4p12 = FPM_FPXQ_MUL(frob_pows[12], mx4,  t, p);
    GEN mx0p12 = FPM_FPXQ_MUL(frob_pows[12], mx0,  t, p);
    GEN imx8p12 = FPM_FPXQ_MUL(frob_pows[12], imx8,  t, p);
    GEN imx10p12 = FPM_FPXQ_MUL(frob_pows[12], imx10,  t, p);
    GEN imx12p12 = FPM_FPXQ_MUL(frob_pows[12], imx12,  t, p);

    GEN mx2p13 = FPM_FPXQ_MUL(frob_pows[13], mx2,  t, p);
    GEN mx4p13 = FPM_FPXQ_MUL(frob_pows[13], mx4,  t, p);
    GEN mx6p13 = FPM_FPXQ_MUL(frob_pows[13], mx6,  t, p);
    GEN mx8p13 = FPM_FPXQ_MUL(frob_pows[13], mx8,  t, p);
    GEN mx10p13 = FPM_FPXQ_MUL(frob_pows[13], mx10,  t, p);
    GEN mx0p13 = FPM_FPXQ_MUL(frob_pows[13], mx0,  t, p);

    GEN mx2p14 = FPM_FPXQ_MUL(frob_pows[14], mx2,  t, p);
    GEN mx0p14 = FPM_FPXQ_MUL(frob_pows[14], mx0,  t, p);
    GEN imx8p14 = FPM_FPXQ_MUL(frob_pows[14], imx8,  t, p);
    GEN imx6p14 = FPM_FPXQ_MUL(frob_pows[14], imx6,  t, p);

    GEN mx6p15 = FPM_FPXQ_MUL(frob_pows[15], mx6,  t, p);
    GEN mx4p15 = FPM_FPXQ_MUL(frob_pows[15], mx4,  t, p);
    GEN mx2p15 = FPM_FPXQ_MUL(frob_pows[15], mx2,  t, p);
    GEN mx0p15 = FPM_FPXQ_MUL(frob_pows[15], mx0,  t, p);

    //
    // Coef 1
    //
    GEN y0;
    y0 = FPXQ_MUL(mx4, imx32, t, p);
    y0 = FPXQ_MUL(y0, imx36, t, p);
    y0 = FPXQ_MUL(y0, mx4p1, t, p);
    y0 = FPXQ_MUL(y0, mx30p1, t, p);
    y0 = FPXQ_MUL(y0, mx34p1, t, p);
    y0 = FPXQ_MUL(y0, mx0p1, t, p);
    y0 = FPXQ_MUL(y0, mx4p2, t, p);
    y0 = FPXQ_MUL(y0, mx0p2, t, p);
    y0 = FPXQ_MUL(y0, imx28p2, t, p);
    y0 = FPXQ_MUL(y0, imx32p2, t, p);
    y0 = FPXQ_MUL(y0, mx4p3, t, p);
    y0 = FPXQ_MUL(y0, mx26p3, t, p);
    y0 = FPXQ_MUL(y0, mx30p3, t, p);
    y0 = FPXQ_MUL(y0, mx0p3, t, p);
    y0 = FPXQ_MUL(y0, mx4p4, t, p);
    y0 = FPXQ_MUL(y0, mx0p4, t, p);
    y0 = FPXQ_MUL(y0, imx24p4, t, p);
    y0 = FPXQ_MUL(y0, imx28p4, t, p);
    y0 = FPXQ_MUL(y0, mx4p5, t, p);
    y0 = FPXQ_MUL(y0, mx22p5, t, p);
    y0 = FPXQ_MUL(y0, mx26p5, t, p);
    y0 = FPXQ_MUL(y0, mx0p5, t, p);
    y0 = FPXQ_MUL(y0, mx4p6, t, p);
    y0 = FPXQ_MUL(y0, mx0p6, t, p);
    y0 = FPXQ_MUL(y0, imx20p6, t, p);
    y0 = FPXQ_MUL(y0, imx24p6, t, p);
    y0 = FPXQ_MUL(y0, mx4p7, t, p);
    y0 = FPXQ_MUL(y0, mx18p7, t, p);
    y0 = FPXQ_MUL(y0, mx22p7, t, p);
    y0 = FPXQ_MUL(y0, mx0p7, t, p);
    y0 = FPXQ_MUL(y0, mx4p8, t, p);
    y0 = FPXQ_MUL(y0, mx0p8, t, p);
    y0 = FPXQ_MUL(y0, imx16p8, t, p);
    y0 = FPXQ_MUL(y0, imx20p8, t, p);
    y0 = FPXQ_MUL(y0, mx4p9, t, p);
    y0 = FPXQ_MUL(y0, mx14p9, t, p);
    y0 = FPXQ_MUL(y0, mx18p9, t, p);
    y0 = FPXQ_MUL(y0, mx0p9, t, p);
    y0 = FPXQ_MUL(y0, mx4p10, t, p);
    y0 = FPXQ_MUL(y0, mx0p10, t, p);
    y0 = FPXQ_MUL(y0, imx12p10, t, p);
    y0 = FPXQ_MUL(y0, imx16p10, t, p);
    y0 = FPXQ_MUL(y0, mx4p11, t, p);
    y0 = FPXQ_MUL(y0, mx10p11, t, p);
    y0 = FPXQ_MUL(y0, mx14p11, t, p);
    y0 = FPXQ_MUL(y0, mx0p11, t, p);
    y0 = FPXQ_MUL(y0, mx4p12, t, p);
    y0 = FPXQ_MUL(y0, mx0p12, t, p);
    y0 = FPXQ_MUL(y0, imx8p12, t, p);
    y0 = FPXQ_MUL(y0, imx12p12, t, p);
    y0 = FPXQ_MUL(y0, mx4p13, t, p);
    y0 = FPXQ_MUL(y0, mx6p13, t, p);
    y0 = FPXQ_MUL(y0, mx10p13, t, p);
    y0 = FPXQ_MUL(y0, mx0p13, t, p);
    y0 = FPXQ_MUL(y0, mx0p14, t, p);
    y0 = FPXQ_MUL(y0, imx8p14, t, p);
    y0 = FPXQ_MUL(y0, mx6p15, t, p);
    y0 = FPXQ_MUL(y0, mx0p15, t, p);

    //
    // Coef 2
    //
    GEN y1;
    y1 = FPXQ_MUL(mx2, imx34, t, p);
    y1 = FPXQ_MUL(y1, mx2p1, t, p);
    y1 = FPXQ_MUL(y1, mx32p1, t, p);
    y1 = FPXQ_MUL(y1, mx2p2, t, p);
    y1 = FPXQ_MUL(y1, imx30p2, t, p);
    y1 = FPXQ_MUL(y1, mx2p3, t, p);
    y1 = FPXQ_MUL(y1, mx28p3, t, p);
    y1 = FPXQ_MUL(y1, mx2p4, t, p);
    y1 = FPXQ_MUL(y1, imx26p4, t, p);
    y1 = FPXQ_MUL(y1, mx2p5, t, p);
    y1 = FPXQ_MUL(y1, mx24p5, t, p);
    y1 = FPXQ_MUL(y1, mx2p6, t, p);
    y1 = FPXQ_MUL(y1, imx22p6, t, p);
    y1 = FPXQ_MUL(y1, mx2p7, t, p);
    y1 = FPXQ_MUL(y1, mx20p7, t, p);
    y1 = FPXQ_MUL(y1, mx2p8, t, p);
    y1 = FPXQ_MUL(y1, imx18p8, t, p);
    y1 = FPXQ_MUL(y1, mx2p9, t, p);
    y1 = FPXQ_MUL(y1, mx16p9, t, p);
    y1 = FPXQ_MUL(y1, mx2p10, t, p);
    y1 = FPXQ_MUL(y1, imx14p10, t, p);
    y1 = FPXQ_MUL(y1, mx2p11, t, p);
    y1 = FPXQ_MUL(y1, mx12p11, t, p);
    y1 = FPXQ_MUL(y1, mx2p12, t, p);
    y1 = FPXQ_MUL(y1, imx10p12, t, p);
    y1 = FPXQ_MUL(y1, mx2p13, t, p);
    y1 = FPXQ_MUL(y1, mx8p13, t, p);
    y1 = FPXQ_MUL(y1, mx2p14, t, p);
    y1 = FPXQ_MUL(y1, imx6p14, t, p);

    //
    // Coef 3
    //
    GEN y2 = FPXQ_MUL(mx4p15, mx2p15, t, p);

    //
    // Coef 5
    //
    GEN y3 = mx0;
    
    GEN T0 = FPXQ_MUL(y3, y2, t, p);
    GEN T1 = FPXQ_MUL(y3, y1, t, p);
    GEN T2 = FPXQ_MUL(T1, T0, t, p);
    GEN T3 = FPXQ_MUL(T0, y0, t, p);
    GEN T4 = FPXQ_SQR(T2, t, p);
    GEN T5 = FPXQ_MUL(T4, T3, t, p);
    
    return T5;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_duan_et_al_15_1500(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;

    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);

    GEN imx0  = FPXQ_INV(mx0,  t, p);
    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx2  = FPXQ_INV(mx2,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);

    GEN mx8p1  = FPM_FPXQ_MUL(frob_pows[1], mx8,  t, p);
    GEN mx5p1  = FPM_FPXQ_MUL(frob_pows[1], mx5,  t, p);
    GEN mx3p1  = FPM_FPXQ_MUL(frob_pows[1], mx3,  t, p);
    GEN mx10p1 = FPM_FPXQ_MUL(frob_pows[1], mx10, t, p);
    GEN mx1p1  = FPM_FPXQ_MUL(frob_pows[1], mx1,  t, p);
    GEN imx4p1 = FPM_FPXQ_MUL(frob_pows[1], imx4, t, p);
    GEN imx2p1 = FPM_FPXQ_MUL(frob_pows[1], imx2, t, p);
    GEN imx9p1 = FPM_FPXQ_MUL(frob_pows[1], imx9, t, p);

    GEN mx7p2  = FPM_FPXQ_MUL(frob_pows[2], mx7,  t, p);
    GEN mx4p2  = FPM_FPXQ_MUL(frob_pows[2], mx4,  t, p);
    GEN mx2p2  = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx9p2  = FPM_FPXQ_MUL(frob_pows[2], mx9,  t, p);
    GEN mx0p2  = FPM_FPXQ_MUL(frob_pows[2], mx0,  t, p);
    GEN imx8p2 = FPM_FPXQ_MUL(frob_pows[2], imx8, t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx1p2 = FPM_FPXQ_MUL(frob_pows[2], imx1, t, p);

    GEN mx8p3  = FPM_FPXQ_MUL(frob_pows[3], mx8,  t, p);
    GEN mx6p3  = FPM_FPXQ_MUL(frob_pows[3], mx6,  t, p);
    GEN mx1p3  = FPM_FPXQ_MUL(frob_pows[3], mx1,  t, p);
    GEN imx7p3 = FPM_FPXQ_MUL(frob_pows[3], imx7, t, p);
    GEN imx2p3 = FPM_FPXQ_MUL(frob_pows[3], imx2, t, p);
    GEN imx0p3 = FPM_FPXQ_MUL(frob_pows[3], imx0, t, p);

    GEN mx7p4  = FPM_FPXQ_MUL(frob_pows[4], mx7,  t, p);
    GEN mx3p4  = FPM_FPXQ_MUL(frob_pows[4], mx3,  t, p);
    GEN mx5p4  = FPM_FPXQ_MUL(frob_pows[4], mx5,  t, p);
    GEN mx0p4  = FPM_FPXQ_MUL(frob_pows[4], mx0,  t, p);
    GEN imx6p4 = FPM_FPXQ_MUL(frob_pows[4], imx6, t, p);
    GEN imx2p4 = FPM_FPXQ_MUL(frob_pows[4], imx2, t, p);
    GEN imx1p4 = FPM_FPXQ_MUL(frob_pows[4], imx1, t, p);

    GEN mx4p5  = FPM_FPXQ_MUL(frob_pows[5], mx4,  t, p);
    GEN mx6p5  = FPM_FPXQ_MUL(frob_pows[5], mx6,  t, p);
    GEN mx2p5  = FPM_FPXQ_MUL(frob_pows[5], mx2,  t, p);
    GEN imx5p5 = FPM_FPXQ_MUL(frob_pows[5], imx5, t, p);
    GEN imx3p5 = FPM_FPXQ_MUL(frob_pows[5], imx3, t, p);
    GEN imx1p5 = FPM_FPXQ_MUL(frob_pows[5], imx1, t, p);

    GEN mx5p6  = FPM_FPXQ_MUL(frob_pows[6], mx5,  t, p);
    GEN mx3p6  = FPM_FPXQ_MUL(frob_pows[6], mx3,  t, p);
    GEN mx1p6  = FPM_FPXQ_MUL(frob_pows[6], mx1,  t, p);
    GEN imx4p6 = FPM_FPXQ_MUL(frob_pows[6], imx4, t, p);
    GEN imx2p6 = FPM_FPXQ_MUL(frob_pows[6], imx2, t, p);
    GEN imx0p6 = FPM_FPXQ_MUL(frob_pows[6], imx0, t, p);

    GEN mx4p7  = FPM_FPXQ_MUL(frob_pows[7], mx4,  t, p);
    GEN mx0p7  = FPM_FPXQ_MUL(frob_pows[7], mx0,  t, p);
    GEN imx3p7 = FPM_FPXQ_MUL(frob_pows[7], imx3, t, p);
    GEN imx1p7 = FPM_FPXQ_MUL(frob_pows[7], imx1, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx4, mx11,  t, p);
    y0 = FPXQ_MUL(y0, mx6,    t, p);
    y0 = FPXQ_MUL(y0, mx2,    t, p);
    y0 = FPXQ_MUL(y0, mx9,    t, p);
    y0 = FPXQ_MUL(y0, mx1,    t, p);
    y0 = FPXQ_MUL(y0, imx3,   t, p);
    y0 = FPXQ_MUL(y0, mx8p1,  t, p);
    y0 = FPXQ_MUL(y0, mx5p1,  t, p);
    y0 = FPXQ_MUL(y0, mx10p1, t, p);
    y0 = FPXQ_MUL(y0, mx1p1,  t, p);
    y0 = FPXQ_MUL(y0, mx7p2,  t, p);
    y0 = FPXQ_MUL(y0, mx4p2,  t, p);
    y0 = FPXQ_MUL(y0, mx9p2,  t, p);
    y0 = FPXQ_MUL(y0, mx0p2,  t, p);
    y0 = FPXQ_MUL(y0, mx8p3,  t, p);
    y0 = FPXQ_MUL(y0, mx6p3,  t, p);
    y0 = FPXQ_MUL(y0, imx2p3, t, p);
    y0 = FPXQ_MUL(y0, imx0p3, t, p);
    y0 = FPXQ_MUL(y0, mx7p4,  t, p);
    y0 = FPXQ_MUL(y0, mx3p4,  t, p);
    y0 = FPXQ_MUL(y0, mx5p4,  t, p);
    y0 = FPXQ_MUL(y0, mx0p4,  t, p);
    y0 = FPXQ_MUL(y0, imx2p4, t, p);
    y0 = FPXQ_MUL(y0, imx1p4, t, p);
    y0 = FPXQ_MUL(y0, mx4p5,  t, p);
    y0 = FPXQ_MUL(y0, mx6p5,  t, p);
    y0 = FPXQ_MUL(y0, imx3p5, t, p);
    y0 = FPXQ_MUL(y0, imx1p5, t, p);
    y0 = FPXQ_MUL(y0, mx5p6,  t, p);
    y0 = FPXQ_MUL(y0, mx3p6,  t, p);
    y0 = FPXQ_MUL(y0, imx2p6, t, p);
    y0 = FPXQ_MUL(y0, imx0p6, t, p);
    y0 = FPXQ_MUL(y0, mx4p7,  t, p);
    y0 = FPXQ_MUL(y0, mx0p7,  t, p);
    y0 = FPXQ_MUL(y0, imx3p7, t, p);
    y0 = FPXQ_MUL(y0, imx1p7, t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx0, imx5,  t, p);
    y1 = FPXQ_MUL(y1, imx10,  t, p);
    y1 = FPXQ_MUL(y1, mx3p1,  t, p);
    y1 = FPXQ_MUL(y1, imx4p1, t, p);
    y1 = FPXQ_MUL(y1, imx2p1, t, p);
    y1 = FPXQ_MUL(y1, imx9p1, t, p);
    y1 = FPXQ_MUL(y1, mx2p2,  t, p);
    y1 = FPXQ_MUL(y1, imx8p2, t, p);
    y1 = FPXQ_MUL(y1, imx3p2, t, p);
    y1 = FPXQ_MUL(y1, imx1p2, t, p);
    y1 = FPXQ_MUL(y1, mx1p3,  t, p);
    y1 = FPXQ_MUL(y1, imx7p3, t, p);
    y1 = FPXQ_MUL(y1, imx6p4, t, p);
    y1 = FPXQ_MUL(y1, mx2p5,  t, p);
    y1 = FPXQ_MUL(y1, imx5p5, t, p);
    y1 = FPXQ_MUL(y1, mx1p6,  t, p);
    y1 = FPXQ_MUL(y1, imx4p6, t, p);
    
    GEN T0 = FPXQ_SQR(y1, t, p);
    GEN T1 = FPXQ_MUL(T0, y0, t, p);
    
    return T1;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_18_1333(pairing_data_t* const pdata, GEN unreduced)
{    
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0 = m;
    GEN mx1 = FPXQ_POW(mx0,  x, t, p);
    GEN mx2 = FPXQ_POW(mx1,  x, t, p);
    GEN mx3 = FPXQ_POW(mx2,  x, t, p);
    GEN mx4 = FPXQ_POW(mx3,  x, t, p);
    GEN mx5 = FPXQ_POW(mx4,  x, t, p);
    GEN mx6 = FPXQ_POW(mx5,  x, t, p);
    GEN mx7 = FPXQ_POW(mx6,  x, t, p);

    GEN imx0 = FPXQ_INV(mx0, t, p);
    GEN imx1 = FPXQ_INV(mx1, t, p);
    GEN imx2 = FPXQ_INV(mx2, t, p);
    GEN imx3 = FPXQ_INV(mx3, t, p);
    GEN imx4 = FPXQ_INV(mx4, t, p);
    GEN imx5 = FPXQ_INV(mx5, t, p);
    GEN imx6 = FPXQ_INV(mx6, t, p);
    GEN imx7 = FPXQ_INV(mx7, t, p);

    GEN mx6p1 = FPM_FPXQ_MUL(frob_pows[1], mx6, t, p);
    GEN mx5p1 = FPM_FPXQ_MUL(frob_pows[1], mx5, t, p);
    GEN mx4p1 = FPM_FPXQ_MUL(frob_pows[1], mx4, t, p);
    GEN mx3p1 = FPM_FPXQ_MUL(frob_pows[1], mx3, t, p);
    GEN mx2p1 = FPM_FPXQ_MUL(frob_pows[1], mx2, t, p);
    GEN mx1p1 = FPM_FPXQ_MUL(frob_pows[1], mx1, t, p);

    GEN imx5p2 = FPM_FPXQ_MUL(frob_pows[2], imx5, t, p);
    GEN imx4p2 = FPM_FPXQ_MUL(frob_pows[2], imx4, t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx2p2 = FPM_FPXQ_MUL(frob_pows[2], imx2, t, p);
    GEN imx1p2 = FPM_FPXQ_MUL(frob_pows[2], imx1, t, p);
    GEN imx0p2 = FPM_FPXQ_MUL(frob_pows[2], imx0, t, p);

    GEN mx0p3  = FPM_FPXQ_MUL(frob_pows[3], mx0,  t, p);
    GEN imx7p3 = FPM_FPXQ_MUL(frob_pows[3], imx7, t, p);
    GEN imx6p3 = FPM_FPXQ_MUL(frob_pows[3], imx6, t, p);
    GEN imx5p3 = FPM_FPXQ_MUL(frob_pows[3], imx5, t, p);
    GEN imx4p3 = FPM_FPXQ_MUL(frob_pows[3], imx4, t, p);
    GEN imx3p3 = FPM_FPXQ_MUL(frob_pows[3], imx3, t, p);
    GEN imx2p3 = FPM_FPXQ_MUL(frob_pows[3], imx2, t, p);

    GEN mx6p4 = FPM_FPXQ_MUL(frob_pows[4], mx6, t, p);
    GEN mx5p4 = FPM_FPXQ_MUL(frob_pows[4], mx5, t, p);
    GEN mx4p4 = FPM_FPXQ_MUL(frob_pows[4], mx4, t, p);
    GEN mx3p4 = FPM_FPXQ_MUL(frob_pows[4], mx3, t, p);
    GEN mx2p4 = FPM_FPXQ_MUL(frob_pows[4], mx2, t, p);
    GEN mx1p4 = FPM_FPXQ_MUL(frob_pows[4], mx1, t, p);

    GEN mx2p5 = FPM_FPXQ_MUL(frob_pows[5], mx2, t, p);
    GEN mx1p5 = FPM_FPXQ_MUL(frob_pows[5], mx1, t, p);
    GEN mx0p5 = FPM_FPXQ_MUL(frob_pows[5], mx0, t, p);
    
    GEN y0 = FPXQ_MUL(mx0, imx7, t, p);
    GEN y1 = imx7p3;
    GEN y2 = mx6p4;
    GEN y3 = mx6p1;
    GEN y4 = imx6;
    GEN y5 = imx5;
    GEN y6 = imx6p3;
    GEN y7 = FPXQ_MUL(imx5p3, mx5p4, t, p);

    GEN y8;
    y8 = FPXQ_MUL(imx5p2, mx4p4, t, p);
    y8 = FPXQ_MUL(y8, mx2p5, t, p);

    GEN y9  = mx0p3;
    GEN y10 = imx4;
    GEN y11 = mx5p1;
    GEN y12 = imx4p3;
    GEN y13 = mx4p1;
    GEN y14 = mx3p4;
    GEN y15 = FPXQ_MUL(imx4p2, mx1p5, t, p);
    GEN y16 = mx3p1;
    GEN y17 = imx3;
    GEN y18 = FPXQ_MUL(imx3p2, mx0p5, t, p);
    GEN y19 = imx2;
    GEN y20 = imx3p3;
    GEN y21 = mx2p4;
    GEN y22 = imx2p3;
    GEN y23 = mx1p4;
    GEN y24 = imx2p2;
    GEN y25 = mx2p1;
    GEN y26 = mx1p1;
    GEN y27 = imx1p2;
    GEN y28 = imx0p2;    
    
    GEN T0  = FPXQ_MUL(y28, y27, t, p);
    GEN T1  = FPXQ_MUL(T0, y24, t, p);
    GEN T2  = FPXQ_SQR(T0, t, p);
    GEN T3  = FPXQ_MUL(T2, y28, t, p);
    GEN T4  = FPXQ_MUL(T3, y26, t, p);
    GEN T5  = FPXQ_MUL(y26, y8, t, p);
    GEN T6  = FPXQ_SQR(T4, t, p);
    GEN T7  = FPXQ_MUL(T6, T1, t, p);
    GEN T8  = FPXQ_MUL(y25, y15, t, p);
    GEN T9  = FPXQ_SQR(y25, t, p);
    GEN T10 = FPXQ_MUL(T9, y21, t, p);
    GEN T11 = FPXQ_MUL(T7, y23, t, p);
    GEN T12 = FPXQ_MUL(T7, y27, t, p);
    GEN T13 = FPXQ_SQR(T11, t, p);
    GEN T14 = FPXQ_MUL(T10, y22, t, p);
    GEN T15 = FPXQ_MUL(T14, y19, t, p);
    GEN T16 = FPXQ_MUL(T14, T12, t, p);
    GEN T17 = FPXQ_MUL(T15, y20, t, p);
    GEN T18 = FPXQ_MUL(T17, T13, t, p);
    GEN T19 = FPXQ_MUL(T18, y18, t, p);
    GEN T20 = FPXQ_MUL(T18, T5, t, p);
    GEN T21 = FPXQ_MUL(T20, T19, t, p);
    GEN T22 = FPXQ_MUL(T19, y17, t, p);
    GEN T23 = FPXQ_MUL(y17, y6, t, p);
    GEN T24 = FPXQ_MUL(T22, y16, t, p);
    GEN T25 = FPXQ_MUL(T22, y5, t, p);
    GEN T26 = FPXQ_MUL(T24, T8, t, p);
    GEN T27 = FPXQ_MUL(T24, y22, t, p);
    GEN T28 = FPXQ_MUL(T26, T16, t, p);
    GEN T29 = FPXQ_MUL(T26, y13, t, p);
    GEN T30 = FPXQ_MUL(T28, y14, t, p);
    GEN T31 = FPXQ_MUL(T28, y7, t, p);
    GEN T32 = FPXQ_MUL(T30, T29, t, p);
    GEN T33 = FPXQ_MUL(T30, y3, t, p);
    GEN T34 = FPXQ_MUL(T32, y11, t, p);
    GEN T35 = FPXQ_MUL(T32, T27, t, p);
    GEN T36 = FPXQ_MUL(y12, y10, t, p);
    GEN T37 = FPXQ_MUL(T23, y12, t, p);
    GEN T38 = FPXQ_MUL(T36, T34, t, p);
    GEN T39 = FPXQ_MUL(T38, y9, t, p);
    GEN T40 = FPXQ_MUL(T38, T34, t, p);
    GEN T41 = FPXQ_MUL(T39, T21, t, p);
    GEN T42 = FPXQ_MUL(T39, y1, t, p);
    GEN T43 = FPXQ_MUL(T41, T17, t, p);
    GEN T44 = FPXQ_MUL(T41, y2, t, p);
    GEN T45 = FPXQ_MUL(T43, T31, t, p);
    GEN T46 = FPXQ_MUL(T44, T43, t, p);
    GEN T47 = FPXQ_MUL(T45, T35, t, p);
    GEN T48 = FPXQ_MUL(T46, T45, t, p);
    GEN T49 = FPXQ_MUL(T47, T37, t, p);
    GEN T50 = FPXQ_MUL(T47, y0, t, p);
    GEN T51 = FPXQ_MUL(T49, T25, t, p);
    GEN T52 = FPXQ_MUL(T51, y20, t, p);
    GEN T53 = FPXQ_MUL(T51, T42, t, p);
    GEN T54 = FPXQ_MUL(T52, y4, t, p);
    GEN T55 = FPXQ_MUL(T54, T33, t, p);
    GEN T56 = FPXQ_MUL(T54, T52, t, p);
    GEN T57 = FPXQ_SQR(T55, t, p);
    GEN T58 = FPXQ_MUL(T57, T48, t, p);
    GEN T59 = FPXQ_MUL(T58, T40, t, p);
    GEN T60 = FPXQ_MUL(T56, T40, t, p);
    GEN T61 = FPXQ_MUL(T59, T53, t, p);
    GEN T62 = FPXQ_MUL(T61, T49, t, p);
    GEN T63 = FPXQ_MUL(T61, T60, t, p);
    GEN T64 = FPXQ_MUL(T62, T50, t, p);
    GEN T65 = FPXQ_MUL(T63, T62, t, p);
    GEN T66 = FPXQ_MUL(T64, T59, t, p);
    GEN T67 = FPXQ_MUL(T65, T64, t, p);
    GEN T68 = FPXQ_SQR(T66, t, p);
    GEN T69 = FPXQ_MUL(T68, T67, t, p);
    
    OUT();
    return T69;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_19_1111(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);
    GEN mx12 = FPXQ_POW(mx11, x, t, p);
    GEN mx13 = FPXQ_POW(mx12, x, t, p);
    GEN mx14 = FPXQ_POW(mx13, x, t, p);
    GEN mx15 = FPXQ_POW(mx14, x, t, p);
    GEN mx16 = FPXQ_POW(mx15, x, t, p);
    GEN mx17 = FPXQ_POW(mx16, x, t, p);
    GEN mx18 = FPXQ_POW(mx17, x, t, p);
    GEN mx19 = FPXQ_POW(mx18, x, t, p);
    GEN mx20 = FPXQ_POW(mx19, x, t, p);
    GEN mx21 = FPXQ_POW(mx20, x, t, p);
    GEN mx22 = FPXQ_POW(mx21, x, t, p);
    GEN mx23 = FPXQ_POW(mx22, x, t, p);
    GEN mx24 = FPXQ_POW(mx23, x, t, p);
    GEN mx25 = FPXQ_POW(mx24, x, t, p);
    GEN mx26 = FPXQ_POW(mx25, x, t, p);
    GEN mx27 = FPXQ_POW(mx26, x, t, p);
    GEN mx28 = FPXQ_POW(mx27, x, t, p);
    GEN mx29 = FPXQ_POW(mx28, x, t, p);
    GEN mx30 = FPXQ_POW(mx29, x, t, p);
    GEN mx31 = FPXQ_POW(mx30, x, t, p);
    GEN mx32 = FPXQ_POW(mx31, x, t, p);
    GEN mx33 = FPXQ_POW(mx32, x, t, p);
    GEN mx34 = FPXQ_POW(mx33, x, t, p);
    GEN mx35 = FPXQ_POW(mx34, x, t, p);
    GEN mx36 = FPXQ_POW(mx35, x, t, p);
    GEN mx37 = FPXQ_POW(mx36, x, t, p);
    GEN mx38 = FPXQ_POW(mx37, x, t, p);
    GEN mx39 = FPXQ_POW(mx38, x, t, p);

    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);
    GEN imx12 = FPXQ_INV(mx12, t, p);
    GEN imx13 = FPXQ_INV(mx13, t, p);
    GEN imx14 = FPXQ_INV(mx14, t, p);
    GEN imx15 = FPXQ_INV(mx15, t, p);
    GEN imx16 = FPXQ_INV(mx16, t, p);
    GEN imx18 = FPXQ_INV(mx18, t, p);
    GEN imx19 = FPXQ_INV(mx19, t, p);
    GEN imx20 = FPXQ_INV(mx20, t, p);
    GEN imx22 = FPXQ_INV(mx22, t, p);
    GEN imx23 = FPXQ_INV(mx23, t, p);
    GEN imx24 = FPXQ_INV(mx24, t, p);
    GEN imx25 = FPXQ_INV(mx25, t, p);
    GEN imx26 = FPXQ_INV(mx26, t, p);
    GEN imx27 = FPXQ_INV(mx27, t, p);
    GEN imx28 = FPXQ_INV(mx28, t, p);
    GEN imx29 = FPXQ_INV(mx29, t, p);
    GEN imx30 = FPXQ_INV(mx30, t, p);
    GEN imx31 = FPXQ_INV(mx31, t, p);
    GEN imx32 = FPXQ_INV(mx32, t, p);
    GEN imx33 = FPXQ_INV(mx33, t, p);
    GEN imx34 = FPXQ_INV(mx34, t, p);
    GEN imx35 = FPXQ_INV(mx35, t, p);
    GEN imx36 = FPXQ_INV(mx36, t, p);
    GEN imx37 = FPXQ_INV(mx37, t, p);
    GEN imx38 = FPXQ_INV(mx38, t, p);

    GEN mx21p1  = FPM_FPXQ_MUL(frob_pows[1], mx21,  t, p);
    GEN mx18p1  = FPM_FPXQ_MUL(frob_pows[1], mx18,  t, p);
    GEN mx39p1  = FPM_FPXQ_MUL(frob_pows[1], mx39,  t, p);
    GEN mx2p1   = FPM_FPXQ_MUL(frob_pows[1], mx2,   t, p);
    GEN mx37p1  = FPM_FPXQ_MUL(frob_pows[1], mx37,  t, p);
    GEN mx0p1   = FPM_FPXQ_MUL(frob_pows[1], mx0,   t, p);
    GEN imx36p1 = FPM_FPXQ_MUL(frob_pows[1], imx36, t, p);
    GEN imx20p1 = FPM_FPXQ_MUL(frob_pows[1], imx20, t, p);
    GEN imx38p1 = FPM_FPXQ_MUL(frob_pows[1], imx38, t, p);
    GEN imx19p1 = FPM_FPXQ_MUL(frob_pows[1], imx19, t, p);
    GEN imx22p1 = FPM_FPXQ_MUL(frob_pows[1], imx22, t, p);
    GEN imx1p1  = FPM_FPXQ_MUL(frob_pows[1], imx1,  t, p);

    GEN mx36p2  = FPM_FPXQ_MUL(frob_pows[2], mx36,  t, p);
    GEN mx21p2  = FPM_FPXQ_MUL(frob_pows[2], mx21,  t, p);
    GEN mx18p2  = FPM_FPXQ_MUL(frob_pows[2], mx18,  t, p);
    GEN mx16p2  = FPM_FPXQ_MUL(frob_pows[2], mx16,  t, p);
    GEN mx2p2   = FPM_FPXQ_MUL(frob_pows[2], mx2,   t, p);
    GEN mx17p2  = FPM_FPXQ_MUL(frob_pows[2], mx17,  t, p);
    GEN mx0p2   = FPM_FPXQ_MUL(frob_pows[2], mx0,   t, p);
    GEN imx35p2 = FPM_FPXQ_MUL(frob_pows[2], imx35, t, p);
    GEN imx20p2 = FPM_FPXQ_MUL(frob_pows[2], imx20, t, p);
    GEN imx19p2 = FPM_FPXQ_MUL(frob_pows[2], imx19, t, p);
    GEN imx22p2 = FPM_FPXQ_MUL(frob_pows[2], imx22, t, p);
    GEN imx37p2 = FPM_FPXQ_MUL(frob_pows[2], imx37, t, p);
    GEN imx1p2  = FPM_FPXQ_MUL(frob_pows[2], imx1,  t, p);

    GEN mx36p3  = FPM_FPXQ_MUL(frob_pows[3], mx36,  t, p);
    GEN mx21p3  = FPM_FPXQ_MUL(frob_pows[3], mx21,  t, p);
    GEN mx15p3  = FPM_FPXQ_MUL(frob_pows[3], mx15,  t, p);
    GEN mx2p3   = FPM_FPXQ_MUL(frob_pows[3], mx2,   t, p);
    GEN mx17p3  = FPM_FPXQ_MUL(frob_pows[3], mx17,  t, p);
    GEN mx0p3   = FPM_FPXQ_MUL(frob_pows[3], mx0,   t, p);
    GEN imx35p3 = FPM_FPXQ_MUL(frob_pows[3], imx35, t, p);
    GEN imx20p3 = FPM_FPXQ_MUL(frob_pows[3], imx20, t, p);
    GEN imx16p3 = FPM_FPXQ_MUL(frob_pows[3], imx16, t, p);
    GEN imx22p3 = FPM_FPXQ_MUL(frob_pows[3], imx22, t, p);
    GEN imx37p3 = FPM_FPXQ_MUL(frob_pows[3], imx37, t, p);
    GEN imx1p3  = FPM_FPXQ_MUL(frob_pows[3], imx1,  t, p);

    GEN mx35p4  = FPM_FPXQ_MUL(frob_pows[4], mx35,  t, p);
    GEN mx33p4  = FPM_FPXQ_MUL(frob_pows[4], mx33,  t, p);
    GEN mx21p4  = FPM_FPXQ_MUL(frob_pows[4], mx21,  t, p);
    GEN mx16p4  = FPM_FPXQ_MUL(frob_pows[4], mx16,  t, p);
    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,   t, p);
    GEN mx17p4  = FPM_FPXQ_MUL(frob_pows[4], mx17,  t, p);
    GEN mx0p4   = FPM_FPXQ_MUL(frob_pows[4], mx0,   t, p);
    GEN imx36p4 = FPM_FPXQ_MUL(frob_pows[4], imx36, t, p);
    GEN imx20p4 = FPM_FPXQ_MUL(frob_pows[4], imx20, t, p);
    GEN imx15p4 = FPM_FPXQ_MUL(frob_pows[4], imx15, t, p);
    GEN imx22p4 = FPM_FPXQ_MUL(frob_pows[4], imx22, t, p);
    GEN imx34p4 = FPM_FPXQ_MUL(frob_pows[4], imx34, t, p);
    GEN imx1p4  = FPM_FPXQ_MUL(frob_pows[4], imx1,  t, p);

    GEN mx32p5  = FPM_FPXQ_MUL(frob_pows[5], mx32,  t, p);
    GEN mx21p5  = FPM_FPXQ_MUL(frob_pows[5], mx21,  t, p);
    GEN mx16p5  = FPM_FPXQ_MUL(frob_pows[5], mx16,  t, p);
    GEN mx2p5   = FPM_FPXQ_MUL(frob_pows[5], mx2,   t, p);
    GEN mx34p5  = FPM_FPXQ_MUL(frob_pows[5], mx34,  t, p);
    GEN mx0p5   = FPM_FPXQ_MUL(frob_pows[5], mx0,   t, p);
    GEN imx33p5 = FPM_FPXQ_MUL(frob_pows[5], imx33, t, p);
    GEN imx20p5 = FPM_FPXQ_MUL(frob_pows[5], imx20, t, p);
    GEN imx15p5 = FPM_FPXQ_MUL(frob_pows[5], imx15, t, p);
    GEN imx13p5 = FPM_FPXQ_MUL(frob_pows[5], imx13, t, p);
    GEN imx14p5 = FPM_FPXQ_MUL(frob_pows[5], imx14, t, p);
    GEN imx22p5 = FPM_FPXQ_MUL(frob_pows[5], imx22, t, p);
    GEN imx1p5  = FPM_FPXQ_MUL(frob_pows[5], imx1,  t, p);

    GEN mx32p6  = FPM_FPXQ_MUL(frob_pows[6], mx32,  t, p);
    GEN mx21p6  = FPM_FPXQ_MUL(frob_pows[6], mx21,  t, p);
    GEN mx13p6  = FPM_FPXQ_MUL(frob_pows[6], mx13,  t, p);
    GEN mx2p6   = FPM_FPXQ_MUL(frob_pows[6], mx2,   t, p);
    GEN mx34p6  = FPM_FPXQ_MUL(frob_pows[6], mx34,  t, p);
    GEN mx0p6   = FPM_FPXQ_MUL(frob_pows[6], mx0,   t, p);
    GEN imx33p6 = FPM_FPXQ_MUL(frob_pows[6], imx33, t, p);
    GEN imx20p6 = FPM_FPXQ_MUL(frob_pows[6], imx20, t, p);
    GEN imx12p6 = FPM_FPXQ_MUL(frob_pows[6], imx12, t, p);
    GEN imx14p6 = FPM_FPXQ_MUL(frob_pows[6], imx14, t, p);
    GEN imx22p6 = FPM_FPXQ_MUL(frob_pows[6], imx22, t, p);
    GEN imx1p6  = FPM_FPXQ_MUL(frob_pows[6], imx1,  t, p);

    GEN mx33p7  = FPM_FPXQ_MUL(frob_pows[7], mx33,  t, p);
    GEN mx21p7  = FPM_FPXQ_MUL(frob_pows[7], mx21,  t, p);
    GEN mx12p7  = FPM_FPXQ_MUL(frob_pows[7], mx12,  t, p);
    GEN mx2p7   = FPM_FPXQ_MUL(frob_pows[7], mx2,   t, p);
    GEN mx31p7  = FPM_FPXQ_MUL(frob_pows[7], mx31,  t, p);
    GEN mx0p7   = FPM_FPXQ_MUL(frob_pows[7], mx0,   t, p);
    GEN imx32p7 = FPM_FPXQ_MUL(frob_pows[7], imx32, t, p);
    GEN imx30p7 = FPM_FPXQ_MUL(frob_pows[7], imx30, t, p);
    GEN imx20p7 = FPM_FPXQ_MUL(frob_pows[7], imx20, t, p);
    GEN imx13p7 = FPM_FPXQ_MUL(frob_pows[7], imx13, t, p);
    GEN imx14p7 = FPM_FPXQ_MUL(frob_pows[7], imx14, t, p);
    GEN imx22p7 = FPM_FPXQ_MUL(frob_pows[7], imx22, t, p);
    GEN imx1p7  = FPM_FPXQ_MUL(frob_pows[7], imx1,  t, p);

    GEN mx30p8  = FPM_FPXQ_MUL(frob_pows[8], mx30,  t, p);
    GEN mx21p8  = FPM_FPXQ_MUL(frob_pows[8], mx21,  t, p);
    GEN mx12p8  = FPM_FPXQ_MUL(frob_pows[8], mx12,  t, p);
    GEN mx10p8  = FPM_FPXQ_MUL(frob_pows[8], mx10,  t, p);
    GEN mx2p8   = FPM_FPXQ_MUL(frob_pows[8], mx2,   t, p);
    GEN mx11p8  = FPM_FPXQ_MUL(frob_pows[8], mx11,  t, p);
    GEN mx0p8   = FPM_FPXQ_MUL(frob_pows[8], mx0,   t, p);
    GEN imx29p8 = FPM_FPXQ_MUL(frob_pows[8], imx29, t, p);
    GEN imx20p8 = FPM_FPXQ_MUL(frob_pows[8], imx20, t, p);
    GEN imx13p8 = FPM_FPXQ_MUL(frob_pows[8], imx13, t, p);
    GEN imx22p8 = FPM_FPXQ_MUL(frob_pows[8], imx22, t, p);
    GEN imx31p8 = FPM_FPXQ_MUL(frob_pows[8], imx31, t, p);
    GEN imx1p8  = FPM_FPXQ_MUL(frob_pows[8], imx1,  t, p);

    GEN mx30p9  = FPM_FPXQ_MUL(frob_pows[9], mx30,  t, p);
    GEN mx21p9  = FPM_FPXQ_MUL(frob_pows[9], mx21,  t, p);
    GEN mx9p9   = FPM_FPXQ_MUL(frob_pows[9], mx9,   t, p);
    GEN mx2p9   = FPM_FPXQ_MUL(frob_pows[9], mx2,   t, p);
    GEN mx11p9  = FPM_FPXQ_MUL(frob_pows[9], mx11,  t, p);
    GEN mx0p9   = FPM_FPXQ_MUL(frob_pows[9], mx0,   t, p);
    GEN imx29p9 = FPM_FPXQ_MUL(frob_pows[9], imx29, t, p);
    GEN imx20p9 = FPM_FPXQ_MUL(frob_pows[9], imx20, t, p);
    GEN imx10p9 = FPM_FPXQ_MUL(frob_pows[9], imx10, t, p);
    GEN imx22p9 = FPM_FPXQ_MUL(frob_pows[9], imx22, t, p);
    GEN imx31p9 = FPM_FPXQ_MUL(frob_pows[9], imx31, t, p);
    GEN imx1p9  = FPM_FPXQ_MUL(frob_pows[9], imx1,  t, p);

    GEN mx29p10  = FPM_FPXQ_MUL(frob_pows[10], mx29,  t, p);
    GEN mx27p10  = FPM_FPXQ_MUL(frob_pows[10], mx27,  t, p);
    GEN mx21p10  = FPM_FPXQ_MUL(frob_pows[10], mx21,  t, p);
    GEN mx10p10  = FPM_FPXQ_MUL(frob_pows[10], mx10,  t, p);
    GEN mx2p10   = FPM_FPXQ_MUL(frob_pows[10], mx2,   t, p);
    GEN mx11p10  = FPM_FPXQ_MUL(frob_pows[10], mx11,  t, p);
    GEN mx0p10   = FPM_FPXQ_MUL(frob_pows[10], mx0,   t, p);
    GEN imx30p10 = FPM_FPXQ_MUL(frob_pows[10], imx30, t, p);
    GEN imx20p10 = FPM_FPXQ_MUL(frob_pows[10], imx20, t, p);
    GEN imx9p10  = FPM_FPXQ_MUL(frob_pows[10], imx9,  t, p);
    GEN imx22p10 = FPM_FPXQ_MUL(frob_pows[10], imx22, t, p);
    GEN imx28p10 = FPM_FPXQ_MUL(frob_pows[10], imx28, t, p);
    GEN imx1p10  = FPM_FPXQ_MUL(frob_pows[10], imx1,  t, p);

    GEN mx26p11  = FPM_FPXQ_MUL(frob_pows[11], mx26,  t, p);
    GEN mx21p11  = FPM_FPXQ_MUL(frob_pows[11], mx21,  t, p);
    GEN mx10p11  = FPM_FPXQ_MUL(frob_pows[11], mx10,  t, p);
    GEN mx2p11   = FPM_FPXQ_MUL(frob_pows[11], mx2,   t, p);
    GEN mx28p11  = FPM_FPXQ_MUL(frob_pows[11], mx28,  t, p);
    GEN mx0p11   = FPM_FPXQ_MUL(frob_pows[11], mx0,   t, p);
    GEN imx27p11 = FPM_FPXQ_MUL(frob_pows[11], imx27, t, p);
    GEN imx20p11 = FPM_FPXQ_MUL(frob_pows[11], imx20, t, p);
    GEN imx9p11  = FPM_FPXQ_MUL(frob_pows[11], imx9,  t, p);
    GEN imx7p11  = FPM_FPXQ_MUL(frob_pows[11], imx7,  t, p);
    GEN imx8p11  = FPM_FPXQ_MUL(frob_pows[11], imx8,  t, p);
    GEN imx22p11 = FPM_FPXQ_MUL(frob_pows[11], imx22, t, p);
    GEN imx1p11  = FPM_FPXQ_MUL(frob_pows[11], imx1,  t, p);

    GEN mx26p12  = FPM_FPXQ_MUL(frob_pows[12], mx26,  t, p);
    GEN mx21p12  = FPM_FPXQ_MUL(frob_pows[12], mx21,  t, p);
    GEN mx7p12   = FPM_FPXQ_MUL(frob_pows[12], mx7,   t, p);
    GEN mx2p12   = FPM_FPXQ_MUL(frob_pows[12], mx2,   t, p);
    GEN mx28p12  = FPM_FPXQ_MUL(frob_pows[12], mx28,  t, p);
    GEN mx0p12   = FPM_FPXQ_MUL(frob_pows[12], mx0,   t, p);
    GEN imx27p12 = FPM_FPXQ_MUL(frob_pows[12], imx27, t, p);
    GEN imx20p12 = FPM_FPXQ_MUL(frob_pows[12], imx20, t, p);
    GEN imx6p12  = FPM_FPXQ_MUL(frob_pows[12], imx6,  t, p);
    GEN imx8p12  = FPM_FPXQ_MUL(frob_pows[12], imx8,  t, p);
    GEN imx22p12 = FPM_FPXQ_MUL(frob_pows[12], imx22, t, p);
    GEN imx1p12  = FPM_FPXQ_MUL(frob_pows[12], imx1,  t, p);

    GEN mx27p13  = FPM_FPXQ_MUL(frob_pows[13], mx27,  t, p);
    GEN mx21p13  = FPM_FPXQ_MUL(frob_pows[13], mx21,  t, p);
    GEN mx6p13   = FPM_FPXQ_MUL(frob_pows[13], mx6,   t, p);
    GEN mx2p13   = FPM_FPXQ_MUL(frob_pows[13], mx2,   t, p);
    GEN mx25p13  = FPM_FPXQ_MUL(frob_pows[13], mx25,  t, p);
    GEN mx0p13   = FPM_FPXQ_MUL(frob_pows[13], mx0,   t, p);
    GEN imx26p13 = FPM_FPXQ_MUL(frob_pows[13], imx26, t, p);
    GEN imx24p13 = FPM_FPXQ_MUL(frob_pows[13], imx24, t, p);
    GEN imx20p13 = FPM_FPXQ_MUL(frob_pows[13], imx20, t, p);
    GEN imx7p13  = FPM_FPXQ_MUL(frob_pows[13], imx7,  t, p);
    GEN imx8p13  = FPM_FPXQ_MUL(frob_pows[13], imx8,  t, p);
    GEN imx22p13 = FPM_FPXQ_MUL(frob_pows[13], imx22, t, p);
    GEN imx1p13  = FPM_FPXQ_MUL(frob_pows[13], imx1,  t, p);

    GEN mx24p14  = FPM_FPXQ_MUL(frob_pows[14], mx24,  t, p);
    GEN mx21p14  = FPM_FPXQ_MUL(frob_pows[14], mx21,  t, p);
    GEN mx6p14   = FPM_FPXQ_MUL(frob_pows[14], mx6,   t, p);
    GEN mx4p14   = FPM_FPXQ_MUL(frob_pows[14], mx4,   t, p);
    GEN mx2p14   = FPM_FPXQ_MUL(frob_pows[14], mx2,   t, p);
    GEN mx5p14   = FPM_FPXQ_MUL(frob_pows[14], mx5,   t, p);
    GEN mx0p14   = FPM_FPXQ_MUL(frob_pows[14], mx0,   t, p);
    GEN imx23p14 = FPM_FPXQ_MUL(frob_pows[14], imx23, t, p);
    GEN imx20p14 = FPM_FPXQ_MUL(frob_pows[14], imx20, t, p);
    GEN imx7p14  = FPM_FPXQ_MUL(frob_pows[14], imx7,  t, p);
    GEN imx22p14 = FPM_FPXQ_MUL(frob_pows[14], imx22, t, p);
    GEN imx25p14 = FPM_FPXQ_MUL(frob_pows[14], imx25, t, p);
    GEN imx1p14  = FPM_FPXQ_MUL(frob_pows[14], imx1,  t, p);

    GEN mx24p15  = FPM_FPXQ_MUL(frob_pows[15], mx24,  t, p);
    GEN mx21p15  = FPM_FPXQ_MUL(frob_pows[15], mx21,  t, p);
    GEN mx3p15   = FPM_FPXQ_MUL(frob_pows[15], mx3,   t, p);
    GEN mx2p15   = FPM_FPXQ_MUL(frob_pows[15], mx2,   t, p);
    GEN mx5p15   = FPM_FPXQ_MUL(frob_pows[15], mx5,   t, p);
    GEN mx0p15   = FPM_FPXQ_MUL(frob_pows[15], mx0,   t, p);
    GEN imx23p15 = FPM_FPXQ_MUL(frob_pows[15], imx23, t, p);
    GEN imx20p15 = FPM_FPXQ_MUL(frob_pows[15], imx20, t, p);
    GEN imx4p15  = FPM_FPXQ_MUL(frob_pows[15], imx4,  t, p);
    GEN imx22p15 = FPM_FPXQ_MUL(frob_pows[15], imx22, t, p);
    GEN imx25p15 = FPM_FPXQ_MUL(frob_pows[15], imx25, t, p);
    GEN imx1p15  = FPM_FPXQ_MUL(frob_pows[15], imx1,  t, p);

    GEN mx23p16  = FPM_FPXQ_MUL(frob_pows[16], mx23,  t, p);
    GEN mx21p16  = FPM_FPXQ_MUL(frob_pows[16], mx21,  t, p);
    GEN mx4p16   = FPM_FPXQ_MUL(frob_pows[16], mx4,   t, p);
    GEN mx2p16   = FPM_FPXQ_MUL(frob_pows[16], mx2,   t, p);
    GEN mx5p16   = FPM_FPXQ_MUL(frob_pows[16], mx5,   t, p);
    GEN mx0p16   = FPM_FPXQ_MUL(frob_pows[16], mx0,   t, p);
    GEN imx24p16 = FPM_FPXQ_MUL(frob_pows[16], imx24, t, p);
    GEN imx20p16 = FPM_FPXQ_MUL(frob_pows[16], imx20, t, p);
    GEN imx3p16  = FPM_FPXQ_MUL(frob_pows[16], imx3,  t, p);
    GEN imx22p16 = FPM_FPXQ_MUL(frob_pows[16], imx22, t, p);
    GEN imx1p16  = FPM_FPXQ_MUL(frob_pows[16], imx1,  t, p);

    GEN mx4p17  = FPM_FPXQ_MUL(frob_pows[17], mx4,  t, p);
    GEN mx2p17  = FPM_FPXQ_MUL(frob_pows[17], mx2,  t, p);
    GEN mx0p17  = FPM_FPXQ_MUL(frob_pows[17], mx0,  t, p);
    GEN imx3p17 = FPM_FPXQ_MUL(frob_pows[17], imx3, t, p);
    GEN imx1p17 = FPM_FPXQ_MUL(frob_pows[17], imx1, t, p);

    GEN y0;
    y0 = FPXQ_MUL(imx18, imx22, t, p);
    y0 = FPXQ_MUL(y0, mx21p1,   t, p);
    y0 = FPXQ_MUL(y0, mx39p1,   t, p);
    y0 = FPXQ_MUL(y0, imx36p1,  t, p);
    y0 = FPXQ_MUL(y0, imx22p1,  t, p);
    y0 = FPXQ_MUL(y0, imx1p1,   t, p);
    y0 = FPXQ_MUL(y0, mx36p2,   t, p);
    y0 = FPXQ_MUL(y0, mx21p2,   t, p);
    y0 = FPXQ_MUL(y0, mx16p2,   t, p);
    y0 = FPXQ_MUL(y0, mx17p2,   t, p);
    y0 = FPXQ_MUL(y0, mx0p2,    t, p);
    y0 = FPXQ_MUL(y0, imx35p2,  t, p);
    y0 = FPXQ_MUL(y0, imx20p2,  t, p);
    y0 = FPXQ_MUL(y0, imx19p2,  t, p);
    y0 = FPXQ_MUL(y0, imx22p2,  t, p);
    y0 = FPXQ_MUL(y0, imx37p2,  t, p);
    y0 = FPXQ_MUL(y0, imx1p2,   t, p);
    y0 = FPXQ_MUL(y0, mx36p3,   t, p);
    y0 = FPXQ_MUL(y0, mx21p3,   t, p);
    y0 = FPXQ_MUL(y0, mx15p3,   t, p);
    y0 = FPXQ_MUL(y0, mx0p3,    t, p);
    y0 = FPXQ_MUL(y0, imx35p3,  t, p);
    y0 = FPXQ_MUL(y0, imx20p3,  t, p);
    y0 = FPXQ_MUL(y0, imx16p3,  t, p);
    y0 = FPXQ_MUL(y0, imx22p3,  t, p);
    y0 = FPXQ_MUL(y0, imx37p3,  t, p);
    y0 = FPXQ_MUL(y0, imx1p3,   t, p);
    y0 = FPXQ_MUL(y0, mx33p4,   t, p);
    y0 = FPXQ_MUL(y0, mx21p4,   t, p);
    y0 = FPXQ_MUL(y0, mx17p4,   t, p);
    y0 = FPXQ_MUL(y0, mx0p4,    t, p);
    y0 = FPXQ_MUL(y0, imx36p4,  t, p);
    y0 = FPXQ_MUL(y0, imx20p4,  t, p);
    y0 = FPXQ_MUL(y0, imx22p4,  t, p);
    y0 = FPXQ_MUL(y0, imx1p4,   t, p);
    y0 = FPXQ_MUL(y0, mx32p5,   t, p);
    y0 = FPXQ_MUL(y0, mx21p5,   t, p);
    y0 = FPXQ_MUL(y0, mx16p5,   t, p);
    y0 = FPXQ_MUL(y0, mx34p5,   t, p);
    y0 = FPXQ_MUL(y0, mx0p5,    t, p);
    y0 = FPXQ_MUL(y0, imx33p5,  t, p);
    y0 = FPXQ_MUL(y0, imx20p5,  t, p);
    y0 = FPXQ_MUL(y0, imx13p5,  t, p);
    y0 = FPXQ_MUL(y0, imx14p5,  t, p);
    y0 = FPXQ_MUL(y0, imx22p5,  t, p);
    y0 = FPXQ_MUL(y0, imx1p5,   t, p);
    y0 = FPXQ_MUL(y0, mx32p6,   t, p);
    y0 = FPXQ_MUL(y0, mx21p6,   t, p);
    y0 = FPXQ_MUL(y0, mx13p6,   t, p);
    y0 = FPXQ_MUL(y0, mx34p6,   t, p);
    y0 = FPXQ_MUL(y0, mx0p6,    t, p);
    y0 = FPXQ_MUL(y0, imx33p6,  t, p);
    y0 = FPXQ_MUL(y0, imx20p6,  t, p);
    y0 = FPXQ_MUL(y0, imx12p6,  t, p);
    y0 = FPXQ_MUL(y0, imx22p6,  t, p);
    y0 = FPXQ_MUL(y0, imx1p6,   t, p);
    y0 = FPXQ_MUL(y0, mx33p7,   t, p);
    y0 = FPXQ_MUL(y0, mx21p7,   t, p);
    y0 = FPXQ_MUL(y0, mx0p7,    t, p);
    y0 = FPXQ_MUL(y0, imx30p7,  t, p);
    y0 = FPXQ_MUL(y0, imx20p7,  t, p);
    y0 = FPXQ_MUL(y0, imx14p7,  t, p);
    y0 = FPXQ_MUL(y0, imx22p7,  t, p);
    y0 = FPXQ_MUL(y0, imx1p7,   t, p);
    y0 = FPXQ_MUL(y0, mx30p8,   t, p);
    y0 = FPXQ_MUL(y0, mx21p8,   t, p);
    y0 = FPXQ_MUL(y0, mx10p8,   t, p);
    y0 = FPXQ_MUL(y0, mx11p8,   t, p);
    y0 = FPXQ_MUL(y0, mx0p8,    t, p);
    y0 = FPXQ_MUL(y0, imx29p8,  t, p);
    y0 = FPXQ_MUL(y0, imx20p8,  t, p);
    y0 = FPXQ_MUL(y0, imx13p8,  t, p);
    y0 = FPXQ_MUL(y0, imx22p8,  t, p);
    y0 = FPXQ_MUL(y0, imx31p8,  t, p);
    y0 = FPXQ_MUL(y0, imx1p8,   t, p);
    y0 = FPXQ_MUL(y0, mx30p9,   t, p);
    y0 = FPXQ_MUL(y0, mx21p9,   t, p);
    y0 = FPXQ_MUL(y0, mx9p9,    t, p);
    y0 = FPXQ_MUL(y0, mx0p9,    t, p);
    y0 = FPXQ_MUL(y0, imx29p9,  t, p);
    y0 = FPXQ_MUL(y0, imx20p9,  t, p);
    y0 = FPXQ_MUL(y0, imx10p9,  t, p);
    y0 = FPXQ_MUL(y0, imx22p9,  t, p);
    y0 = FPXQ_MUL(y0, imx31p9,  t, p);
    y0 = FPXQ_MUL(y0, imx1p9,   t, p);
    y0 = FPXQ_MUL(y0, mx27p10,  t, p);
    y0 = FPXQ_MUL(y0, mx21p10,  t, p);
    y0 = FPXQ_MUL(y0, mx11p10,  t, p);
    y0 = FPXQ_MUL(y0, mx0p10,   t, p);
    y0 = FPXQ_MUL(y0, imx30p10, t, p);
    y0 = FPXQ_MUL(y0, imx20p10, t, p);
    y0 = FPXQ_MUL(y0, imx22p10, t, p);
    y0 = FPXQ_MUL(y0, imx1p10,  t, p);
    y0 = FPXQ_MUL(y0, mx26p11,  t, p);
    y0 = FPXQ_MUL(y0, mx21p11,  t, p);
    y0 = FPXQ_MUL(y0, mx10p11,  t, p);
    y0 = FPXQ_MUL(y0, mx28p11,  t, p);
    y0 = FPXQ_MUL(y0, mx0p11,   t, p);
    y0 = FPXQ_MUL(y0, imx27p11, t, p);
    y0 = FPXQ_MUL(y0, imx20p11, t, p);
    y0 = FPXQ_MUL(y0, imx7p11,  t, p);
    y0 = FPXQ_MUL(y0, imx8p11,  t, p);
    y0 = FPXQ_MUL(y0, imx22p11, t, p);
    y0 = FPXQ_MUL(y0, imx1p11,  t, p);
    y0 = FPXQ_MUL(y0, mx26p12,  t, p);
    y0 = FPXQ_MUL(y0, mx21p12,  t, p);
    y0 = FPXQ_MUL(y0, mx7p12,   t, p);
    y0 = FPXQ_MUL(y0, mx28p12,  t, p);
    y0 = FPXQ_MUL(y0, mx0p12,   t, p);
    y0 = FPXQ_MUL(y0, imx27p12, t, p);
    y0 = FPXQ_MUL(y0, imx20p12, t, p);
    y0 = FPXQ_MUL(y0, imx6p12,  t, p);
    y0 = FPXQ_MUL(y0, imx22p12, t, p);
    y0 = FPXQ_MUL(y0, imx1p12,  t, p);
    y0 = FPXQ_MUL(y0, mx27p13,  t, p);
    y0 = FPXQ_MUL(y0, mx21p13,  t, p);
    y0 = FPXQ_MUL(y0, mx0p13,   t, p);
    y0 = FPXQ_MUL(y0, imx24p13, t, p);
    y0 = FPXQ_MUL(y0, imx20p13, t, p);
    y0 = FPXQ_MUL(y0, imx8p13,  t, p);
    y0 = FPXQ_MUL(y0, imx22p13, t, p);
    y0 = FPXQ_MUL(y0, imx1p13,  t, p);
    y0 = FPXQ_MUL(y0, mx24p14,  t, p);
    y0 = FPXQ_MUL(y0, mx21p14,  t, p);
    y0 = FPXQ_MUL(y0, mx4p14,   t, p);
    y0 = FPXQ_MUL(y0, mx5p14,   t, p);
    y0 = FPXQ_MUL(y0, mx0p14,   t, p);
    y0 = FPXQ_MUL(y0, imx23p14, t, p);
    y0 = FPXQ_MUL(y0, imx20p14, t, p);
    y0 = FPXQ_MUL(y0, imx7p14,  t, p);
    y0 = FPXQ_MUL(y0, imx22p14, t, p);
    y0 = FPXQ_MUL(y0, imx25p14, t, p);
    y0 = FPXQ_MUL(y0, imx1p14,  t, p);
    y0 = FPXQ_MUL(y0, mx24p15,  t, p);
    y0 = FPXQ_MUL(y0, mx21p15,  t, p);
    y0 = FPXQ_MUL(y0, mx3p15,   t, p);
    y0 = FPXQ_MUL(y0, mx0p15,   t, p);
    y0 = FPXQ_MUL(y0, imx23p15, t, p);
    y0 = FPXQ_MUL(y0, imx20p15, t, p);
    y0 = FPXQ_MUL(y0, imx4p15,  t, p);
    y0 = FPXQ_MUL(y0, imx22p15, t, p);
    y0 = FPXQ_MUL(y0, imx25p15, t, p);
    y0 = FPXQ_MUL(y0, imx1p15,  t, p);
    y0 = FPXQ_MUL(y0, mx5p16,   t, p);
    y0 = FPXQ_MUL(y0, mx0p16,   t, p);
    y0 = FPXQ_MUL(y0, imx24p16, t, p);
    y0 = FPXQ_MUL(y0, imx20p16, t, p);
    y0 = FPXQ_MUL(y0, imx1p16,  t, p);
    y0 = FPXQ_MUL(y0, mx4p17,   t, p);
    y0 = FPXQ_MUL(y0, mx0p17,   t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx21, mx19,   t, p);
    y1 = FPXQ_MUL(y1, mx18p1,   t, p);
    y1 = FPXQ_MUL(y1, mx37p1,   t, p);
    y1 = FPXQ_MUL(y1, imx20p1,  t, p);
    y1 = FPXQ_MUL(y1, imx38p1,  t, p);
    y1 = FPXQ_MUL(y1, imx19p1,  t, p);
    y1 = FPXQ_MUL(y1, mx18p2,   t, p);
    y1 = FPXQ_MUL(y1, mx35p4,   t, p);
    y1 = FPXQ_MUL(y1, mx16p4,   t, p);
    y1 = FPXQ_MUL(y1, imx15p4,  t, p);
    y1 = FPXQ_MUL(y1, imx34p4,  t, p);
    y1 = FPXQ_MUL(y1, imx15p5,  t, p);
    y1 = FPXQ_MUL(y1, mx12p7,   t, p);
    y1 = FPXQ_MUL(y1, mx31p7,   t, p);
    y1 = FPXQ_MUL(y1, imx32p7,  t, p);
    y1 = FPXQ_MUL(y1, imx13p7,  t, p);
    y1 = FPXQ_MUL(y1, mx12p8,   t, p);
    y1 = FPXQ_MUL(y1, mx29p10,  t, p);
    y1 = FPXQ_MUL(y1, mx10p10,  t, p);
    y1 = FPXQ_MUL(y1, imx9p10,  t, p);
    y1 = FPXQ_MUL(y1, imx28p10, t, p);
    y1 = FPXQ_MUL(y1, imx9p11,  t, p);
    y1 = FPXQ_MUL(y1, mx6p13,   t, p);
    y1 = FPXQ_MUL(y1, mx25p13,  t, p);
    y1 = FPXQ_MUL(y1, imx26p13, t, p);
    y1 = FPXQ_MUL(y1, imx7p13,  t, p);
    y1 = FPXQ_MUL(y1, mx6p14,   t, p);
    y1 = FPXQ_MUL(y1, mx23p16,  t, p);
    y1 = FPXQ_MUL(y1, mx21p16,  t, p);
    y1 = FPXQ_MUL(y1, mx4p16,   t, p);
    y1 = FPXQ_MUL(y1, imx3p16,  t, p);
    y1 = FPXQ_MUL(y1, imx3p17,  t, p);
    y1 = FPXQ_MUL(y1, imx1p17,  t, p);

    GEN y2;
    y2 = FPXQ_MUL(mx2, mx0,     t, p);
    y2 = FPXQ_MUL(y2, imx20,    t, p);
    y2 = FPXQ_MUL(y2, imx1,     t, p);
    y2 = FPXQ_MUL(y2, imx22p16, t, p);
    y2 = FPXQ_MUL(y2, mx2p17,   t, p);

    GEN y3;
    y3 = FPXQ_MUL(mx2p1, mx0p1, t, p);
    y3 = FPXQ_MUL(y3, mx2p2,    t, p);
    y3 = FPXQ_MUL(y3, mx2p3,    t, p);
    y3 = FPXQ_MUL(y3, mx17p3,   t, p);
    y3 = FPXQ_MUL(y3, mx2p4,    t, p);
    y3 = FPXQ_MUL(y3, mx2p5,    t, p);
    y3 = FPXQ_MUL(y3, mx2p6,    t, p);
    y3 = FPXQ_MUL(y3, imx14p6,  t, p);
    y3 = FPXQ_MUL(y3, mx2p7,    t, p);
    y3 = FPXQ_MUL(y3, mx2p8,    t, p);
    y3 = FPXQ_MUL(y3, mx2p9,    t, p);
    y3 = FPXQ_MUL(y3, mx11p9,   t, p);
    y3 = FPXQ_MUL(y3, mx2p10,   t, p);
    y3 = FPXQ_MUL(y3, mx2p11,   t, p);
    y3 = FPXQ_MUL(y3, mx2p12,   t, p);
    y3 = FPXQ_MUL(y3, imx8p12,  t, p);
    y3 = FPXQ_MUL(y3, mx2p13,   t, p);
    y3 = FPXQ_MUL(y3, mx2p14,   t, p);
    y3 = FPXQ_MUL(y3, mx2p15,   t, p);
    y3 = FPXQ_MUL(y3, mx5p15,   t, p);
    y3 = FPXQ_MUL(y3, mx2p16,   t, p);
    
    GEN T0 = FPXQ_MUL(y3, y2, t, p);
    GEN T1 = FPXQ_MUL(y3, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y1, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_SQR(T2, t, p);
    GEN T5 = FPXQ_MUL(T4, T3, t, p);
    
    return T5;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_25_1300(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;

    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);
    GEN mx12 = FPXQ_POW(mx11, x, t, p);
    GEN mx13 = FPXQ_POW(mx12, x, t, p);
    GEN mx14 = FPXQ_POW(mx13, x, t, p);
    GEN mx15 = FPXQ_POW(mx14, x, t, p);
    GEN mx16 = FPXQ_POW(mx15, x, t, p);
    GEN mx17 = FPXQ_POW(mx16, x, t, p);
    GEN mx18 = FPXQ_POW(mx17, x, t, p);
    GEN mx19 = FPXQ_POW(mx18, x, t, p);
    GEN mx20 = FPXQ_POW(mx19, x, t, p);
    GEN mx21 = FPXQ_POW(mx20, x, t, p);
    GEN mx22 = FPXQ_POW(mx21, x, t, p);
    GEN mx23 = FPXQ_POW(mx22, x, t, p);
    GEN mx24 = FPXQ_POW(mx23, x, t, p);
    GEN mx25 = FPXQ_POW(mx24, x, t, p);
    GEN mx26 = FPXQ_POW(mx25, x, t, p);
    GEN mx27 = FPXQ_POW(mx26, x, t, p);
    GEN mx28 = FPXQ_POW(mx27, x, t, p);
    GEN mx29 = FPXQ_POW(mx28, x, t, p);
    GEN mx30 = FPXQ_POW(mx29, x, t, p);
    GEN mx31 = FPXQ_POW(mx30, x, t, p);
    GEN mx32 = FPXQ_POW(mx31, x, t, p);
    GEN mx33 = FPXQ_POW(mx32, x, t, p);
    GEN mx34 = FPXQ_POW(mx33, x, t, p);
    GEN mx35 = FPXQ_POW(mx34, x, t, p);
    GEN mx36 = FPXQ_POW(mx35, x, t, p);
    GEN mx37 = FPXQ_POW(mx36, x, t, p);
    GEN mx38 = FPXQ_POW(mx37, x, t, p);
    GEN mx39 = FPXQ_POW(mx38, x, t, p);
    GEN mx40 = FPXQ_POW(mx39, x, t, p);
    GEN mx41 = FPXQ_POW(mx40, x, t, p);
    GEN mx42 = FPXQ_POW(mx41, x, t, p);
    GEN mx43 = FPXQ_POW(mx42, x, t, p);
    GEN mx44 = FPXQ_POW(mx43, x, t, p);
    GEN mx45 = FPXQ_POW(mx44, x, t, p);
    GEN mx46 = FPXQ_POW(mx45, x, t, p);
    GEN mx47 = FPXQ_POW(mx46, x, t, p);
    GEN mx48 = FPXQ_POW(mx47, x, t, p);
    GEN mx49 = FPXQ_POW(mx48, x, t, p);
    GEN mx50 = FPXQ_POW(mx49, x, t, p);
    GEN mx51 = FPXQ_POW(mx50, x, t, p);

    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx2  = FPXQ_INV(mx2,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx11 = FPXQ_INV(mx11, t, p);
    GEN imx12 = FPXQ_INV(mx12, t, p);
    GEN imx13 = FPXQ_INV(mx13, t, p);
    GEN imx14 = FPXQ_INV(mx14, t, p);
    GEN imx15 = FPXQ_INV(mx15, t, p);
    GEN imx16 = FPXQ_INV(mx16, t, p);
    GEN imx17 = FPXQ_INV(mx17, t, p);
    GEN imx18 = FPXQ_INV(mx18, t, p);
    GEN imx19 = FPXQ_INV(mx19, t, p);
    GEN imx20 = FPXQ_INV(mx20, t, p);
    GEN imx21 = FPXQ_INV(mx21, t, p);
    GEN imx22 = FPXQ_INV(mx22, t, p);
    GEN imx23 = FPXQ_INV(mx23, t, p);
    GEN imx24 = FPXQ_INV(mx24, t, p);
    GEN imx25 = FPXQ_INV(mx25, t, p);
    GEN imx26 = FPXQ_INV(mx26, t, p);
    GEN imx27 = FPXQ_INV(mx27, t, p);
    GEN imx28 = FPXQ_INV(mx28, t, p);
    GEN imx29 = FPXQ_INV(mx29, t, p);
    GEN imx30 = FPXQ_INV(mx30, t, p);
    GEN imx31 = FPXQ_INV(mx31, t, p);
    GEN imx32 = FPXQ_INV(mx32, t, p);
    GEN imx33 = FPXQ_INV(mx33, t, p);
    GEN imx34 = FPXQ_INV(mx34, t, p);
    GEN imx35 = FPXQ_INV(mx35, t, p);
    GEN imx36 = FPXQ_INV(mx36, t, p);
    GEN imx37 = FPXQ_INV(mx37, t, p);
    GEN imx38 = FPXQ_INV(mx38, t, p);
    GEN imx39 = FPXQ_INV(mx39, t, p);
    GEN imx40 = FPXQ_INV(mx40, t, p);
    GEN imx41 = FPXQ_INV(mx41, t, p);
    GEN imx42 = FPXQ_INV(mx42, t, p);
    GEN imx43 = FPXQ_INV(mx43, t, p);
    GEN imx44 = FPXQ_INV(mx44, t, p);
    GEN imx45 = FPXQ_INV(mx45, t, p);
    GEN imx46 = FPXQ_INV(mx46, t, p);
    GEN imx47 = FPXQ_INV(mx47, t, p);
    GEN imx48 = FPXQ_INV(mx48, t, p);
    GEN imx49 = FPXQ_INV(mx49, t, p);
    GEN imx50 = FPXQ_INV(mx50, t, p);
    GEN imx51 = FPXQ_INV(mx51, t, p);

    GEN mx10p1  = FPM_FPXQ_MUL(frob_pows[1], mx10,  t, p);
    GEN mx8p1   = FPM_FPXQ_MUL(frob_pows[1], mx8,   t, p);
    GEN mx9p1   = FPM_FPXQ_MUL(frob_pows[1], mx9,   t, p);
    GEN mx24p1  = FPM_FPXQ_MUL(frob_pows[1], mx24,  t, p);
    GEN mx28p1  = FPM_FPXQ_MUL(frob_pows[1], mx28,  t, p);
    GEN mx30p1  = FPM_FPXQ_MUL(frob_pows[1], mx30,  t, p);
    GEN mx34p1  = FPM_FPXQ_MUL(frob_pows[1], mx34,  t, p);
    GEN mx49p1  = FPM_FPXQ_MUL(frob_pows[1], mx49,  t, p);
    GEN imx35p1 = FPM_FPXQ_MUL(frob_pows[1], imx35, t, p);
    GEN imx5p1  = FPM_FPXQ_MUL(frob_pows[1], imx5,  t, p);
    GEN imx50p1 = FPM_FPXQ_MUL(frob_pows[1], imx50, t, p);
    GEN imx3p1  = FPM_FPXQ_MUL(frob_pows[1], imx3,  t, p);
    GEN imx4p1  = FPM_FPXQ_MUL(frob_pows[1], imx4,  t, p);
    GEN imx29p1 = FPM_FPXQ_MUL(frob_pows[1], imx29, t, p);
    GEN imx33p1 = FPM_FPXQ_MUL(frob_pows[1], imx33, t, p);
    GEN imx48p1 = FPM_FPXQ_MUL(frob_pows[1], imx48, t, p);

    GEN mx3p2   = FPM_FPXQ_MUL(frob_pows[2], mx3,   t, p);
    GEN mx7p2   = FPM_FPXQ_MUL(frob_pows[2], mx7,   t, p);
    GEN mx9p2   = FPM_FPXQ_MUL(frob_pows[2], mx9,   t, p);
    GEN mx22p2  = FPM_FPXQ_MUL(frob_pows[2], mx22,  t, p);
    GEN mx23p2  = FPM_FPXQ_MUL(frob_pows[2], mx23,  t, p);
    GEN mx24p2  = FPM_FPXQ_MUL(frob_pows[2], mx24,  t, p);
    GEN mx48p2  = FPM_FPXQ_MUL(frob_pows[2], mx48,  t, p);
    GEN imx2p2  = FPM_FPXQ_MUL(frob_pows[2], imx2,  t, p);
    GEN imx4p2  = FPM_FPXQ_MUL(frob_pows[2], imx4,  t, p);
    GEN imx8p2  = FPM_FPXQ_MUL(frob_pows[2], imx8,  t, p);
    GEN imx47p2 = FPM_FPXQ_MUL(frob_pows[2], imx47, t, p);
    GEN imx49p2 = FPM_FPXQ_MUL(frob_pows[2], imx49, t, p);

    GEN mx27p3  = FPM_FPXQ_MUL(frob_pows[3], mx27,  t, p);
    GEN mx2p3   = FPM_FPXQ_MUL(frob_pows[3], mx2,   t, p);
    GEN mx21p3  = FPM_FPXQ_MUL(frob_pows[3], mx21,  t, p);
    GEN mx23p3  = FPM_FPXQ_MUL(frob_pows[3], mx23,  t, p);
    GEN mx31p3  = FPM_FPXQ_MUL(frob_pows[3], mx31,  t, p);
    GEN mx33p3  = FPM_FPXQ_MUL(frob_pows[3], mx33,  t, p);
    GEN imx26p3 = FPM_FPXQ_MUL(frob_pows[3], imx26, t, p);
    GEN imx7p3  = FPM_FPXQ_MUL(frob_pows[3], imx7,  t, p);
    GEN imx22p3 = FPM_FPXQ_MUL(frob_pows[3], imx22, t, p);
    GEN imx28p3 = FPM_FPXQ_MUL(frob_pows[3], imx28, t, p);
    GEN imx32p3 = FPM_FPXQ_MUL(frob_pows[3], imx32, t, p);

    GEN mx26p4  = FPM_FPXQ_MUL(frob_pows[4], mx26,  t, p);
    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,   t, p);
    GEN mx30p4  = FPM_FPXQ_MUL(frob_pows[4], mx30,  t, p);
    GEN mx32p4  = FPM_FPXQ_MUL(frob_pows[4], mx32,  t, p);
    GEN mx45p4  = FPM_FPXQ_MUL(frob_pows[4], mx45,  t, p);
    GEN mx47p4  = FPM_FPXQ_MUL(frob_pows[4], mx47,  t, p);
    GEN mx1p4   = FPM_FPXQ_MUL(frob_pows[4], mx1,   t, p);
    GEN mx0p4   = FPM_FPXQ_MUL(frob_pows[4], mx0,   t, p);
    GEN imx25p4 = FPM_FPXQ_MUL(frob_pows[4], imx25, t, p);
    GEN imx5p4  = FPM_FPXQ_MUL(frob_pows[4], imx5,  t, p);
    GEN imx27p4 = FPM_FPXQ_MUL(frob_pows[4], imx27, t, p);
    GEN imx6p4  = FPM_FPXQ_MUL(frob_pows[4], imx6,  t, p);
    GEN imx7p4  = FPM_FPXQ_MUL(frob_pows[4], imx7,  t, p);
    GEN imx21p4 = FPM_FPXQ_MUL(frob_pows[4], imx21, t, p);
    GEN imx31p4 = FPM_FPXQ_MUL(frob_pows[4], imx31, t, p);
    GEN imx46p4 = FPM_FPXQ_MUL(frob_pows[4], imx46, t, p);

    GEN mx35p5  = FPM_FPXQ_MUL(frob_pows[5], mx35,  t, p);
    GEN mx25p5  = FPM_FPXQ_MUL(frob_pows[5], mx25,  t, p);
    GEN mx50p5  = FPM_FPXQ_MUL(frob_pows[5], mx50,  t, p);
    GEN mx10p5  = FPM_FPXQ_MUL(frob_pows[5], mx10,  t, p);
    GEN mx29p5  = FPM_FPXQ_MUL(frob_pows[5], mx29,  t, p);
    GEN mx31p5  = FPM_FPXQ_MUL(frob_pows[5], mx31,  t, p);
    GEN mx44p5  = FPM_FPXQ_MUL(frob_pows[5], mx44,  t, p);
    GEN mx46p5  = FPM_FPXQ_MUL(frob_pows[5], mx46,  t, p);
    GEN imx20p5 = FPM_FPXQ_MUL(frob_pows[5], imx20, t, p);
    GEN imx5p5  = FPM_FPXQ_MUL(frob_pows[5], imx5,  t, p);
    GEN imx51p5 = FPM_FPXQ_MUL(frob_pows[5], imx51, t, p);
    GEN imx4p5  = FPM_FPXQ_MUL(frob_pows[5], imx4,  t, p);
    GEN imx6p5  = FPM_FPXQ_MUL(frob_pows[5], imx6,  t, p);
    GEN imx19p5 = FPM_FPXQ_MUL(frob_pows[5], imx19, t, p);
    GEN imx21p5 = FPM_FPXQ_MUL(frob_pows[5], imx21, t, p);
    GEN imx30p5 = FPM_FPXQ_MUL(frob_pows[5], imx30, t, p);
    GEN imx34p5 = FPM_FPXQ_MUL(frob_pows[5], imx34, t, p);
    GEN imx36p5 = FPM_FPXQ_MUL(frob_pows[5], imx36, t, p);
    GEN imx45p5 = FPM_FPXQ_MUL(frob_pows[5], imx45, t, p);
    GEN imx49p5 = FPM_FPXQ_MUL(frob_pows[5], imx49, t, p);

    GEN mx25p6  = FPM_FPXQ_MUL(frob_pows[6], mx25,  t, p);
    GEN mx10p6  = FPM_FPXQ_MUL(frob_pows[6], mx10,  t, p);
    GEN mx4p6   = FPM_FPXQ_MUL(frob_pows[6], mx4,   t, p);
    GEN mx8p6   = FPM_FPXQ_MUL(frob_pows[6], mx8,   t, p);
    GEN mx9p6   = FPM_FPXQ_MUL(frob_pows[6], mx9,   t, p);
    GEN mx19p6  = FPM_FPXQ_MUL(frob_pows[6], mx19,  t, p);
    GEN mx23p6  = FPM_FPXQ_MUL(frob_pows[6], mx23,  t, p);
    GEN mx24p6  = FPM_FPXQ_MUL(frob_pows[6], mx24,  t, p);
    GEN mx34p6  = FPM_FPXQ_MUL(frob_pows[6], mx34,  t, p);
    GEN mx49p6  = FPM_FPXQ_MUL(frob_pows[6], mx49,  t, p);
    GEN imx35p6 = FPM_FPXQ_MUL(frob_pows[6], imx35, t, p);
    GEN imx20p6 = FPM_FPXQ_MUL(frob_pows[6], imx20, t, p);
    GEN imx5p6  = FPM_FPXQ_MUL(frob_pows[6], imx5,  t, p);
    GEN imx50p6 = FPM_FPXQ_MUL(frob_pows[6], imx50, t, p);
    GEN imx3p6  = FPM_FPXQ_MUL(frob_pows[6], imx3,  t, p);
    GEN imx18p6 = FPM_FPXQ_MUL(frob_pows[6], imx18, t, p);
    GEN imx33p6 = FPM_FPXQ_MUL(frob_pows[6], imx33, t, p);
    GEN imx48p6 = FPM_FPXQ_MUL(frob_pows[6], imx48, t, p);

    GEN mx3p7   = FPM_FPXQ_MUL(frob_pows[7], mx3,   t, p);
    GEN mx7p7   = FPM_FPXQ_MUL(frob_pows[7], mx7,   t, p);
    GEN mx9p7   = FPM_FPXQ_MUL(frob_pows[7], mx9,   t, p);
    GEN mx18p7  = FPM_FPXQ_MUL(frob_pows[7], mx18,  t, p);
    GEN mx22p7  = FPM_FPXQ_MUL(frob_pows[7], mx22,  t, p);
    GEN mx24p7  = FPM_FPXQ_MUL(frob_pows[7], mx24,  t, p);
    GEN mx28p7  = FPM_FPXQ_MUL(frob_pows[7], mx28,  t, p);
    GEN mx43p7  = FPM_FPXQ_MUL(frob_pows[7], mx43,  t, p);
    GEN imx27p7 = FPM_FPXQ_MUL(frob_pows[7], imx27, t, p);
    GEN imx8p7  = FPM_FPXQ_MUL(frob_pows[7], imx8,  t, p);
    GEN imx23p7 = FPM_FPXQ_MUL(frob_pows[7], imx23, t, p);
    GEN imx29p7 = FPM_FPXQ_MUL(frob_pows[7], imx29, t, p);
    GEN imx42p7 = FPM_FPXQ_MUL(frob_pows[7], imx42, t, p);
    GEN imx44p7 = FPM_FPXQ_MUL(frob_pows[7], imx44, t, p);

    GEN mx17p8  = FPM_FPXQ_MUL(frob_pows[8], mx17,  t, p);
    GEN mx2p8   = FPM_FPXQ_MUL(frob_pows[8], mx2,   t, p);
    GEN mx48p8  = FPM_FPXQ_MUL(frob_pows[8], mx48,  t, p);
    GEN mx1p8   = FPM_FPXQ_MUL(frob_pows[8], mx1,   t, p);
    GEN mx3p8   = FPM_FPXQ_MUL(frob_pows[8], mx3,   t, p);
    GEN mx16p8  = FPM_FPXQ_MUL(frob_pows[8], mx16,  t, p);
    GEN mx18p8  = FPM_FPXQ_MUL(frob_pows[8], mx18,  t, p);
    GEN mx27p8  = FPM_FPXQ_MUL(frob_pows[8], mx27,  t, p);
    GEN mx31p8  = FPM_FPXQ_MUL(frob_pows[8], mx31,  t, p);
    GEN mx33p8  = FPM_FPXQ_MUL(frob_pows[8], mx33,  t, p);
    GEN mx42p8  = FPM_FPXQ_MUL(frob_pows[8], mx42,  t, p);
    GEN mx46p8  = FPM_FPXQ_MUL(frob_pows[8], mx46,  t, p);
    GEN imx32p8 = FPM_FPXQ_MUL(frob_pows[8], imx32, t, p);
    GEN imx22p8 = FPM_FPXQ_MUL(frob_pows[8], imx22, t, p);
    GEN imx47p8 = FPM_FPXQ_MUL(frob_pows[8], imx47, t, p);
    GEN imx7p8  = FPM_FPXQ_MUL(frob_pows[8], imx7,  t, p);
    GEN imx26p8 = FPM_FPXQ_MUL(frob_pows[8], imx26, t, p);
    GEN imx28p8 = FPM_FPXQ_MUL(frob_pows[8], imx28, t, p);
    GEN imx41p8 = FPM_FPXQ_MUL(frob_pows[8], imx41, t, p);
    GEN imx43p8 = FPM_FPXQ_MUL(frob_pows[8], imx43, t, p);

    GEN mx32p9  = FPM_FPXQ_MUL(frob_pows[9], mx32,  t, p);
    GEN mx17p9  = FPM_FPXQ_MUL(frob_pows[9], mx17,  t, p);
    GEN mx2p9   = FPM_FPXQ_MUL(frob_pows[9], mx2,   t, p);
    GEN mx47p9  = FPM_FPXQ_MUL(frob_pows[9], mx47,  t, p);
    GEN mx15p9  = FPM_FPXQ_MUL(frob_pows[9], mx15,  t, p);
    GEN mx30p9  = FPM_FPXQ_MUL(frob_pows[9], mx30,  t, p);
    GEN mx45p9  = FPM_FPXQ_MUL(frob_pows[9], mx45,  t, p);
    GEN mx0p9   = FPM_FPXQ_MUL(frob_pows[9], mx0,   t, p);
    GEN imx22p9 = FPM_FPXQ_MUL(frob_pows[9], imx22, t, p);
    GEN imx7p9  = FPM_FPXQ_MUL(frob_pows[9], imx7,  t, p);
    GEN imx1p9  = FPM_FPXQ_MUL(frob_pows[9], imx1,  t, p);
    GEN imx5p9  = FPM_FPXQ_MUL(frob_pows[9], imx5,  t, p);
    GEN imx6p9  = FPM_FPXQ_MUL(frob_pows[9], imx6,  t, p);
    GEN imx16p9 = FPM_FPXQ_MUL(frob_pows[9], imx16, t, p);
    GEN imx20p9 = FPM_FPXQ_MUL(frob_pows[9], imx20, t, p);
    GEN imx21p9 = FPM_FPXQ_MUL(frob_pows[9], imx21, t, p);
    GEN imx31p9 = FPM_FPXQ_MUL(frob_pows[9], imx31, t, p);
    GEN imx46p9 = FPM_FPXQ_MUL(frob_pows[9], imx46, t, p);

    GEN mx35p10  = FPM_FPXQ_MUL(frob_pows[10], mx35,  t, p);
    GEN mx20p10  = FPM_FPXQ_MUL(frob_pows[10], mx20,  t, p);
    GEN mx10p10  = FPM_FPXQ_MUL(frob_pows[10], mx10,  t, p);
    GEN mx29p10  = FPM_FPXQ_MUL(frob_pows[10], mx29,  t, p);
    GEN mx31p10  = FPM_FPXQ_MUL(frob_pows[10], mx31,  t, p);
    GEN mx39p10  = FPM_FPXQ_MUL(frob_pows[10], mx39,  t, p);
    GEN mx41p10  = FPM_FPXQ_MUL(frob_pows[10], mx41,  t, p);
    GEN imx40p10 = FPM_FPXQ_MUL(frob_pows[10], imx40, t, p);
    GEN imx15p10 = FPM_FPXQ_MUL(frob_pows[10], imx15, t, p);
    GEN imx5p10  = FPM_FPXQ_MUL(frob_pows[10], imx5,  t, p);
    GEN imx4p10  = FPM_FPXQ_MUL(frob_pows[10], imx4,  t, p);
    GEN imx6p10  = FPM_FPXQ_MUL(frob_pows[10], imx6,  t, p);
    GEN imx19p10 = FPM_FPXQ_MUL(frob_pows[10], imx19, t, p);
    GEN imx21p10 = FPM_FPXQ_MUL(frob_pows[10], imx21, t, p);
    GEN imx30p10 = FPM_FPXQ_MUL(frob_pows[10], imx30, t, p);
    GEN imx34p10 = FPM_FPXQ_MUL(frob_pows[10], imx34, t, p);
    GEN imx36p10 = FPM_FPXQ_MUL(frob_pows[10], imx36, t, p);

    GEN mx40p11  = FPM_FPXQ_MUL(frob_pows[11], mx40,  t, p);
    GEN mx10p11  = FPM_FPXQ_MUL(frob_pows[11], mx10,  t, p);
    GEN mx4p11   = FPM_FPXQ_MUL(frob_pows[11], mx4,   t, p);
    GEN mx8p11   = FPM_FPXQ_MUL(frob_pows[11], mx8,   t, p);
    GEN mx9p11   = FPM_FPXQ_MUL(frob_pows[11], mx9,   t, p);
    GEN mx19p11  = FPM_FPXQ_MUL(frob_pows[11], mx19,  t, p);
    GEN mx34p11  = FPM_FPXQ_MUL(frob_pows[11], mx34,  t, p);
    GEN mx38p11  = FPM_FPXQ_MUL(frob_pows[11], mx38,  t, p);
    GEN mx44p11  = FPM_FPXQ_MUL(frob_pows[11], mx44,  t, p);
    GEN imx35p11 = FPM_FPXQ_MUL(frob_pows[11], imx35, t, p);
    GEN imx15p11 = FPM_FPXQ_MUL(frob_pows[11], imx15, t, p);
    GEN imx5p11  = FPM_FPXQ_MUL(frob_pows[11], imx5,  t, p);
    GEN imx3p11  = FPM_FPXQ_MUL(frob_pows[11], imx3,  t, p);
    GEN imx13p11 = FPM_FPXQ_MUL(frob_pows[11], imx13, t, p);
    GEN imx14p11 = FPM_FPXQ_MUL(frob_pows[11], imx14, t, p);
    GEN imx33p11 = FPM_FPXQ_MUL(frob_pows[11], imx33, t, p);
    GEN imx39p11 = FPM_FPXQ_MUL(frob_pows[11], imx39, t, p);
    GEN imx43p11 = FPM_FPXQ_MUL(frob_pows[11], imx43, t, p);
    GEN imx45p11 = FPM_FPXQ_MUL(frob_pows[11], imx45, t, p);

    GEN mx3p12   = FPM_FPXQ_MUL(frob_pows[12], mx3,   t, p);
    GEN mx7p12   = FPM_FPXQ_MUL(frob_pows[12], mx7,   t, p);
    GEN mx9p12   = FPM_FPXQ_MUL(frob_pows[12], mx9,   t, p);
    GEN mx13p12  = FPM_FPXQ_MUL(frob_pows[12], mx13,  t, p);
    GEN mx17p12  = FPM_FPXQ_MUL(frob_pows[12], mx17,  t, p);
    GEN mx18p12  = FPM_FPXQ_MUL(frob_pows[12], mx18,  t, p);
    GEN mx19p12  = FPM_FPXQ_MUL(frob_pows[12], mx19,  t, p);
    GEN mx28p12  = FPM_FPXQ_MUL(frob_pows[12], mx28,  t, p);
    GEN mx43p12  = FPM_FPXQ_MUL(frob_pows[12], mx43,  t, p);
    GEN imx27p12 = FPM_FPXQ_MUL(frob_pows[12], imx27, t, p);
    GEN imx8p12  = FPM_FPXQ_MUL(frob_pows[12], imx8,  t, p);
    GEN imx12p12 = FPM_FPXQ_MUL(frob_pows[12], imx12, t, p);
    GEN imx14p12 = FPM_FPXQ_MUL(frob_pows[12], imx14, t, p);
    GEN imx29p12 = FPM_FPXQ_MUL(frob_pows[12], imx29, t, p);
    GEN imx42p12 = FPM_FPXQ_MUL(frob_pows[12], imx42, t, p);
    GEN imx44p12 = FPM_FPXQ_MUL(frob_pows[12], imx44, t, p);

    GEN mx37p13  = FPM_FPXQ_MUL(frob_pows[13], mx37,  t, p);
    GEN mx12p13  = FPM_FPXQ_MUL(frob_pows[13], mx12,  t, p);
    GEN mx2p13   = FPM_FPXQ_MUL(frob_pows[13], mx2,   t, p);
    GEN mx1p13   = FPM_FPXQ_MUL(frob_pows[13], mx1,   t, p);
    GEN mx3p13   = FPM_FPXQ_MUL(frob_pows[13], mx3,   t, p);
    GEN mx16p13  = FPM_FPXQ_MUL(frob_pows[13], mx16,  t, p);
    GEN mx18p13  = FPM_FPXQ_MUL(frob_pows[13], mx18,  t, p);
    GEN mx27p13  = FPM_FPXQ_MUL(frob_pows[13], mx27,  t, p);
    GEN mx31p13  = FPM_FPXQ_MUL(frob_pows[13], mx31,  t, p);
    GEN mx33p13  = FPM_FPXQ_MUL(frob_pows[13], mx33,  t, p);
    GEN imx32p13 = FPM_FPXQ_MUL(frob_pows[13], imx32, t, p);
    GEN imx17p13 = FPM_FPXQ_MUL(frob_pows[13], imx17, t, p);
    GEN imx7p13  = FPM_FPXQ_MUL(frob_pows[13], imx7,  t, p);
    GEN imx26p13 = FPM_FPXQ_MUL(frob_pows[13], imx26, t, p);
    GEN imx28p13 = FPM_FPXQ_MUL(frob_pows[13], imx28, t, p);
    GEN imx36p13 = FPM_FPXQ_MUL(frob_pows[13], imx36, t, p);
    GEN imx38p13 = FPM_FPXQ_MUL(frob_pows[13], imx38, t, p);

    GEN mx32p14  = FPM_FPXQ_MUL(frob_pows[14], mx32,  t, p);
    GEN mx12p14  = FPM_FPXQ_MUL(frob_pows[14], mx12,  t, p);
    GEN mx2p14   = FPM_FPXQ_MUL(frob_pows[14], mx2,   t, p);
    GEN mx10p14  = FPM_FPXQ_MUL(frob_pows[14], mx10,  t, p);
    GEN mx11p14  = FPM_FPXQ_MUL(frob_pows[14], mx11,  t, p);
    GEN mx30p14  = FPM_FPXQ_MUL(frob_pows[14], mx30,  t, p);
    GEN mx36p14  = FPM_FPXQ_MUL(frob_pows[14], mx36,  t, p);
    GEN mx40p14  = FPM_FPXQ_MUL(frob_pows[14], mx40,  t, p);
    GEN mx42p14  = FPM_FPXQ_MUL(frob_pows[14], mx42,  t, p);
    GEN mx0p14   = FPM_FPXQ_MUL(frob_pows[14], mx0,   t, p);
    GEN imx37p14 = FPM_FPXQ_MUL(frob_pows[14], imx37, t, p);
    GEN imx7p14  = FPM_FPXQ_MUL(frob_pows[14], imx7,  t, p);
    GEN imx1p14  = FPM_FPXQ_MUL(frob_pows[14], imx1,  t, p);
    GEN imx5p14  = FPM_FPXQ_MUL(frob_pows[14], imx5,  t, p);
    GEN imx6p14  = FPM_FPXQ_MUL(frob_pows[14], imx6,  t, p);
    GEN imx16p14 = FPM_FPXQ_MUL(frob_pows[14], imx16, t, p);
    GEN imx31p14 = FPM_FPXQ_MUL(frob_pows[14], imx31, t, p);
    GEN imx35p14 = FPM_FPXQ_MUL(frob_pows[14], imx35, t, p);
    GEN imx41p14 = FPM_FPXQ_MUL(frob_pows[14], imx41, t, p);

    GEN mx35p15  = FPM_FPXQ_MUL(frob_pows[15], mx35,  t, p);
    GEN mx10p15  = FPM_FPXQ_MUL(frob_pows[15], mx10,  t, p);
    GEN mx9p15   = FPM_FPXQ_MUL(frob_pows[15], mx9,   t, p);
    GEN mx11p15  = FPM_FPXQ_MUL(frob_pows[15], mx11,  t, p);
    GEN mx29p15  = FPM_FPXQ_MUL(frob_pows[15], mx29,  t, p);
    GEN mx31p15  = FPM_FPXQ_MUL(frob_pows[15], mx31,  t, p);
    GEN mx39p15  = FPM_FPXQ_MUL(frob_pows[15], mx39,  t, p);
    GEN mx41p15  = FPM_FPXQ_MUL(frob_pows[15], mx41,  t, p);
    GEN imx40p15 = FPM_FPXQ_MUL(frob_pows[15], imx40, t, p);
    GEN imx15p15 = FPM_FPXQ_MUL(frob_pows[15], imx15, t, p);
    GEN imx5p15  = FPM_FPXQ_MUL(frob_pows[15], imx5,  t, p);
    GEN imx4p15  = FPM_FPXQ_MUL(frob_pows[15], imx4,  t, p);
    GEN imx6p15  = FPM_FPXQ_MUL(frob_pows[15], imx6,  t, p);
    GEN imx14p15 = FPM_FPXQ_MUL(frob_pows[15], imx14, t, p);
    GEN imx16p15 = FPM_FPXQ_MUL(frob_pows[15], imx16, t, p);
    GEN imx30p15 = FPM_FPXQ_MUL(frob_pows[15], imx30, t, p);
    GEN imx34p15 = FPM_FPXQ_MUL(frob_pows[15], imx34, t, p);
    GEN imx36p15 = FPM_FPXQ_MUL(frob_pows[15], imx36, t, p);

    GEN mx10p16  = FPM_FPXQ_MUL(frob_pows[16], mx10,  t, p);
    GEN mx4p16   = FPM_FPXQ_MUL(frob_pows[16], mx4,   t, p);
    GEN mx8p16   = FPM_FPXQ_MUL(frob_pows[16], mx8,   t, p);
    GEN mx14p16  = FPM_FPXQ_MUL(frob_pows[16], mx14,  t, p);
    GEN imx15p16 = FPM_FPXQ_MUL(frob_pows[16], imx15, t, p);
    GEN imx5p16  = FPM_FPXQ_MUL(frob_pows[16], imx5,  t, p);
    GEN imx3p16  = FPM_FPXQ_MUL(frob_pows[16], imx3,  t, p);
    GEN imx9p16  = FPM_FPXQ_MUL(frob_pows[16], imx9,  t, p);
    GEN imx13p16 = FPM_FPXQ_MUL(frob_pows[16], imx13, t, p);

    GEN mx38p17  = FPM_FPXQ_MUL(frob_pows[17], mx38,  t, p);
    GEN mx28p17  = FPM_FPXQ_MUL(frob_pows[17], mx28,  t, p);
    GEN mx13p17  = FPM_FPXQ_MUL(frob_pows[17], mx13,  t, p);
    GEN mx32p17  = FPM_FPXQ_MUL(frob_pows[17], mx32,  t, p);
    GEN mx34p17  = FPM_FPXQ_MUL(frob_pows[17], mx34,  t, p);
    GEN mx3p17   = FPM_FPXQ_MUL(frob_pows[17], mx3,   t, p);
    GEN imx8p17  = FPM_FPXQ_MUL(frob_pows[17], imx8,  t, p);
    GEN imx29p17 = FPM_FPXQ_MUL(frob_pows[17], imx29, t, p);
    GEN imx27p17 = FPM_FPXQ_MUL(frob_pows[17], imx27, t, p);
    GEN imx33p17 = FPM_FPXQ_MUL(frob_pows[17], imx33, t, p);
    GEN imx37p17 = FPM_FPXQ_MUL(frob_pows[17], imx37, t, p);
    GEN imx39p17 = FPM_FPXQ_MUL(frob_pows[17], imx39, t, p);

    GEN mx37p18  = FPM_FPXQ_MUL(frob_pows[18], mx37,  t, p);
    GEN mx12p18  = FPM_FPXQ_MUL(frob_pows[18], mx12,  t, p);
    GEN mx2p18   = FPM_FPXQ_MUL(frob_pows[18], mx2,   t, p);
    GEN mx1p18   = FPM_FPXQ_MUL(frob_pows[18], mx1,   t, p);
    GEN mx3p18   = FPM_FPXQ_MUL(frob_pows[18], mx3,   t, p);
    GEN mx11p18  = FPM_FPXQ_MUL(frob_pows[18], mx11,  t, p);
    GEN mx13p18  = FPM_FPXQ_MUL(frob_pows[18], mx13,  t, p);
    GEN mx27p18  = FPM_FPXQ_MUL(frob_pows[18], mx27,  t, p);
    GEN mx31p18  = FPM_FPXQ_MUL(frob_pows[18], mx31,  t, p);
    GEN mx33p18  = FPM_FPXQ_MUL(frob_pows[18], mx33,  t, p);
    GEN imx32p18 = FPM_FPXQ_MUL(frob_pows[18], imx32, t, p);
    GEN imx7p18  = FPM_FPXQ_MUL(frob_pows[18], imx7,  t, p);
    GEN imx6p18  = FPM_FPXQ_MUL(frob_pows[18], imx6,  t, p);
    GEN imx8p18  = FPM_FPXQ_MUL(frob_pows[18], imx8,  t, p);
    GEN imx26p18 = FPM_FPXQ_MUL(frob_pows[18], imx26, t, p);
    GEN imx28p18 = FPM_FPXQ_MUL(frob_pows[18], imx28, t, p);
    GEN imx36p18 = FPM_FPXQ_MUL(frob_pows[18], imx36, t, p);
    GEN imx38p18 = FPM_FPXQ_MUL(frob_pows[18], imx38, t, p);

    GEN mx12p19  = FPM_FPXQ_MUL(frob_pows[19], mx12,  t, p);
    GEN mx2p19   = FPM_FPXQ_MUL(frob_pows[19], mx2,   t, p);
    GEN mx6p19   = FPM_FPXQ_MUL(frob_pows[19], mx6,   t, p);
    GEN mx10p19  = FPM_FPXQ_MUL(frob_pows[19], mx10,  t, p);
    GEN mx0p19   = FPM_FPXQ_MUL(frob_pows[19], mx0,   t, p);
    GEN imx7p19  = FPM_FPXQ_MUL(frob_pows[19], imx7,  t, p);
    GEN imx1p19  = FPM_FPXQ_MUL(frob_pows[19], imx1,  t, p);
    GEN imx5p19  = FPM_FPXQ_MUL(frob_pows[19], imx5,  t, p);
    GEN imx11p19 = FPM_FPXQ_MUL(frob_pows[19], imx11, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx35, mx25,   t, p);
    y0 = FPXQ_MUL(y0, mx29,     t, p);
    y0 = FPXQ_MUL(y0, mx31,     t, p);
    y0 = FPXQ_MUL(y0, imx26,    t, p);
    y0 = FPXQ_MUL(y0, imx24,    t, p);
    y0 = FPXQ_MUL(y0, imx30,    t, p);
    y0 = FPXQ_MUL(y0, imx34,    t, p);
    y0 = FPXQ_MUL(y0, imx36,    t, p);
    y0 = FPXQ_MUL(y0, mx10p1,   t, p);
    y0 = FPXQ_MUL(y0, mx8p1,    t, p);
    y0 = FPXQ_MUL(y0, mx28p1,   t, p);
    y0 = FPXQ_MUL(y0, mx30p1,   t, p);
    y0 = FPXQ_MUL(y0, mx34p1,   t, p);
    y0 = FPXQ_MUL(y0, mx49p1,   t, p);
    y0 = FPXQ_MUL(y0, imx35p1,  t, p);
    y0 = FPXQ_MUL(y0, imx5p1,   t, p);
    y0 = FPXQ_MUL(y0, imx50p1,  t, p);
    y0 = FPXQ_MUL(y0, imx3p1,   t, p);
    y0 = FPXQ_MUL(y0, imx29p1,  t, p);
    y0 = FPXQ_MUL(y0, imx33p1,  t, p);
    y0 = FPXQ_MUL(y0, imx48p1,  t, p);
    y0 = FPXQ_MUL(y0, mx3p2,    t, p);
    y0 = FPXQ_MUL(y0, mx7p2,    t, p);
    y0 = FPXQ_MUL(y0, mx9p2,    t, p);
    y0 = FPXQ_MUL(y0, mx22p2,   t, p);
    y0 = FPXQ_MUL(y0, mx24p2,   t, p);
    y0 = FPXQ_MUL(y0, mx48p2,   t, p);
    y0 = FPXQ_MUL(y0, imx2p2,   t, p);
    y0 = FPXQ_MUL(y0, imx4p2,   t, p);
    y0 = FPXQ_MUL(y0, imx8p2,   t, p);
    y0 = FPXQ_MUL(y0, imx47p2,  t, p);
    y0 = FPXQ_MUL(y0, imx49p2,  t, p);
    y0 = FPXQ_MUL(y0, mx27p3,   t, p);
    y0 = FPXQ_MUL(y0, mx21p3,   t, p);
    y0 = FPXQ_MUL(y0, mx23p3,   t, p);
    y0 = FPXQ_MUL(y0, mx31p3,   t, p);
    y0 = FPXQ_MUL(y0, mx33p3,   t, p);
    y0 = FPXQ_MUL(y0, imx26p3,  t, p);
    y0 = FPXQ_MUL(y0, imx22p3,  t, p);
    y0 = FPXQ_MUL(y0, imx28p3,  t, p);
    y0 = FPXQ_MUL(y0, imx32p3,  t, p);
    y0 = FPXQ_MUL(y0, mx26p4,   t, p);
    y0 = FPXQ_MUL(y0, mx2p4,    t, p);
    y0 = FPXQ_MUL(y0, mx30p4,   t, p);
    y0 = FPXQ_MUL(y0, mx32p4,   t, p);
    y0 = FPXQ_MUL(y0, mx45p4,   t, p);
    y0 = FPXQ_MUL(y0, mx47p4,   t, p);
    y0 = FPXQ_MUL(y0, mx0p4,    t, p);
    y0 = FPXQ_MUL(y0, imx25p4,  t, p);
    y0 = FPXQ_MUL(y0, imx5p4,   t, p);
    y0 = FPXQ_MUL(y0, imx27p4,  t, p);
    y0 = FPXQ_MUL(y0, imx7p4,   t, p);
    y0 = FPXQ_MUL(y0, imx31p4,  t, p);
    y0 = FPXQ_MUL(y0, imx46p4,  t, p);
    y0 = FPXQ_MUL(y0, mx35p5,   t, p);
    y0 = FPXQ_MUL(y0, mx50p5,   t, p);
    y0 = FPXQ_MUL(y0, mx29p5,   t, p);
    y0 = FPXQ_MUL(y0, mx31p5,   t, p);
    y0 = FPXQ_MUL(y0, mx44p5,   t, p);
    y0 = FPXQ_MUL(y0, mx46p5,   t, p);
    y0 = FPXQ_MUL(y0, imx51p5,  t, p);
    y0 = FPXQ_MUL(y0, imx4p5,   t, p);
    y0 = FPXQ_MUL(y0, imx6p5,   t, p);
    y0 = FPXQ_MUL(y0, imx19p5,  t, p);
    y0 = FPXQ_MUL(y0, imx21p5,  t, p);
    y0 = FPXQ_MUL(y0, imx30p5,  t, p);
    y0 = FPXQ_MUL(y0, imx34p5,  t, p);
    y0 = FPXQ_MUL(y0, imx36p5,  t, p);
    y0 = FPXQ_MUL(y0, imx45p5,  t, p);
    y0 = FPXQ_MUL(y0, imx49p5,  t, p);
    y0 = FPXQ_MUL(y0, mx25p6,   t, p);
    y0 = FPXQ_MUL(y0, mx10p6,   t, p);
    y0 = FPXQ_MUL(y0, mx4p6,    t, p);
    y0 = FPXQ_MUL(y0, mx8p6,    t, p);
    y0 = FPXQ_MUL(y0, mx19p6,   t, p);
    y0 = FPXQ_MUL(y0, mx23p6,   t, p);
    y0 = FPXQ_MUL(y0, mx34p6,   t, p);
    y0 = FPXQ_MUL(y0, mx49p6,   t, p);
    y0 = FPXQ_MUL(y0, imx35p6,  t, p);
    y0 = FPXQ_MUL(y0, imx20p6,  t, p);
    y0 = FPXQ_MUL(y0, imx5p6,   t, p);
    y0 = FPXQ_MUL(y0, imx50p6,  t, p);
    y0 = FPXQ_MUL(y0, imx3p6,   t, p);
    y0 = FPXQ_MUL(y0, imx18p6,  t, p);
    y0 = FPXQ_MUL(y0, imx33p6,  t, p);
    y0 = FPXQ_MUL(y0, imx48p6,  t, p);
    y0 = FPXQ_MUL(y0, mx7p7,    t, p);
    y0 = FPXQ_MUL(y0, mx9p7,    t, p);
    y0 = FPXQ_MUL(y0, mx22p7,   t, p);
    y0 = FPXQ_MUL(y0, mx24p7,   t, p);
    y0 = FPXQ_MUL(y0, mx28p7,   t, p);
    y0 = FPXQ_MUL(y0, mx43p7,   t, p);
    y0 = FPXQ_MUL(y0, imx27p7,  t, p);
    y0 = FPXQ_MUL(y0, imx8p7,   t, p);
    y0 = FPXQ_MUL(y0, imx23p7,  t, p);
    y0 = FPXQ_MUL(y0, imx29p7,  t, p);
    y0 = FPXQ_MUL(y0, imx42p7,  t, p);
    y0 = FPXQ_MUL(y0, imx44p7,  t, p);
    y0 = FPXQ_MUL(y0, mx48p8,   t, p);
    y0 = FPXQ_MUL(y0, mx1p8,    t, p);
    y0 = FPXQ_MUL(y0, mx3p8,    t, p);
    y0 = FPXQ_MUL(y0, mx16p8,   t, p);
    y0 = FPXQ_MUL(y0, mx18p8,   t, p);
    y0 = FPXQ_MUL(y0, mx27p8,   t, p);
    y0 = FPXQ_MUL(y0, mx31p8,   t, p);
    y0 = FPXQ_MUL(y0, mx33p8,   t, p);
    y0 = FPXQ_MUL(y0, mx42p8,   t, p);
    y0 = FPXQ_MUL(y0, mx46p8,   t, p);
    y0 = FPXQ_MUL(y0, imx32p8,  t, p);
    y0 = FPXQ_MUL(y0, imx47p8,  t, p);
    y0 = FPXQ_MUL(y0, imx26p8,  t, p);
    y0 = FPXQ_MUL(y0, imx28p8,  t, p);
    y0 = FPXQ_MUL(y0, imx41p8,  t, p);
    y0 = FPXQ_MUL(y0, imx43p8,  t, p);
    y0 = FPXQ_MUL(y0, mx32p9,   t, p);
    y0 = FPXQ_MUL(y0, mx17p9,   t, p);
    y0 = FPXQ_MUL(y0, mx2p9,    t, p);
    y0 = FPXQ_MUL(y0, mx47p9,   t, p);
    y0 = FPXQ_MUL(y0, mx15p9,   t, p);
    y0 = FPXQ_MUL(y0, mx30p9,   t, p);
    y0 = FPXQ_MUL(y0, mx45p9,   t, p);
    y0 = FPXQ_MUL(y0, mx0p9,    t, p);
    y0 = FPXQ_MUL(y0, imx22p9,  t, p);
    y0 = FPXQ_MUL(y0, imx7p9,   t, p);
    y0 = FPXQ_MUL(y0, imx1p9,   t, p);
    y0 = FPXQ_MUL(y0, imx5p9,   t, p);
    y0 = FPXQ_MUL(y0, imx16p9,  t, p);
    y0 = FPXQ_MUL(y0, imx20p9,  t, p);
    y0 = FPXQ_MUL(y0, imx31p9,  t, p);
    y0 = FPXQ_MUL(y0, imx46p9,  t, p);
    y0 = FPXQ_MUL(y0, mx35p10,  t, p);
    y0 = FPXQ_MUL(y0, mx20p10,  t, p);
    y0 = FPXQ_MUL(y0, mx29p10,  t, p);
    y0 = FPXQ_MUL(y0, mx31p10,  t, p);
    y0 = FPXQ_MUL(y0, mx39p10,  t, p);
    y0 = FPXQ_MUL(y0, mx41p10,  t, p);
    y0 = FPXQ_MUL(y0, imx40p10, t, p);
    y0 = FPXQ_MUL(y0, imx4p10,  t, p);
    y0 = FPXQ_MUL(y0, imx6p10,  t, p);
    y0 = FPXQ_MUL(y0, imx19p10, t, p);
    y0 = FPXQ_MUL(y0, imx21p10, t, p);
    y0 = FPXQ_MUL(y0, imx30p10, t, p);
    y0 = FPXQ_MUL(y0, imx34p10, t, p);
    y0 = FPXQ_MUL(y0, imx36p10, t, p);
    y0 = FPXQ_MUL(y0, mx40p11,  t, p);
    y0 = FPXQ_MUL(y0, mx10p11,  t, p);
    y0 = FPXQ_MUL(y0, mx4p11,   t, p);
    y0 = FPXQ_MUL(y0, mx8p11,   t, p);
    y0 = FPXQ_MUL(y0, mx34p11,  t, p);
    y0 = FPXQ_MUL(y0, mx38p11,  t, p);
    y0 = FPXQ_MUL(y0, mx44p11,  t, p);
    y0 = FPXQ_MUL(y0, imx35p11, t, p);
    y0 = FPXQ_MUL(y0, imx15p11, t, p);
    y0 = FPXQ_MUL(y0, imx5p11,  t, p);
    y0 = FPXQ_MUL(y0, imx3p11,  t, p);
    y0 = FPXQ_MUL(y0, imx13p11, t, p);
    y0 = FPXQ_MUL(y0, imx33p11, t, p);
    y0 = FPXQ_MUL(y0, imx39p11, t, p);
    y0 = FPXQ_MUL(y0, imx43p11, t, p);
    y0 = FPXQ_MUL(y0, imx45p11, t, p);
    y0 = FPXQ_MUL(y0, mx7p12,   t, p);
    y0 = FPXQ_MUL(y0, mx9p12,   t, p);
    y0 = FPXQ_MUL(y0, mx13p12,  t, p);
    y0 = FPXQ_MUL(y0, mx17p12,  t, p);
    y0 = FPXQ_MUL(y0, mx19p12,  t, p);
    y0 = FPXQ_MUL(y0, mx28p12,  t, p);
    y0 = FPXQ_MUL(y0, mx43p12,  t, p);
    y0 = FPXQ_MUL(y0, imx27p12, t, p);
    y0 = FPXQ_MUL(y0, imx8p12,  t, p);
    y0 = FPXQ_MUL(y0, imx12p12, t, p);
    y0 = FPXQ_MUL(y0, imx14p12, t, p);
    y0 = FPXQ_MUL(y0, imx29p12, t, p);
    y0 = FPXQ_MUL(y0, imx42p12, t, p);
    y0 = FPXQ_MUL(y0, imx44p12, t, p);
    y0 = FPXQ_MUL(y0, mx37p13,  t, p);
    y0 = FPXQ_MUL(y0, mx1p13,   t, p);
    y0 = FPXQ_MUL(y0, mx3p13,   t, p);
    y0 = FPXQ_MUL(y0, mx16p13,  t, p);
    y0 = FPXQ_MUL(y0, mx18p13,  t, p);
    y0 = FPXQ_MUL(y0, mx27p13,  t, p);
    y0 = FPXQ_MUL(y0, mx31p13,  t, p);
    y0 = FPXQ_MUL(y0, mx33p13,  t, p);
    y0 = FPXQ_MUL(y0, imx32p13, t, p);
    y0 = FPXQ_MUL(y0, imx17p13, t, p);
    y0 = FPXQ_MUL(y0, imx26p13, t, p);
    y0 = FPXQ_MUL(y0, imx28p13, t, p);
    y0 = FPXQ_MUL(y0, imx36p13, t, p);
    y0 = FPXQ_MUL(y0, imx38p13, t, p);
    y0 = FPXQ_MUL(y0, mx32p14,  t, p);
    y0 = FPXQ_MUL(y0, mx12p14,  t, p);
    y0 = FPXQ_MUL(y0, mx2p14,   t, p);
    y0 = FPXQ_MUL(y0, mx10p14,  t, p);
    y0 = FPXQ_MUL(y0, mx30p14,  t, p);
    y0 = FPXQ_MUL(y0, mx36p14,  t, p);
    y0 = FPXQ_MUL(y0, mx40p14,  t, p);
    y0 = FPXQ_MUL(y0, mx42p14,  t, p);
    y0 = FPXQ_MUL(y0, mx0p14,   t, p);
    y0 = FPXQ_MUL(y0, imx37p14, t, p);
    y0 = FPXQ_MUL(y0, imx7p14,  t, p);
    y0 = FPXQ_MUL(y0, imx1p14,  t, p);
    y0 = FPXQ_MUL(y0, imx5p14,  t, p);
    y0 = FPXQ_MUL(y0, imx31p14, t, p);
    y0 = FPXQ_MUL(y0, imx35p14, t, p);
    y0 = FPXQ_MUL(y0, imx41p14, t, p);
    y0 = FPXQ_MUL(y0, mx35p15,  t, p);
    y0 = FPXQ_MUL(y0, mx9p15,   t, p);
    y0 = FPXQ_MUL(y0, mx11p15,  t, p);
    y0 = FPXQ_MUL(y0, mx29p15,  t, p);
    y0 = FPXQ_MUL(y0, mx31p15,  t, p);
    y0 = FPXQ_MUL(y0, mx39p15,  t, p);
    y0 = FPXQ_MUL(y0, mx41p15,  t, p);
    y0 = FPXQ_MUL(y0, imx40p15, t, p);
    y0 = FPXQ_MUL(y0, imx4p15,  t, p);
    y0 = FPXQ_MUL(y0, imx6p15,  t, p);
    y0 = FPXQ_MUL(y0, imx14p15, t, p);
    y0 = FPXQ_MUL(y0, imx16p15, t, p);
    y0 = FPXQ_MUL(y0, imx30p15, t, p);
    y0 = FPXQ_MUL(y0, imx34p15, t, p);
    y0 = FPXQ_MUL(y0, imx36p15, t, p);
    y0 = FPXQ_MUL(y0, mx10p16,  t, p);
    y0 = FPXQ_MUL(y0, mx4p16,   t, p);
    y0 = FPXQ_MUL(y0, mx8p16,   t, p);
    y0 = FPXQ_MUL(y0, mx14p16,  t, p);
    y0 = FPXQ_MUL(y0, imx15p16, t, p);
    y0 = FPXQ_MUL(y0, imx5p16,  t, p);
    y0 = FPXQ_MUL(y0, imx3p16,  t, p);
    y0 = FPXQ_MUL(y0, imx9p16,  t, p);
    y0 = FPXQ_MUL(y0, imx13p16, t, p);
    y0 = FPXQ_MUL(y0, mx38p17,  t, p);
    y0 = FPXQ_MUL(y0, mx28p17,  t, p);
    y0 = FPXQ_MUL(y0, mx32p17,  t, p);
    y0 = FPXQ_MUL(y0, mx34p17,  t, p);
    y0 = FPXQ_MUL(y0, imx29p17, t, p);
    y0 = FPXQ_MUL(y0, imx27p17, t, p);
    y0 = FPXQ_MUL(y0, imx33p17, t, p);
    y0 = FPXQ_MUL(y0, imx37p17, t, p);
    y0 = FPXQ_MUL(y0, imx39p17, t, p);
    y0 = FPXQ_MUL(y0, mx37p18,  t, p);
    y0 = FPXQ_MUL(y0, mx1p18,   t, p);
    y0 = FPXQ_MUL(y0, mx3p18,   t, p);
    y0 = FPXQ_MUL(y0, mx11p18,  t, p);
    y0 = FPXQ_MUL(y0, mx13p18,  t, p);
    y0 = FPXQ_MUL(y0, mx27p18,  t, p);
    y0 = FPXQ_MUL(y0, mx31p18,  t, p);
    y0 = FPXQ_MUL(y0, mx33p18,  t, p);
    y0 = FPXQ_MUL(y0, imx32p18, t, p);
    y0 = FPXQ_MUL(y0, imx6p18,  t, p);
    y0 = FPXQ_MUL(y0, imx8p18,  t, p);
    y0 = FPXQ_MUL(y0, imx26p18, t, p);
    y0 = FPXQ_MUL(y0, imx28p18, t, p);
    y0 = FPXQ_MUL(y0, imx36p18, t, p);
    y0 = FPXQ_MUL(y0, imx38p18, t, p);
    y0 = FPXQ_MUL(y0, mx12p19,  t, p);
    y0 = FPXQ_MUL(y0, mx2p19,   t, p);
    y0 = FPXQ_MUL(y0, mx6p19,   t, p);
    y0 = FPXQ_MUL(y0, mx10p19,  t, p);
    y0 = FPXQ_MUL(y0, mx0p19,   t, p);
    y0 = FPXQ_MUL(y0, imx7p19,  t, p);
    y0 = FPXQ_MUL(y0, imx1p19,  t, p);
    y0 = FPXQ_MUL(y0, imx5p19,  t, p);
    y0 = FPXQ_MUL(y0, imx11p19, t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx9p1, imx4p1, t, p);
    y1 = FPXQ_MUL(y1, mx23p2,    t, p);
    y1 = FPXQ_MUL(y1, mx1p4,     t, p);
    y1 = FPXQ_MUL(y1, imx6p4,    t, p);
    y1 = FPXQ_MUL(y1, imx20p5,   t, p);
    y1 = FPXQ_MUL(y1, imx5p5,    t, p);
    y1 = FPXQ_MUL(y1, mx9p6,     t, p);
    y1 = FPXQ_MUL(y1, mx24p6,    t, p);
    y1 = FPXQ_MUL(y1, mx17p8,    t, p);
    y1 = FPXQ_MUL(y1, mx2p8,     t, p);
    y1 = FPXQ_MUL(y1, imx6p9,    t, p);
    y1 = FPXQ_MUL(y1, imx21p9,   t, p);
    y1 = FPXQ_MUL(y1, imx5p10,   t, p);
    y1 = FPXQ_MUL(y1, mx9p11,    t, p);
    y1 = FPXQ_MUL(y1, imx14p11,  t, p);
    y1 = FPXQ_MUL(y1, mx18p12,   t, p);
    y1 = FPXQ_MUL(y1, mx2p13,    t, p);
    y1 = FPXQ_MUL(y1, mx11p14,   t, p);
    y1 = FPXQ_MUL(y1, imx6p14,   t, p);
    y1 = FPXQ_MUL(y1, mx10p15,   t, p);
    y1 = FPXQ_MUL(y1, imx15p15,  t, p);
    y1 = FPXQ_MUL(y1, imx5p15,   t, p);
    y1 = FPXQ_MUL(y1, mx12p18,   t, p);
    y1 = FPXQ_MUL(y1, mx2p18,    t, p);
    y1 = FPXQ_MUL(y1, imx7p18,   t, p);

    GEN y2;
    y2 = FPXQ_MUL(mx10, mx0,    t, p);
    y2 = FPXQ_MUL(y2, imx5,     t, p);
    y2 = FPXQ_MUL(y2, mx24p1,   t, p);
    y2 = FPXQ_MUL(y2, mx2p3,    t, p);
    y2 = FPXQ_MUL(y2, imx7p3,   t, p);
    y2 = FPXQ_MUL(y2, imx21p4,  t, p);
    y2 = FPXQ_MUL(y2, mx25p5,   t, p);
    y2 = FPXQ_MUL(y2, mx10p5,   t, p);
    y2 = FPXQ_MUL(y2, mx3p7,    t, p);
    y2 = FPXQ_MUL(y2, mx18p7,   t, p);
    y2 = FPXQ_MUL(y2, imx22p8,  t, p);
    y2 = FPXQ_MUL(y2, imx7p8,   t, p);
    y2 = FPXQ_MUL(y2, mx10p10,  t, p);
    y2 = FPXQ_MUL(y2, imx15p10, t, p);
    y2 = FPXQ_MUL(y2, mx19p11,  t, p);
    y2 = FPXQ_MUL(y2, mx3p12,   t, p);
    y2 = FPXQ_MUL(y2, mx12p13,  t, p);
    y2 = FPXQ_MUL(y2, imx7p13,  t, p);
    y2 = FPXQ_MUL(y2, imx16p14, t, p);
    y2 = FPXQ_MUL(y2, mx13p17,  t, p);
    y2 = FPXQ_MUL(y2, mx3p17,   t, p);
    y2 = FPXQ_MUL(y2, imx8p17,  t, p);

    GEN T0 = FPXQ_MUL(y2, y1, t, p);
    GEN T1 = FPXQ_MUL(y2, y0, t, p);
    GEN T2 = FPXQ_SQR(T0, t, p);
    GEN T3 = FPXQ_MUL(T2, T1, t, p);
    
    return T3;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_26_1166(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);
    GEN mx12 = FPXQ_POW(mx11, x, t, p);
    GEN mx13 = FPXQ_POW(mx12, x, t, p);
    GEN mx14 = FPXQ_POW(mx13, x, t, p);
    GEN mx15 = FPXQ_POW(mx14, x, t, p);
    GEN mx16 = FPXQ_POW(mx15, x, t, p);
    GEN mx17 = FPXQ_POW(mx16, x, t, p);
    GEN mx18 = FPXQ_POW(mx17, x, t, p);
    GEN mx19 = FPXQ_POW(mx18, x, t, p);
    GEN mx20 = FPXQ_POW(mx19, x, t, p);
    GEN mx21 = FPXQ_POW(mx20, x, t, p);
    GEN mx22 = FPXQ_POW(mx21, x, t, p);
    GEN mx23 = FPXQ_POW(mx22, x, t, p);
    GEN mx24 = FPXQ_POW(mx23, x, t, p);
    GEN mx25 = FPXQ_POW(mx24, x, t, p);
    GEN mx26 = FPXQ_POW(mx25, x, t, p);
    GEN mx27 = FPXQ_POW(mx26, x, t, p);

    GEN imx0  = FPXQ_INV(mx0,  t, p);
    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx2  = FPXQ_INV(mx2,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);
    GEN imx11 = FPXQ_INV(mx11, t, p);
    GEN imx12 = FPXQ_INV(mx12, t, p);
    GEN imx14 = FPXQ_INV(mx14, t, p);
    GEN imx15 = FPXQ_INV(mx15, t, p);
    GEN imx16 = FPXQ_INV(mx16, t, p);
    GEN imx17 = FPXQ_INV(mx17, t, p);
    GEN imx18 = FPXQ_INV(mx18, t, p);
    GEN imx19 = FPXQ_INV(mx19, t, p);
    GEN imx20 = FPXQ_INV(mx20, t, p);
    GEN imx21 = FPXQ_INV(mx21, t, p);
    GEN imx22 = FPXQ_INV(mx22, t, p);
    GEN imx23 = FPXQ_INV(mx23, t, p);
    GEN imx24 = FPXQ_INV(mx24, t, p);
    GEN imx25 = FPXQ_INV(mx25, t, p);

    GEN mx27p1  = FPM_FPXQ_MUL(frob_pows[1], mx27,  t, p);
    GEN mx13p1  = FPM_FPXQ_MUL(frob_pows[1], mx13,  t, p);
    GEN mx1p1   = FPM_FPXQ_MUL(frob_pows[1], mx1,   t, p);
    GEN imx24p1 = FPM_FPXQ_MUL(frob_pows[1], imx24, t, p);
    GEN imx15p1 = FPM_FPXQ_MUL(frob_pows[1], imx15, t, p);
    GEN imx14p1 = FPM_FPXQ_MUL(frob_pows[1], imx14, t, p);
    GEN imx12p1 = FPM_FPXQ_MUL(frob_pows[1], imx12, t, p);
    GEN imx2p1  = FPM_FPXQ_MUL(frob_pows[1], imx2,  t, p);
    GEN imx16p1 = FPM_FPXQ_MUL(frob_pows[1], imx16, t, p);
    GEN imx0p1  = FPM_FPXQ_MUL(frob_pows[1], imx0,  t, p);

    GEN mx24p2  = FPM_FPXQ_MUL(frob_pows[2], mx24,  t, p);
    GEN mx23p2  = FPM_FPXQ_MUL(frob_pows[2], mx23,  t, p);
    GEN mx15p2  = FPM_FPXQ_MUL(frob_pows[2], mx15,  t, p);
    GEN mx14p2  = FPM_FPXQ_MUL(frob_pows[2], mx14,  t, p);
    GEN mx13p2  = FPM_FPXQ_MUL(frob_pows[2], mx13,  t, p);
    GEN mx2p2   = FPM_FPXQ_MUL(frob_pows[2], mx2,   t, p);
    GEN mx11p2  = FPM_FPXQ_MUL(frob_pows[2], mx11,  t, p);
    GEN mx16p2  = FPM_FPXQ_MUL(frob_pows[2], mx16,  t, p);
    GEN mx25p2  = FPM_FPXQ_MUL(frob_pows[2], mx25,  t, p);
    GEN imx10p2 = FPM_FPXQ_MUL(frob_pows[2], imx10, t, p);
    GEN imx1p2  = FPM_FPXQ_MUL(frob_pows[2], imx1,  t, p);
    GEN imx0p2  = FPM_FPXQ_MUL(frob_pows[2], imx0,  t, p);

    GEN mx10p3  = FPM_FPXQ_MUL(frob_pows[3], mx10,  t, p);
    GEN mx9p3   = FPM_FPXQ_MUL(frob_pows[3], mx9,   t, p);
    GEN mx1p3   = FPM_FPXQ_MUL(frob_pows[3], mx1,   t, p);
    GEN mx0p3   = FPM_FPXQ_MUL(frob_pows[3], mx0,   t, p);
    GEN imx24p3 = FPM_FPXQ_MUL(frob_pows[3], imx24, t, p);
    GEN imx23p3 = FPM_FPXQ_MUL(frob_pows[3], imx23, t, p);
    GEN imx15p3 = FPM_FPXQ_MUL(frob_pows[3], imx15, t, p);
    GEN imx14p3 = FPM_FPXQ_MUL(frob_pows[3], imx14, t, p);
    GEN imx2p3  = FPM_FPXQ_MUL(frob_pows[3], imx2,  t, p);
    GEN imx11p3 = FPM_FPXQ_MUL(frob_pows[3], imx11, t, p);
    GEN imx16p3 = FPM_FPXQ_MUL(frob_pows[3], imx16, t, p);
    GEN imx25p3 = FPM_FPXQ_MUL(frob_pows[3], imx25, t, p);

    GEN mx24p4  = FPM_FPXQ_MUL(frob_pows[4], mx24,  t, p);
    GEN mx15p4  = FPM_FPXQ_MUL(frob_pows[4], mx15,  t, p);
    GEN mx14p4  = FPM_FPXQ_MUL(frob_pows[4], mx14,  t, p);
    GEN mx10p4  = FPM_FPXQ_MUL(frob_pows[4], mx10,  t, p);
    GEN mx2p4   = FPM_FPXQ_MUL(frob_pows[4], mx2,   t, p);
    GEN mx16p4  = FPM_FPXQ_MUL(frob_pows[4], mx16,  t, p);
    GEN imx21p4 = FPM_FPXQ_MUL(frob_pows[4], imx21, t, p);
    GEN imx9p4  = FPM_FPXQ_MUL(frob_pows[4], imx9,  t, p);
    GEN imx11p4 = FPM_FPXQ_MUL(frob_pows[4], imx11, t, p);
    GEN imx1p4  = FPM_FPXQ_MUL(frob_pows[4], imx1,  t, p);
    GEN imx0p4  = FPM_FPXQ_MUL(frob_pows[4], imx0,  t, p);

    GEN mx21p5  = FPM_FPXQ_MUL(frob_pows[5], mx21,  t, p);
    GEN mx20p5  = FPM_FPXQ_MUL(frob_pows[5], mx20,  t, p);
    GEN mx10p5  = FPM_FPXQ_MUL(frob_pows[5], mx10,  t, p);
    GEN mx8p5   = FPM_FPXQ_MUL(frob_pows[5], mx8,   t, p);
    GEN mx22p5  = FPM_FPXQ_MUL(frob_pows[5], mx22,  t, p);
    GEN mx1p5   = FPM_FPXQ_MUL(frob_pows[5], mx1,   t, p);
    GEN mx0p5   = FPM_FPXQ_MUL(frob_pows[5], mx0,   t, p);
    GEN imx15p5 = FPM_FPXQ_MUL(frob_pows[5], imx15, t, p);
    GEN imx14p5 = FPM_FPXQ_MUL(frob_pows[5], imx14, t, p);
    GEN imx7p5  = FPM_FPXQ_MUL(frob_pows[5], imx7,  t, p);
    GEN imx2p5  = FPM_FPXQ_MUL(frob_pows[5], imx2,  t, p);
    GEN imx16p5 = FPM_FPXQ_MUL(frob_pows[5], imx16, t, p);

    GEN mx15p6  = FPM_FPXQ_MUL(frob_pows[6], mx15,  t, p);
    GEN mx14p6  = FPM_FPXQ_MUL(frob_pows[6], mx14,  t, p);
    GEN mx7p6   = FPM_FPXQ_MUL(frob_pows[6], mx7,   t, p);
    GEN mx6p6   = FPM_FPXQ_MUL(frob_pows[6], mx6,   t, p);
    GEN mx2p6   = FPM_FPXQ_MUL(frob_pows[6], mx2,   t, p);
    GEN mx16p6  = FPM_FPXQ_MUL(frob_pows[6], mx16,  t, p);
    GEN imx21p6 = FPM_FPXQ_MUL(frob_pows[6], imx21, t, p);
    GEN imx20p6 = FPM_FPXQ_MUL(frob_pows[6], imx20, t, p);
    GEN imx8p6  = FPM_FPXQ_MUL(frob_pows[6], imx8,  t, p);
    GEN imx22p6 = FPM_FPXQ_MUL(frob_pows[6], imx22, t, p);
    GEN imx1p6  = FPM_FPXQ_MUL(frob_pows[6], imx1,  t, p);
    GEN imx0p6  = FPM_FPXQ_MUL(frob_pows[6], imx0,  t, p);

    GEN mx21p7  = FPM_FPXQ_MUL(frob_pows[7], mx21,  t, p);
    GEN mx7p7   = FPM_FPXQ_MUL(frob_pows[7], mx7,   t, p);
    GEN mx1p7   = FPM_FPXQ_MUL(frob_pows[7], mx1,   t, p);
    GEN mx0p7   = FPM_FPXQ_MUL(frob_pows[7], mx0,   t, p);
    GEN imx18p7 = FPM_FPXQ_MUL(frob_pows[7], imx18, t, p);
    GEN imx15p7 = FPM_FPXQ_MUL(frob_pows[7], imx15, t, p);
    GEN imx14p7 = FPM_FPXQ_MUL(frob_pows[7], imx14, t, p);
    GEN imx6p7  = FPM_FPXQ_MUL(frob_pows[7], imx6,  t, p);
    GEN imx2p7  = FPM_FPXQ_MUL(frob_pows[7], imx2,  t, p);
    GEN imx8p7  = FPM_FPXQ_MUL(frob_pows[7], imx8,  t, p);
    GEN imx16p7 = FPM_FPXQ_MUL(frob_pows[7], imx16, t, p);

    GEN mx18p8 = FPM_FPXQ_MUL(frob_pows[8], mx18,  t, p);
    GEN mx17p8 = FPM_FPXQ_MUL(frob_pows[8], mx17,  t, p);
    GEN mx15p8 = FPM_FPXQ_MUL(frob_pows[8], mx15,  t, p);
    GEN mx14p8 = FPM_FPXQ_MUL(frob_pows[8], mx14,  t, p);
    GEN mx7p8  = FPM_FPXQ_MUL(frob_pows[8], mx7,   t, p);
    GEN mx2p8  = FPM_FPXQ_MUL(frob_pows[8], mx2,   t, p);
    GEN mx5p8  = FPM_FPXQ_MUL(frob_pows[8], mx5,   t, p);
    GEN mx16p8 = FPM_FPXQ_MUL(frob_pows[8], mx16,  t, p);
    GEN mx19p8 = FPM_FPXQ_MUL(frob_pows[8], mx19,  t, p);
    GEN imx4p8 = FPM_FPXQ_MUL(frob_pows[8], imx4,  t, p);
    GEN imx1p8 = FPM_FPXQ_MUL(frob_pows[8], imx1,  t, p);
    GEN imx0p8 = FPM_FPXQ_MUL(frob_pows[8], imx0,  t, p);

    GEN mx4p9   = FPM_FPXQ_MUL(frob_pows[9], mx4,   t, p);
    GEN mx3p9   = FPM_FPXQ_MUL(frob_pows[9], mx3,   t, p);
    GEN mx1p9   = FPM_FPXQ_MUL(frob_pows[9], mx1,   t, p);
    GEN mx0p9   = FPM_FPXQ_MUL(frob_pows[9], mx0,   t, p);
    GEN imx18p9 = FPM_FPXQ_MUL(frob_pows[9], imx18, t, p);
    GEN imx17p9 = FPM_FPXQ_MUL(frob_pows[9], imx17, t, p);
    GEN imx15p9 = FPM_FPXQ_MUL(frob_pows[9], imx15, t, p);
    GEN imx14p9 = FPM_FPXQ_MUL(frob_pows[9], imx14, t, p);
    GEN imx2p9  = FPM_FPXQ_MUL(frob_pows[9], imx2,  t, p);
    GEN imx5p9  = FPM_FPXQ_MUL(frob_pows[9], imx5,  t, p);
    GEN imx16p9 = FPM_FPXQ_MUL(frob_pows[9], imx16, t, p);
    GEN imx19p9 = FPM_FPXQ_MUL(frob_pows[9], imx19, t, p);

    GEN mx18p10 = FPM_FPXQ_MUL(frob_pows[10], mx18, t, p);
    GEN mx14p10 = FPM_FPXQ_MUL(frob_pows[10], mx14, t, p);
    GEN mx4p10  = FPM_FPXQ_MUL(frob_pows[10], mx4,  t, p);
    GEN mx2p10  = FPM_FPXQ_MUL(frob_pows[10], mx2,  t, p);
    GEN mx16p10 = FPM_FPXQ_MUL(frob_pows[10], mx16, t, p);
    GEN imx3p10 = FPM_FPXQ_MUL(frob_pows[10], imx3, t, p);
    GEN imx5p10 = FPM_FPXQ_MUL(frob_pows[10], imx5, t, p);
    GEN imx1p10 = FPM_FPXQ_MUL(frob_pows[10], imx1, t, p);
    GEN imx0p10 = FPM_FPXQ_MUL(frob_pows[10], imx0, t, p);

    GEN mx4p11 = FPM_FPXQ_MUL(frob_pows[11], mx4, t, p);
    GEN mx2p11 = FPM_FPXQ_MUL(frob_pows[11], mx2, t, p);
    GEN mx0p11 = FPM_FPXQ_MUL(frob_pows[11], mx0, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx14, mx12,  t, p);
    y0 = FPXQ_MUL(y0, mx16,    t, p);
    y0 = FPXQ_MUL(y0, mx27p1,  t, p);
    y0 = FPXQ_MUL(y0, mx1p1,   t, p);
    y0 = FPXQ_MUL(y0, imx24p1, t, p);
    y0 = FPXQ_MUL(y0, imx15p1, t, p);
    y0 = FPXQ_MUL(y0, imx16p1, t, p);
    y0 = FPXQ_MUL(y0, mx24p2,  t, p);
    y0 = FPXQ_MUL(y0, mx23p2,  t, p);
    y0 = FPXQ_MUL(y0, mx15p2,  t, p);
    y0 = FPXQ_MUL(y0, mx14p2,  t, p);
    y0 = FPXQ_MUL(y0, mx13p2,  t, p);
    y0 = FPXQ_MUL(y0, mx16p2,  t, p);
    y0 = FPXQ_MUL(y0, mx25p2,  t, p);
    y0 = FPXQ_MUL(y0, imx10p2, t, p);
    y0 = FPXQ_MUL(y0, imx1p2,  t, p);
    y0 = FPXQ_MUL(y0, imx0p2,  t, p);
    y0 = FPXQ_MUL(y0, mx10p3,  t, p);
    y0 = FPXQ_MUL(y0, mx9p3,   t, p);
    y0 = FPXQ_MUL(y0, mx1p3,   t, p);
    y0 = FPXQ_MUL(y0, mx0p3,   t, p);
    y0 = FPXQ_MUL(y0, imx24p3, t, p);
    y0 = FPXQ_MUL(y0, imx23p3, t, p);
    y0 = FPXQ_MUL(y0, imx15p3, t, p);
    y0 = FPXQ_MUL(y0, imx14p3, t, p);
    y0 = FPXQ_MUL(y0, imx16p3, t, p);
    y0 = FPXQ_MUL(y0, imx25p3, t, p);
    y0 = FPXQ_MUL(y0, mx24p4,  t, p);
    y0 = FPXQ_MUL(y0, mx15p4,  t, p);
    y0 = FPXQ_MUL(y0, mx14p4,  t, p);
    y0 = FPXQ_MUL(y0, mx16p4,  t, p);
    y0 = FPXQ_MUL(y0, imx21p4, t, p);
    y0 = FPXQ_MUL(y0, imx11p4, t, p);
    y0 = FPXQ_MUL(y0, imx1p4,  t, p);
    y0 = FPXQ_MUL(y0, imx0p4,  t, p);
    y0 = FPXQ_MUL(y0, mx21p5,  t, p);
    y0 = FPXQ_MUL(y0, mx20p5,  t, p);
    y0 = FPXQ_MUL(y0, mx10p5,  t, p);
    y0 = FPXQ_MUL(y0, mx22p5,  t, p);
    y0 = FPXQ_MUL(y0, mx1p5,   t, p);
    y0 = FPXQ_MUL(y0, mx0p5,   t, p);
    y0 = FPXQ_MUL(y0, imx15p5, t, p);
    y0 = FPXQ_MUL(y0, imx14p5, t, p);
    y0 = FPXQ_MUL(y0, imx7p5,  t, p);
    y0 = FPXQ_MUL(y0, imx16p5, t, p);
    y0 = FPXQ_MUL(y0, mx15p6,  t, p);
    y0 = FPXQ_MUL(y0, mx14p6,  t, p);
    y0 = FPXQ_MUL(y0, mx7p6,   t, p);
    y0 = FPXQ_MUL(y0, mx6p6,   t, p);
    y0 = FPXQ_MUL(y0, mx16p6,  t, p);
    y0 = FPXQ_MUL(y0, imx21p6, t, p);
    y0 = FPXQ_MUL(y0, imx20p6, t, p);
    y0 = FPXQ_MUL(y0, imx22p6, t, p);
    y0 = FPXQ_MUL(y0, imx1p6,  t, p);
    y0 = FPXQ_MUL(y0, imx0p6,  t, p);
    y0 = FPXQ_MUL(y0, mx21p7,  t, p);
    y0 = FPXQ_MUL(y0, mx1p7,   t, p);
    y0 = FPXQ_MUL(y0, mx0p7,   t, p);
    y0 = FPXQ_MUL(y0, imx18p7, t, p);
    y0 = FPXQ_MUL(y0, imx15p7, t, p);
    y0 = FPXQ_MUL(y0, imx14p7, t, p);
    y0 = FPXQ_MUL(y0, imx8p7,  t, p);
    y0 = FPXQ_MUL(y0, imx16p7, t, p);
    y0 = FPXQ_MUL(y0, mx18p8,  t, p);
    y0 = FPXQ_MUL(y0, mx17p8,  t, p);
    y0 = FPXQ_MUL(y0, mx15p8,  t, p);
    y0 = FPXQ_MUL(y0, mx14p8,  t, p);
    y0 = FPXQ_MUL(y0, mx7p8,   t, p);
    y0 = FPXQ_MUL(y0, mx16p8,  t, p);
    y0 = FPXQ_MUL(y0, mx19p8,  t, p);
    y0 = FPXQ_MUL(y0, imx4p8,  t, p);
    y0 = FPXQ_MUL(y0, imx1p8,  t, p);
    y0 = FPXQ_MUL(y0, imx0p8,  t, p);
    y0 = FPXQ_MUL(y0, mx4p9,   t, p);
    y0 = FPXQ_MUL(y0, mx3p9,   t, p);
    y0 = FPXQ_MUL(y0, mx1p9,   t, p);
    y0 = FPXQ_MUL(y0, mx0p9,   t, p);
    y0 = FPXQ_MUL(y0, imx18p9, t, p);
    y0 = FPXQ_MUL(y0, imx17p9, t, p);
    y0 = FPXQ_MUL(y0, imx15p9, t, p);
    y0 = FPXQ_MUL(y0, imx14p9, t, p);
    y0 = FPXQ_MUL(y0, imx16p9, t, p);
    y0 = FPXQ_MUL(y0, imx19p9, t, p);
    y0 = FPXQ_MUL(y0, mx18p10, t, p);
    y0 = FPXQ_MUL(y0, mx14p10, t, p);
    y0 = FPXQ_MUL(y0, mx16p10, t, p);
    y0 = FPXQ_MUL(y0, imx5p10, t, p);
    y0 = FPXQ_MUL(y0, imx1p10, t, p);
    y0 = FPXQ_MUL(y0, imx0p10, t, p);
    y0 = FPXQ_MUL(y0, mx4p11,  t, p);
    y0 = FPXQ_MUL(y0, mx2p11,  t, p);
    y0 = FPXQ_MUL(y0, mx0p11,  t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx13p1, imx14p1, t, p);
    y1 = FPXQ_MUL(y1, imx2p1,  t, p);
    y1 = FPXQ_MUL(y1, imx0p1,  t, p);
    y1 = FPXQ_MUL(y1, mx2p2,   t, p);
    y1 = FPXQ_MUL(y1, imx2p3,  t, p);
    y1 = FPXQ_MUL(y1, imx11p3, t, p);
    y1 = FPXQ_MUL(y1, mx10p4,  t, p);
    y1 = FPXQ_MUL(y1, mx2p4,   t, p);
    y1 = FPXQ_MUL(y1, imx2p5,  t, p);
    y1 = FPXQ_MUL(y1, mx2p6,   t, p);
    y1 = FPXQ_MUL(y1, imx8p6,  t, p);
    y1 = FPXQ_MUL(y1, mx7p7,   t, p);
    y1 = FPXQ_MUL(y1, imx2p7,  t, p);
    y1 = FPXQ_MUL(y1, mx2p8,   t, p);
    y1 = FPXQ_MUL(y1, imx2p9,  t, p);
    y1 = FPXQ_MUL(y1, imx5p9,  t, p);
    y1 = FPXQ_MUL(y1, mx4p10,  t, p);
    y1 = FPXQ_MUL(y1, mx2p10,  t, p);

    GEN y2;
    y2 = FPXQ_MUL(mx2, mx0,   t, p);
    y2 = FPXQ_MUL(y2, imx1,   t, p);
    y2 = FPXQ_MUL(y2, mx11p2, t, p);
    y2 = FPXQ_MUL(y2, mx8p5,  t, p);
    y2 = FPXQ_MUL(y2, mx5p8,  t, p);

    GEN y3;
    y3 = FPXQ_MUL(imx12p1, imx9p4, t, p);
    y3 = FPXQ_MUL(y3, imx6p7,  t, p);
    y3 = FPXQ_MUL(y3, imx3p10, t, p);
    
    GEN T0 = FPXQ_MUL(y3, y2, t, p);
    GEN T1 = FPXQ_MUL(y3, y0, t, p);
    GEN T2 = FPXQ_MUL(T0, y1, t, p);
    GEN T3 = FPXQ_MUL(T1, T0, t, p);
    GEN T4 = FPXQ_SQR(T2, t, p);
    GEN T5 = FPXQ_MUL(T4, T3, t, p);
    
    return T5;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_27_1111(pairing_data_t* const pdata, GEN unreduced)
{    
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mx0  = m;
    GEN mx1  = FPXQ_POW(mx0,  x, t, p);
    GEN mx2  = FPXQ_POW(mx1,  x, t, p);
    GEN mx3  = FPXQ_POW(mx2,  x, t, p);
    GEN mx4  = FPXQ_POW(mx3,  x, t, p);
    GEN mx5  = FPXQ_POW(mx4,  x, t, p);
    GEN mx6  = FPXQ_POW(mx5,  x, t, p);
    GEN mx7  = FPXQ_POW(mx6,  x, t, p);
    GEN mx8  = FPXQ_POW(mx7,  x, t, p);
    GEN mx9  = FPXQ_POW(mx8,  x, t, p);
    GEN mx10 = FPXQ_POW(mx9,  x, t, p);
    GEN mx11 = FPXQ_POW(mx10, x, t, p);
    GEN mx12 = FPXQ_POW(mx11, x, t, p);
    GEN mx13 = FPXQ_POW(mx12, x, t, p);
    GEN mx14 = FPXQ_POW(mx13, x, t, p);
    GEN mx15 = FPXQ_POW(mx14, x, t, p);
    GEN mx16 = FPXQ_POW(mx15, x, t, p);
    GEN mx17 = FPXQ_POW(mx16, x, t, p);
    GEN mx18 = FPXQ_POW(mx17, x, t, p);
    GEN mx19 = FPXQ_POW(mx18, x, t, p);

    GEN imx1  = FPXQ_INV(mx1,  t, p);
    GEN imx2  = FPXQ_INV(mx2,  t, p);
    GEN imx3  = FPXQ_INV(mx3,  t, p);
    GEN imx4  = FPXQ_INV(mx4,  t, p);
    GEN imx5  = FPXQ_INV(mx5,  t, p);
    GEN imx6  = FPXQ_INV(mx6,  t, p);
    GEN imx7  = FPXQ_INV(mx7,  t, p);
    GEN imx8  = FPXQ_INV(mx8,  t, p);
    GEN imx9  = FPXQ_INV(mx9,  t, p);
    GEN imx10 = FPXQ_INV(mx10, t, p);
    GEN imx11 = FPXQ_INV(mx11, t, p);
    GEN imx12 = FPXQ_INV(mx12, t, p);
    GEN imx13 = FPXQ_INV(mx13, t, p);
    GEN imx14 = FPXQ_INV(mx14, t, p);
    GEN imx15 = FPXQ_INV(mx15, t, p);
    GEN imx16 = FPXQ_INV(mx16, t, p);
    GEN imx17 = FPXQ_INV(mx17, t, p);
    GEN imx18 = FPXQ_INV(mx18, t, p);
    GEN imx19 = FPXQ_INV(mx19, t, p);

    GEN mx8p1   = FPM_FPXQ_MUL(frob_pows[1], mx8,   t, p);
    GEN mx17p1  = FPM_FPXQ_MUL(frob_pows[1], mx17,  t, p);
    GEN imx16p1 = FPM_FPXQ_MUL(frob_pows[1], imx16, t, p);
    GEN imx18p1 = FPM_FPXQ_MUL(frob_pows[1], imx18, t, p);

    GEN mx8p2   = FPM_FPXQ_MUL(frob_pows[2], mx8,   t, p);
    GEN mx7p2   = FPM_FPXQ_MUL(frob_pows[2], mx7,   t, p);
    GEN mx6p2   = FPM_FPXQ_MUL(frob_pows[2], mx6,   t, p);
    GEN mx16p2  = FPM_FPXQ_MUL(frob_pows[2], mx16,  t, p);
    GEN imx15p2 = FPM_FPXQ_MUL(frob_pows[2], imx15, t, p);
    GEN imx17p2 = FPM_FPXQ_MUL(frob_pows[2], imx17, t, p);

    GEN mx7p3  = FPM_FPXQ_MUL(frob_pows[3], mx7,  t, p);
    GEN mx5p3  = FPM_FPXQ_MUL(frob_pows[3], mx5,  t, p);
    GEN imx6p3 = FPM_FPXQ_MUL(frob_pows[3], imx6, t, p);

    GEN mx15p4  = FPM_FPXQ_MUL(frob_pows[4], mx15,  t, p);
    GEN mx13p4  = FPM_FPXQ_MUL(frob_pows[4], mx13,  t, p);
    GEN imx5p4  = FPM_FPXQ_MUL(frob_pows[4], imx5,  t, p);
    GEN imx14p4 = FPM_FPXQ_MUL(frob_pows[4], imx14, t, p);

    GEN mx12p5  = FPM_FPXQ_MUL(frob_pows[5], mx12,  t, p);
    GEN mx14p5  = FPM_FPXQ_MUL(frob_pows[5], mx14,  t, p);
    GEN imx5p5  = FPM_FPXQ_MUL(frob_pows[5], imx5,  t, p);
    GEN imx13p5 = FPM_FPXQ_MUL(frob_pows[5], imx13, t, p);
    GEN imx4p5  = FPM_FPXQ_MUL(frob_pows[5], imx4,  t, p);
    GEN imx3p5  = FPM_FPXQ_MUL(frob_pows[5], imx3,  t, p);

    GEN mx3p6  = FPM_FPXQ_MUL(frob_pows[6], mx3,  t, p);
    GEN imx4p6 = FPM_FPXQ_MUL(frob_pows[6], imx4, t, p);
    GEN imx2p6 = FPM_FPXQ_MUL(frob_pows[6], imx2, t, p);

    GEN mx11p7  = FPM_FPXQ_MUL(frob_pows[7], mx11,  t, p);
    GEN mx2p7   = FPM_FPXQ_MUL(frob_pows[7], mx2,   t, p);
    GEN imx12p7 = FPM_FPXQ_MUL(frob_pows[7], imx12, t, p);
    GEN imx10p7 = FPM_FPXQ_MUL(frob_pows[7], imx10, t, p);

    GEN mx10p8  = FPM_FPXQ_MUL(frob_pows[8], mx10,  t, p);
    GEN mx2p8   = FPM_FPXQ_MUL(frob_pows[8], mx2,   t, p);
    GEN mx1p8   = FPM_FPXQ_MUL(frob_pows[8], mx1,   t, p);
    GEN mx0p8   = FPM_FPXQ_MUL(frob_pows[8], mx0,   t, p);
    GEN imx11p8 = FPM_FPXQ_MUL(frob_pows[8], imx11, t, p);
    GEN imx9p8  = FPM_FPXQ_MUL(frob_pows[8], imx9,  t, p);

    GEN mx9p9   = FPM_FPXQ_MUL(frob_pows[9], mx9,   t, p);
    GEN mx18p9  = FPM_FPXQ_MUL(frob_pows[9], mx18,  t, p);
    GEN imx17p9 = FPM_FPXQ_MUL(frob_pows[9], imx17, t, p);
    GEN imx19p9 = FPM_FPXQ_MUL(frob_pows[9], imx19, t, p);

    GEN mx8p10   = FPM_FPXQ_MUL(frob_pows[10], mx8,   t, p);
    GEN mx7p10   = FPM_FPXQ_MUL(frob_pows[10], mx7,   t, p);
    GEN mx9p10   = FPM_FPXQ_MUL(frob_pows[10], mx9,   t, p);
    GEN mx17p10  = FPM_FPXQ_MUL(frob_pows[10], mx17,  t, p);
    GEN imx16p10 = FPM_FPXQ_MUL(frob_pows[10], imx16, t, p);
    GEN imx18p10 = FPM_FPXQ_MUL(frob_pows[10], imx18, t, p);

    GEN mx8p11  = FPM_FPXQ_MUL(frob_pows[11], mx8,  t, p);
    GEN mx6p11  = FPM_FPXQ_MUL(frob_pows[11], mx6,  t, p);
    GEN imx7p11 = FPM_FPXQ_MUL(frob_pows[11], imx7, t, p);

    GEN mx16p12  = FPM_FPXQ_MUL(frob_pows[12], mx16,  t, p);
    GEN mx14p12  = FPM_FPXQ_MUL(frob_pows[12], mx14,  t, p);
    GEN imx6p12  = FPM_FPXQ_MUL(frob_pows[12], imx6,  t, p);
    GEN imx15p12 = FPM_FPXQ_MUL(frob_pows[12], imx15, t, p);

    GEN mx15p13  = FPM_FPXQ_MUL(frob_pows[13], mx15,  t, p);
    GEN mx13p13  = FPM_FPXQ_MUL(frob_pows[13], mx13,  t, p);
    GEN imx6p13  = FPM_FPXQ_MUL(frob_pows[13], imx6,  t, p);
    GEN imx5p13  = FPM_FPXQ_MUL(frob_pows[13], imx5,  t, p);
    GEN imx14p13 = FPM_FPXQ_MUL(frob_pows[13], imx14, t, p);
    GEN imx4p13  = FPM_FPXQ_MUL(frob_pows[13], imx4,  t, p);

    GEN mx4p14  = FPM_FPXQ_MUL(frob_pows[14], mx4,  t, p);
    GEN imx5p14 = FPM_FPXQ_MUL(frob_pows[14], imx5, t, p);
    GEN imx3p14 = FPM_FPXQ_MUL(frob_pows[14], imx3, t, p);

    GEN mx12p15  = FPM_FPXQ_MUL(frob_pows[15], mx12,  t, p);
    GEN mx3p15   = FPM_FPXQ_MUL(frob_pows[15], mx3,   t, p);
    GEN imx11p15 = FPM_FPXQ_MUL(frob_pows[15], imx11, t, p);
    GEN imx13p15 = FPM_FPXQ_MUL(frob_pows[15], imx13, t, p);

    GEN mx11p16  = FPM_FPXQ_MUL(frob_pows[16], mx11,  t, p);
    GEN mx3p16   = FPM_FPXQ_MUL(frob_pows[16], mx3,   t, p);
    GEN mx2p16   = FPM_FPXQ_MUL(frob_pows[16], mx2,   t, p);
    GEN mx1p16   = FPM_FPXQ_MUL(frob_pows[16], mx1,   t, p);
    GEN imx12p16 = FPM_FPXQ_MUL(frob_pows[16], imx12, t, p);
    GEN imx10p16 = FPM_FPXQ_MUL(frob_pows[16], imx10, t, p);

    GEN mx2p17  = FPM_FPXQ_MUL(frob_pows[17], mx2,  t, p);
    GEN mx0p17  = FPM_FPXQ_MUL(frob_pows[17], mx0,  t, p);
    GEN imx1p17 = FPM_FPXQ_MUL(frob_pows[17], imx1, t, p);

    GEN y0;
    y0 = FPXQ_MUL(mx9, imx8,    t, p);
    y0 = FPXQ_MUL(y0, imx10,    t, p);
    y0 = FPXQ_MUL(y0, mx17p1,   t, p);
    y0 = FPXQ_MUL(y0, imx16p1,  t, p);
    y0 = FPXQ_MUL(y0, imx18p1,  t, p);
    y0 = FPXQ_MUL(y0, mx8p2,    t, p);
    y0 = FPXQ_MUL(y0, mx6p2,    t, p);
    y0 = FPXQ_MUL(y0, mx16p2,   t, p);
    y0 = FPXQ_MUL(y0, imx15p2,  t, p);
    y0 = FPXQ_MUL(y0, imx17p2,  t, p);
    y0 = FPXQ_MUL(y0, mx7p3,    t, p);
    y0 = FPXQ_MUL(y0, mx5p3,    t, p);
    y0 = FPXQ_MUL(y0, imx6p3,   t, p);
    y0 = FPXQ_MUL(y0, mx15p4,   t, p);
    y0 = FPXQ_MUL(y0, mx13p4,   t, p);
    y0 = FPXQ_MUL(y0, imx14p4,  t, p);
    y0 = FPXQ_MUL(y0, mx12p5,   t, p);
    y0 = FPXQ_MUL(y0, mx14p5,   t, p);
    y0 = FPXQ_MUL(y0, imx5p5,   t, p);
    y0 = FPXQ_MUL(y0, imx13p5,  t, p);
    y0 = FPXQ_MUL(y0, imx3p5,   t, p);
    y0 = FPXQ_MUL(y0, mx3p6,    t, p);
    y0 = FPXQ_MUL(y0, imx4p6,   t, p);
    y0 = FPXQ_MUL(y0, imx2p6,   t, p);
    y0 = FPXQ_MUL(y0, mx11p7,   t, p);
    y0 = FPXQ_MUL(y0, imx12p7,  t, p);
    y0 = FPXQ_MUL(y0, imx10p7,  t, p);
    y0 = FPXQ_MUL(y0, mx10p8,   t, p);
    y0 = FPXQ_MUL(y0, mx2p8,    t, p);
    y0 = FPXQ_MUL(y0, mx0p8,    t, p);
    y0 = FPXQ_MUL(y0, imx11p8,  t, p);
    y0 = FPXQ_MUL(y0, imx9p8,   t, p);
    y0 = FPXQ_MUL(y0, mx18p9,   t, p);
    y0 = FPXQ_MUL(y0, imx17p9,  t, p);
    y0 = FPXQ_MUL(y0, imx19p9,  t, p);
    y0 = FPXQ_MUL(y0, mx7p10,   t, p);
    y0 = FPXQ_MUL(y0, mx9p10,   t, p);
    y0 = FPXQ_MUL(y0, mx17p10,  t, p);
    y0 = FPXQ_MUL(y0, imx16p10, t, p);
    y0 = FPXQ_MUL(y0, imx18p10, t, p);
    y0 = FPXQ_MUL(y0, mx8p11,   t, p);
    y0 = FPXQ_MUL(y0, mx6p11,   t, p);
    y0 = FPXQ_MUL(y0, imx7p11,  t, p);
    y0 = FPXQ_MUL(y0, mx16p12,  t, p);
    y0 = FPXQ_MUL(y0, mx14p12,  t, p);
    y0 = FPXQ_MUL(y0, imx15p12, t, p);
    y0 = FPXQ_MUL(y0, mx15p13,  t, p);
    y0 = FPXQ_MUL(y0, mx13p13,  t, p);
    y0 = FPXQ_MUL(y0, imx6p13,  t, p);
    y0 = FPXQ_MUL(y0, imx14p13, t, p);
    y0 = FPXQ_MUL(y0, imx4p13,  t, p);
    y0 = FPXQ_MUL(y0, mx4p14,   t, p);
    y0 = FPXQ_MUL(y0, imx5p14,  t, p);
    y0 = FPXQ_MUL(y0, imx3p14,  t, p);
    y0 = FPXQ_MUL(y0, mx12p15,  t, p);
    y0 = FPXQ_MUL(y0, imx11p15, t, p);
    y0 = FPXQ_MUL(y0, imx13p15, t, p);
    y0 = FPXQ_MUL(y0, mx11p16,  t, p);
    y0 = FPXQ_MUL(y0, mx3p16,   t, p);
    y0 = FPXQ_MUL(y0, mx1p16,   t, p);
    y0 = FPXQ_MUL(y0, imx12p16, t, p);
    y0 = FPXQ_MUL(y0, imx10p16, t, p);
    y0 = FPXQ_MUL(y0, mx2p17,   t, p);
    y0 = FPXQ_MUL(y0, mx0p17,   t, p);
    y0 = FPXQ_MUL(y0, imx1p17,  t, p);

    GEN y1;
    y1 = FPXQ_MUL(mx7p2, imx4p5, t, p);
    y1 = FPXQ_MUL(y1, mx1p8,     t, p);
    y1 = FPXQ_MUL(y1, mx8p10,    t, p);
    y1 = FPXQ_MUL(y1, imx5p13,   t, p);
    y1 = FPXQ_MUL(y1, mx2p16,    t, p);

    GEN y2;
    y2 = FPXQ_MUL(mx0, mx8p1,  t, p);
    y2 = FPXQ_MUL(y2, imx5p4,  t, p);
    y2 = FPXQ_MUL(y2, mx2p7,   t, p);
    y2 = FPXQ_MUL(y2, mx9p9,   t, p);
    y2 = FPXQ_MUL(y2, imx6p12, t, p);
    y2 = FPXQ_MUL(y2, mx3p15,  t, p);

    GEN T0 = FPXQ_MUL(y2, y1, t, p);
    GEN T1 = FPXQ_MUL(y2, y0, t, p);
    GEN T2 = FPXQ_SQR(T0, t, p);
    GEN T3 = FPXQ_MUL(T2, T1, t, p);
    
    OUT();
    return T3;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_opti_24_1250(pairing_data_t* const pdata, GEN unreduced)
{    
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    GEN x = pdata->u;
    
    GEN* frob_pows = pdata->frobenius_powers;
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }
    
    GEN mx1 = FPXQ_POW(m,   x, t, p);
    GEN mx2 = FPXQ_POW(mx1, x, t, p);
    GEN mx3 = FPXQ_POW(mx2, x, t, p);
    GEN mx4 = FPXQ_POW(mx3, x, t, p);
    GEN mx5 = FPXQ_POW(mx4, x, t, p);
    GEN mx6 = FPXQ_POW(mx5, x, t, p);
    GEN mx7 = FPXQ_POW(mx6, x, t, p);
    GEN mx8 = FPXQ_POW(mx7, x, t, p);
    GEN mx9 = FPXQ_POW(mx8, x, t, p);
    
    GEN im   = FPXQ_INV(m,   t, p);
    GEN imx1 = FPXQ_INV(mx1, t, p);
    GEN imx2 = FPXQ_INV(mx2, t, p);
    GEN imx3 = FPXQ_INV(mx3, t, p);
    GEN imx4 = FPXQ_INV(mx4, t, p);
    GEN imx5 = FPXQ_INV(mx5, t, p);
    GEN imx6 = FPXQ_INV(mx6, t, p);
    GEN imx7 = FPXQ_INV(mx7, t, p);
    GEN imx8 = FPXQ_INV(mx8, t, p);
    
    GEN mx3p1  = FPM_FPXQ_MUL(frob_pows[1], mx3,  t, p);
    GEN mx6p1  = FPM_FPXQ_MUL(frob_pows[1], mx6,  t, p);
    GEN mx8p1  = FPM_FPXQ_MUL(frob_pows[1], mx8,  t, p);
    GEN imx2p1 = FPM_FPXQ_MUL(frob_pows[1], imx2, t, p);
    GEN imx4p1 = FPM_FPXQ_MUL(frob_pows[1], imx4, t, p);
    GEN imx7p1 = FPM_FPXQ_MUL(frob_pows[1], imx7, t, p);
    
    GEN mx2p2  = FPM_FPXQ_MUL(frob_pows[2], mx2,  t, p);
    GEN mx5p2  = FPM_FPXQ_MUL(frob_pows[2], mx5,  t, p);
    GEN mx7p2  = FPM_FPXQ_MUL(frob_pows[2], mx7,  t, p);
    GEN imx1p2 = FPM_FPXQ_MUL(frob_pows[2], imx1, t, p);
    GEN imx3p2 = FPM_FPXQ_MUL(frob_pows[2], imx3, t, p);
    GEN imx6p2 = FPM_FPXQ_MUL(frob_pows[2], imx6, t, p);
    
    GEN mx1p3  = FPM_FPXQ_MUL(frob_pows[3], mx1,  t, p);
    GEN mx4p3  = FPM_FPXQ_MUL(frob_pows[3], mx4,  t, p);
    GEN mx6p3  = FPM_FPXQ_MUL(frob_pows[3], mx6,  t, p);
    GEN imp3   = FPM_FPXQ_MUL(frob_pows[3], im,   t, p);
    GEN imx2p3 = FPM_FPXQ_MUL(frob_pows[3], imx2, t, p);
    GEN imx5p3 = FPM_FPXQ_MUL(frob_pows[3], imx5, t, p);
    
    GEN mx3p4  = FPM_FPXQ_MUL(frob_pows[4], mx3,  t, p);
    GEN mx5p4  = FPM_FPXQ_MUL(frob_pows[4], mx5,  t, p);
    GEN imx4p4 = FPM_FPXQ_MUL(frob_pows[4], imx4, t, p);
    
    GEN mx4p5  = FPM_FPXQ_MUL(frob_pows[5], mx4,  t, p);
    GEN mx2p5  = FPM_FPXQ_MUL(frob_pows[5], mx2,  t, p);
    GEN imx3p5 = FPM_FPXQ_MUL(frob_pows[5], imx3, t, p);
    
    GEN mx1p6  = FPM_FPXQ_MUL(frob_pows[6], mx1,  t, p);
    GEN mx3p6  = FPM_FPXQ_MUL(frob_pows[6], mx3,  t, p);
    GEN imx2p6 = FPM_FPXQ_MUL(frob_pows[6], imx2, t, p);
    
    GEN mp7    = FPM_FPXQ_MUL(frob_pows[7], m,    t, p);
    GEN mx2p7  = FPM_FPXQ_MUL(frob_pows[7], mx2,  t, p);
    GEN imx1p7 = FPM_FPXQ_MUL(frob_pows[7], imx1, t, p);
        
    GEN y0;
    y0 = FPXQ_MUL(mx7, imx5,   t, p);
    y0 = FPXQ_MUL(y0,  mx9,    t, p);
    y0 = FPXQ_MUL(y0,  imx3,   t, p);
    y0 = FPXQ_MUL(y0,  mx8p1,  t, p);
    y0 = FPXQ_MUL(y0,  mx6p1,  t, p);
    y0 = FPXQ_MUL(y0,  imx4p1, t, p);
    y0 = FPXQ_MUL(y0,  imx2p1, t, p);
    y0 = FPXQ_MUL(y0,  mx7p2,  t, p);
    y0 = FPXQ_MUL(y0,  mx5p2,  t, p);
    y0 = FPXQ_MUL(y0,  imx3p2, t, p);
    y0 = FPXQ_MUL(y0,  imx1p2, t, p);
    y0 = FPXQ_MUL(y0,  mx6p3,  t, p);
    y0 = FPXQ_MUL(y0,  mx4p3,  t, p);
    y0 = FPXQ_MUL(y0,  imx2p3, t, p);
    y0 = FPXQ_MUL(y0,  imp3,   t, p);
    y0 = FPXQ_MUL(y0,  mx5p4,  t, p);
    y0 = FPXQ_MUL(y0,  mx3p4,  t, p);
    y0 = FPXQ_MUL(y0,  mx4p5,  t, p);
    y0 = FPXQ_MUL(y0,  mx2p5,  t, p);
    y0 = FPXQ_MUL(y0,  mx3p6,  t, p);
    y0 = FPXQ_MUL(y0,  mx1p6,  t, p);
    y0 = FPXQ_MUL(y0,  mx2p7,  t, p);
    y0 = FPXQ_MUL(y0,  mp7,    t, p);
    
    GEN y1;
    y1 = FPXQ_MUL(mx4, imx8,   t, p);
    y1 = FPXQ_MUL(y1,  mx3p1,  t, p);
    y1 = FPXQ_MUL(y1,  imx7p1, t, p);
    y1 = FPXQ_MUL(y1,  imx6p2, t, p);
    y1 = FPXQ_MUL(y1,  mx2p2,  t, p);
    y1 = FPXQ_MUL(y1,  imx5p3, t, p);
    y1 = FPXQ_MUL(y1,  mx1p3,  t, p);
    y1 = FPXQ_MUL(y1,  imx4p4, t, p);
    y1 = FPXQ_MUL(y1,  imx3p5, t, p);
    y1 = FPXQ_MUL(y1,  imx2p6, t, p);
    y1 = FPXQ_MUL(y1,  imx1p7, t, p);

    GEN y2 = m;
    
    GEN T0 = FPXQ_MUL(y2, y1, t, p);
    GEN T1 = FPXQ_MUL(y2, y0, t, p);
    GEN T2 = FPXQ_SQR(T0, t, p);
    GEN T3 = FPXQ_MUL(T2, T1, t, p);
    
    OUT();
    return T3;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_bn(pairing_data_t* const pdata, GEN unreduced)
{
    //
    // Algorithm from "On the final exponentiation for calculating pairings
    // on ordinary elliptic curves" by M. Scott et al.
    //
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    
    GEN* frob_pows = pdata->frobenius_powers;

    GEN mu1 = FPXQ_POW(m, pdata->u, t, p);
    GEN mu2 = FPXQ_POW(mu1, pdata->u, t, p);
    GEN mu3 = FPXQ_POW(mu2, pdata->u, t, p);
    
    GEN mp1 = FPM_FPXQ_MUL(frob_pows[1], m,   t, p);
    GEN mp2 = FPM_FPXQ_MUL(frob_pows[1], mp1, t, p);
    GEN mp3 = FPM_FPXQ_MUL(frob_pows[1], mp2, t, p);
    
    GEN mu1p1 = FPM_FPXQ_MUL(frob_pows[1], mu1, t, p);
    GEN mu2p1 = FPM_FPXQ_MUL(frob_pows[1], mu2, t, p);
    GEN mu3p1 = FPM_FPXQ_MUL(frob_pows[1], mu3, t, p);
    
    GEN mu2p2 = FPM_FPXQ_MUL(frob_pows[1], mu2p1, t, p);
    
    GEN y0 = FPXQ_MUL(FPXQ_MUL(mp1, mp2, t, p), mp3, t, p);
    GEN y1 = FPM_FPXQ_MUL(frob_pows[6], m, t, p);
    GEN y2 = mu2p2;
    GEN y3 = FPM_FPXQ_MUL(frob_pows[6], mu1p1, t, p);
    GEN y4 = FPM_FPXQ_MUL(frob_pows[6], FPXQ_MUL(mu1, mu2p1, t, p), t, p);
    GEN y5 = FPM_FPXQ_MUL(frob_pows[6], mu2, t,p);
    GEN y6 = FPM_FPXQ_MUL(frob_pows[6], FPXQ_MUL(mu3, mu3p1, t, p), t, p);
    
    GEN T0;
    GEN T1;

    T0 = FPXQ_SQR(y6, t, p);
    T0 = FPXQ_MUL(T0, y4, t, p);
    T0 = FPXQ_MUL(T0, y5, t, p);
    T1 = FPXQ_MUL(y3, y5, t, p);
    T1 = FPXQ_MUL(T1, T0, t, p);
    T0 = FPXQ_MUL(T0, y2, t, p);
    T1 = FPXQ_SQR(T1, t, p);
    T1 = FPXQ_MUL(T1, T0, t, p);
    T1 = FPXQ_SQR(T1, t, p);
    T0 = FPXQ_MUL(T1, y1, t, p);
    T1 = FPXQ_MUL(T1, y0, t, p);
    T0 = FPXQ_SQR(T0, t, p);
    T0 = FPXQ_MUL(T0, T1, t, p);
    
    OUT();
    return T0;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_mnt(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN hard_1 = FPXQ_POW(unreduced, pdata->u, t, p);
    GEN hard_2 = FPM_FPXQ_MUL(pdata->frobenius_powers[1], unreduced, t, p);
    GEN hard   = FPXQ_MUL(hard_1, hard_2, t, p);
    OUT();
    return hard;
}
//-----------------------------------------------------------------------------
static GEN
_hard_pow_freeman(pairing_data_t* const pdata, GEN unreduced)
{
    //
    // Algorithm from "On the final exponentiation for calculating pairings
    // on ordinary elliptic curves" by M. Scott et al.
    //
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN m = unreduced;
    
    GEN* frob_pows = pdata->frobenius_powers;
    
    GEN mu1 = FPXQ_POW(m, pdata->u, t, p);
    GEN mu2 = FPXQ_POW(mu1, pdata->u, t, p);
    GEN mu3 = FPXQ_POW(mu2, pdata->u, t, p);
    
    GEN mp1 = FPM_FPXQ_MUL(frob_pows[1], m,   t, p);
    GEN mp2 = FPM_FPXQ_MUL(frob_pows[1], mp1, t, p);
    GEN mp3 = FPM_FPXQ_MUL(frob_pows[1], mp2, t, p);
    
    GEN mu1p1 = FPM_FPXQ_MUL(frob_pows[1], mu1, t, p);
    GEN mu1p2 = FPM_FPXQ_MUL(frob_pows[2], mu1, t, p);
    
    GEN mu2p1 = FPM_FPXQ_MUL(frob_pows[1], mu2, t, p);
    GEN mu2p2 = FPM_FPXQ_MUL(frob_pows[2], mu2, t, p);
        
    GEN y0 = mp3;
    GEN y1 = FPM_FPXQ_MUL(frob_pows[5], m, t, p);
    GEN y2 = FPM_FPXQ_MUL(frob_pows[5], mp1, t, p);
    GEN y3 = FPM_FPXQ_MUL(frob_pows[5], FPXQ_MUL(mu1p1, mu2p1, t, p), t, p);
        y3 = FPXQ_MUL(y3, FPXQ_MUL(mu1p2, mp2, t, p), t, p);
    GEN y4 = mu2p2;
    GEN y5 = FPM_FPXQ_MUL(frob_pows[5], FPXQ_MUL(mu2, mu1, t, p), t,p);
    GEN y6 = FPM_FPXQ_MUL(frob_pows[5], mu3, t, p);
    
    GEN T0  = FPXQ_MUL(y4, y6, t, p);
    GEN T1  = FPXQ_MUL(y5, y6, t, p);
    GEN T2  = FPXQ_MUL(T1, y3, t, p);
    GEN T3  = FPXQ_MUL(T0, T1, t, p);
    GEN T4  = FPXQ_SQR(T3, t, p);
    GEN T5  = FPXQ_MUL(T4, T2, t, p);
    GEN T6  = FPXQ_MUL(T5, y1, t, p);
    GEN T7  = FPXQ_MUL(T5, y2, t, p);
    GEN T8  = FPXQ_MUL(T7, y0, t, p);
    GEN T9  = FPXQ_MUL(T6, T7, t, p);
    GEN T10 = FPXQ_SQR(T9, t, p);
    GEN T11 = FPXQ_MUL(T8, T10, t, p);
    
    OUT();
    return T11;
}
//-----------------------------------------------------------------------------
MAYBE_UNUSED static GEN
_hard_pow_unknown_interleaving(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    GEN e = pdata->hard_exp;
    GEN m = unreduced;
    int v = varn(unreduced);
    
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    } 
    //
    // Write e in base p: e = sigma ei.p^i. The ei are stored in pdata->fe_coefs
    // once and for all.
    //
    if (pdata->fe_coefs == NULL) {
        int le = expi(e) + 1;
        int lp = expi(p) + 1;
        int nc = le / lp + 2;
        GEN coefs = cgetg(nc + 1, t_VEC);
        GEN hexp  = gcopy(e);
        for (int i = 1; i <= nc; i++) {
            hexp = dvmdii(hexp, p, &gel(coefs, i));
        }
        pdata->fe_coefs = gclone(coefs);
    }
    //
    // Precomputes the m^(2^j) for all 0 < 2^j < p.
    //
    int lp = expi(p) + 1;
    GEN mpows = cgetg(lp + 1, t_VEC);
    gel(mpows, 1) = m;
    for (int j = 2; j <= lp; j++) {
        gel(mpows, j) = FPXQ_SQR(gel(mpows, j - 1), t, p);
    }
    //
    // Compute the (m^ei)^(p^i) using the precomputed m^(2^j) and the
    // Frobenius powers.
    //
    GEN hard = Rg_to_FpXQ(stoi(1), t, p);
    GEN fe_coefs = pdata->fe_coefs;
    int npows = lg(fe_coefs) - 1;
    for (int i = npows - 1; i >= 0; i--) {
        hard = FPM_FPXQ_MUL(pdata->frobenius_powers[1], hard, t, p);
        if (equalis(gel(fe_coefs, i + 1), 0)) {
            continue;
        }
        GEN mei = Rg_to_FpXQ(stoi(1), t, p);
        setvarn(mei, v);
        
        GEN ei = gel(fe_coefs, i + 1);
        long j = expi(ei);
        while (j >= 0) {
            if (_get_bit(ei, j)) {
                mei = FPXQ_MUL(mei, gel(mpows, j + 1), t, p);
            }
            j--;
        }
        hard = FPXQ_MUL(hard, mei, t, p);
    }
    setvarn(hard, varn(unreduced));
    
    OUT();
    return hard;
}
//-----------------------------------------------------------------------------
MAYBE_UNUSED static GEN
_hard_pow_unknown_katoetal(pairing_data_t* const pdata, GEN unreduced)
{
    //
    // Algorithm from the paper "Fast exponentiation in extension field with
    // Frobenius mappings" by Kato, Nekado, Nogami and Morikawa, Jan. 2008.
    //
    IN();
    
    pari_timer mytimer;
    (void)TIMER(&mytimer);
    TIMERstart(&mytimer);
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    GEN e = pdata->hard_exp;
    GEN m = unreduced;
    int v = varn(unreduced);
    
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }
    //
    // Write e in base p: e = sigma ei.p^i. The ei are stored in pdata->fe_coefs
    // once and for all.
    //
    if (pdata->fe_coefs == NULL) {
        int le = expi(e) + 1;
        int lp = expi(p) + 1;
        int nc = le / lp + 2;
        GEN coefs = cgetg(nc + 1, t_VEC);
        GEN hexp  = gcopy(e);
        for (int i = 1; i <= nc; i++) {
            hexp = dvmdii(hexp, p, &gel(coefs, i));
        }
        pdata->fe_coefs = gclone(coefs);
    }
    //
    // Precomputes the m^(2^j) for all 0 < 2^j < p.
    //
    int lp = expi(p) + 1;
    GEN mpows = cgetg(lp + 1, t_VEC);
    gel(mpows, 1) = m;
    for (int j = 2; j <= lp; j++) {
        gel(mpows, j) = FPXQ_SQR(gel(mpows, j - 1), T, p);
    } 
    
    GEN fecoefs = pdata->fe_coefs;
    int t = lp;
    GEN B = mpows;
    int ln = lg(fecoefs) - 1;
    while (equalis(gel(fecoefs, ln), 0)) {
        ln--;
    }
    if (pdata->r == 0) {
        _set_r_kato(pdata, t, ln);
    }
    unsigned short r = pdata->r;
    unsigned short c = (ln / r);
    if (ln % r) {
        c++;
    }
    //
    // We copy the coefficients for two reasons :
    //     a) We have to modify them for a simpler code OR add checks to
    //        be able to safely performs a _get_bit.
    //     b) In order to avoid potentially costly checks, we need to pad
    //        our c * r array with zeros.
    //
    GEN nij = cgetg(c * r + 1, t_VEC);
    int i = 1;
    for (i = 1; i <= ln; i++) {
        gel(nij, i) = gcopy(gel(fecoefs, i));
    }
    for (; i <= c * r; i++) {
        gel(nij, i) = stoi(0);
    }
    //
    // Initializes the C[i] for 0 < i < 2^r and the R[i] for 0 <= i < r
    //
    int lc = 1 << r;
    GEN C  = cgetg(lc, t_VEC);
    for (int i = 1; i < lc; i++) {
        gel(C, i) = Rg_to_FpXQ(stoi(1), T, p);
        setvarn(gel(C, i), v);
    }
    GEN R  = cgetg(r + 1, t_VEC);
    for (int i = 0; i < r; i++) {
        gel(R, i + 1) = Rg_to_FpXQ(stoi(1), T, p);
        setvarn(gel(R, i + 1), v);
    }
    GEN* frob_pows = pdata->frobenius_powers;
    //
    // This is a straitforward implementation of the algorithm as described
    // in the aforementioned paper.
    //
    for (int j = c - 1; j >= 0; j--) {
        for (int k = 0; k < t; k++) {
            int M = 0;
            for (int i = 0; i < r; i++) {
                GEN coef = gel(nij, 1 + j + i * c);
                if (_get_bit(coef, 0)) {
                    M += (1 << i);
                }
                gel(nij, 1 + j + i * c) = shifti(coef, -1);
            }
            if (M != 0) {
                gel(C, M) = FPXQ_MUL(gel(C, M), gel(B, k + 1), T, p);
            }
        }
        if (j != 0) {
            for (int i = 1; i < lc; i++) {
                gel(C, i) = FPM_FPXQ_MUL(frob_pows[1], gel(C, i), T, p);
            }
        }
    }
    for (int i = 0; i < r; i++) {
        for (int j = 1; j < lc; j++) {
            if ((1 << i) & j) {
                gel(R, i + 1) = FPXQ_MUL(gel(R, i + 1), gel(C, j), T, p);
            }
        }
    }
    GEN hard = gel(R, r);
    for (int i = r - 2; i >= 0; i--) {
        hard = FPM_FPXQ_MUL(frob_pows[c], hard, T, p);
        hard = FPXQ_MUL(hard, gel(R, i + 1), T, p);
    }
    OUT();
    return hard;
}
//-----------------------------------------------------------------------------
MAYBE_UNUSED static GEN
_hard_pow_unknown_avanzi(pairing_data_t* const pdata, GEN unreduced) {
    //
    // Frobenius-abusing algorithm from the paper "Generic Efficient Arithmetic
    // Algorithms for PAFFs and Related Algebraic Structures" by Avanzi
    // and Mihailescu, in Selected Areas in Cryptology – SAC 2003, LNCS 3006.
    //
    IN();
    
    pari_timer mytimer;
    (void)TIMER(&mytimer);
    TIMERstart(&mytimer);
    
    GEN p = pdata->charac;
    GEN T = pdata->irredpol;
    GEN n = pdata->hard_exp;
    GEN g = unreduced;
    int v = varn(unreduced);
    
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }    
    //
    // Write e in base p: e = sigma ei.p^i. The ei are stored in pdata->fe_coefs
    // once and for all.
    //
    if (pdata->fe_coefs == NULL) {
        int ln = expi(n) + 1;
        int lp = expi(p) + 1;
        int nc = ln / lp + 2;
        GEN coefs = cgetg(nc + 1, t_VEC);
        GEN hexp  = gcopy(n);
        for (int i = 1; i <= nc; i++) {
            hexp = dvmdii(hexp, p, &gel(coefs, i));
        }
        pdata->fe_coefs = gclone(coefs);
    }
    
    GEN ni = pdata->fe_coefs;
    int m  = lg(ni) - 1;
    while (equalis(gel(ni, m), 0)) {
        m--;
    }
    
    //
    // Precomputes the g^(j) for all 0 < j < 2^l - 1.
    //
    int u = expi(p) + 1;
    if (pdata->r == 0) {
        _set_l_avanzi(pdata, m, u);
    }
    int l      = pdata->r;
    int nj     = (1 << l);
    GEN gi     = cgetg(nj + 1, t_VEC);
    gel(gi, 1) = g;
    for (int j = 2; j < nj; j++) {
        gel(gi, j) = FPXQ_MUL(g, gel(gi, j - 1), T, p);
    }
    int K = u / l;
    if (u % l) {
        K++;
    }    
    //
    // Precomputes the n_{ij} for all 0 <= i < m and 0 <= j < K.
    //
    GEN nij = cgetg(m + 1, t_VEC);
    for (int i = 0; i < m; i++) {
        gel(nij, i + 1) = cgetg(K + 1, t_VEC);
    }
    for (int i = 0; i < m; i++) {
        GEN curn = gcopy(gel(ni, i + 1));
        GEN cnij = gel(nij, i + 1);
        for (int j = 0; j < K; j++) {
            curn = dvmdii(curn, stoi(1 << l), &gel(cnij, j + 1));
        }
    }
    GEN  hard  = Rg_to_FpXQ(stoi(1), T, p);
    GEN* frobs = pdata->frobenius_powers;
    
    for (int j = K - 1; j >= 0; j--) {
        GEN pi = Rg_to_FpXQ(stoi(1), T, p);
        for (int i = m - 1; i >= 0; i--) {
            int cnij = itos(gel(gel(nij, i + 1), j + 1));
            GEN cgi  = (cnij != 0) ? gel(gi, cnij) : pol_1(v);
            pi = FPXQ_MUL(pi, cgi, T, p);
            if (i != 0) {
                pi = FPM_FPXQ_MUL(frobs[1], pi, T, p);
            }
        }
        hard = FPXQ_MUL(hard, pi, T, p);
        if (j != 0) {
            for (int is = 0; is < l; is++) {
                hard = FPXQ_SQR(hard, T, p);
            }
        }
    }    
    OUT();
    return hard;
}
//-----------------------------------------------------------------------------
inline static GEN
_hard_pow_unknown(pairing_data_t* const pdata, GEN unreduced)
{
    //return _hard_pow_unknown_interleaving(pdata, unreduced);
    //return _hard_pow_unknown_katoetal(pdata, unreduced);
    return _hard_pow_unknown_avanzi(pdata, unreduced);
}
//-----------------------------------------------------------------------------
static GEN
_easy_pow(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }
    int k = pdata->embed_deg;
    int d = k / pdata->edeg_sdiv;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN tmp = FPM_FPXQ_MUL(pdata->frobenius_powers[d], unreduced, t, p);
    GEN res = FPXQ_DIV(tmp, unreduced, t, p);
    
    if (equalis(pdata->easy_exp_2, 1) != 1) {
        res = _easy_pow_part_2(pdata, res);
    }
    OUT();
    return res;
}
//-----------------------------------------------------------------------------
static GEN
_easy_pow_part_2(pairing_data_t* const pdata, GEN unreduced)
{
    IN();
    int k = pdata->embed_deg;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    GEN* frob_pows = pdata->frobenius_powers;
    GEN  res;
    
    if (pdata->fields_are_friendly) {
        //
        // In this case, easy_exp_2 = p^(k/6) + 1
        //
        GEN res = FPM_FPXQ_MUL(frob_pows[k / 6], unreduced, t, p);
        res = FPXQ_MUL(res, unreduced, t, p);
        OUT();
        return res;
    }
    GEN  tmp_1;
    GEN  tmp_2;
    GEN  tmp_3;
    GEN  tmp_4;
    GEN  tmp_5;
    
    switch (k) {
        case 15:
        case 21:
        case 33:
        case 39:
        case 51:
        case 57:
            //
            // easy_exp_2 = p^2 + p + 1
            //
            tmp_1 = FPM_FPXQ_MUL(frob_pows[2], unreduced, t, p);
            tmp_2 = FPM_FPXQ_MUL(frob_pows[1], unreduced, t, p);
            res   = FPXQ_MUL(tmp_1, tmp_2, t, p);
            res   = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 35:
        case 55:
            //
            // easy_exp_2 = p^4 + p^3 + p^2 + p + 1
            //
            tmp_1 = FPM_FPXQ_MUL(frob_pows[4], unreduced, t, p);
            tmp_2 = FPM_FPXQ_MUL(frob_pows[3], unreduced, t, p);
            tmp_3 = FPM_FPXQ_MUL(frob_pows[2], unreduced, t, p);
            tmp_4 = FPM_FPXQ_MUL(frob_pows[1], unreduced, t, p);
            
            res = FPXQ_MUL(tmp_1, tmp_2, t, p);
            res = FPXQ_MUL(res, tmp_3, t, p);
            res = FPXQ_MUL(res, tmp_4, t, p);            
            res = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 45:
        case 63:
            //
            // easy_exp_2 = p^6 + p^3 + 1
            //
            tmp_1 = FPM_FPXQ_MUL(frob_pows[6], unreduced, t, p);
            tmp_2 = FPM_FPXQ_MUL(frob_pows[3], unreduced, t, p);
            
            res   = FPXQ_MUL(tmp_1, tmp_2, t, p);
            res   = FPXQ_MUL(res, unreduced, t, p);
            break;
        
        case 20:
        case 28:
        case 44:
        case 52:
            //
            // easy_exp_2 = p^2 + 1
            //
            res = FPM_FPXQ_MUL(frob_pows[2], unreduced, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 30:
            //
            // easy_exp_2 = p^7 + p^5 + p^2 - (p^6 + p) + 1
            //
            tmp_1 = FPM_FPXQ_MUL(frob_pows[7], unreduced, t, p);
            tmp_2 = FPM_FPXQ_MUL(frob_pows[5], unreduced, t, p);
            tmp_3 = FPM_FPXQ_MUL(frob_pows[2], unreduced, t, p);

            tmp_4 = FPM_FPXQ_MUL(frob_pows[6], unreduced, t, p);
            tmp_5 = FPM_FPXQ_MUL(frob_pows[1], unreduced, t, p);
            
            res = FPXQ_MUL(tmp_1, tmp_2, t, p);
            res = FPXQ_MUL(res, tmp_3, t, p);
            
            tmp_5 = FPXQ_MUL(tmp_4, tmp_5, t, p);
            tmp_5 = FPXQ_INV(tmp_5, t, p);
            
            res = FPXQ_MUL(res, tmp_5, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 40:
        case 56:
            //
            // easy_exp_2 = p^4 + 1
            //
            res = FPM_FPXQ_MUL(frob_pows[4], unreduced, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 42:
            //
            // easy_exp_2 = p^9 + p^7 + p^2 - (p^8 + p) + 1
            //
            tmp_1 = FPM_FPXQ_MUL(frob_pows[9], unreduced, t, p);
            tmp_2 = FPM_FPXQ_MUL(frob_pows[7], unreduced, t, p);
            tmp_3 = FPM_FPXQ_MUL(frob_pows[2], unreduced, t, p);

            tmp_4 = FPM_FPXQ_MUL(frob_pows[8], unreduced, t, p);
            tmp_5 = FPM_FPXQ_MUL(frob_pows[1], unreduced, t, p);
            
            res = FPXQ_MUL(tmp_1, tmp_2, t, p);
            res = FPXQ_MUL(res, tmp_3, t, p);
            
            tmp_5 = FPXQ_MUL(tmp_4, tmp_5, t, p);
            tmp_5 = FPXQ_INV(tmp_5, t, p);
            
            res = FPXQ_MUL(res, tmp_5, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 50:
            //
            // easy_exp_2 = p^5 + 1
            //
            res = FPM_FPXQ_MUL(frob_pows[5], unreduced, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;

        case 54:
            //
            // easy_exp_2 = p^9 + 1
            //
            res = FPM_FPXQ_MUL(frob_pows[9], unreduced, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;
        
        case 60:
            //
            // easy_exp_2 = p^14 + p^10 + p^4 - (p^12 + p^2) + 1
            //
            tmp_1 = FPM_FPXQ_MUL(frob_pows[14], unreduced, t, p);
            tmp_2 = FPM_FPXQ_MUL(frob_pows[10], unreduced, t, p);
            tmp_3 = FPM_FPXQ_MUL(frob_pows[4],  unreduced, t, p);

            tmp_4 = FPM_FPXQ_MUL(frob_pows[12], unreduced, t, p);
            tmp_5 = FPM_FPXQ_MUL(frob_pows[2],  unreduced, t, p);
            
            res = FPXQ_MUL(tmp_1, tmp_2, t, p);
            res = FPXQ_MUL(res, tmp_3, t, p);
            
            tmp_5 = FPXQ_MUL(tmp_4, tmp_5, t, p);
            tmp_5 = FPXQ_INV(tmp_5, t, p);
            
            res = FPXQ_MUL(res, tmp_5, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;
        
        case 6:
        case 10:
        case 14:
        case 26:
        case 34:
        case 38:
        case 46:
        case 58:
        case 62:
            //
            // easy_exp_2 = p + 1
            //
            res = FPM_FPXQ_MUL(frob_pows[1], unreduced, t, p);
            res = FPXQ_MUL(res, unreduced, t, p);
            break;
        
        default:
            //
            // easy_exp_2 = 1
            //
            return unreduced;
    }
    OUT();
    return res;
}
//-----------------------------------------------------------------------------
MAYBE_UNUSED static int
_compute_field_maps(pairing_data_t* const pdata)
{
    IN();
    //
    // All the data we need will be cloned on the heap -- we'll thus
    // simply restore avma to its old value before exiting the function,
    //
    pari_sp av = avma;

    GEN P = ffinit_rand(pdata->charac, pdata->embed_deg / 2);
    GEN Q = pdata->irredpol;
    GEN p = pdata->charac;

    int degP = degpol(P);
    int degQ = degpol(Q);
    int vno  = varn(P);

    GEN iso    = FpX_ffisom(P, Q, p);
    GEN isofqx = Rg_to_FpXQ(iso, Q, p);
    GEN isopow = Rg_to_FpXQ(gen_1, Q, p);

    int ncols = degP;
    int nrows = degQ;

    //
    // Computes to_ext, the change of basis matrix from F_{p^(k/2)} to F_{p^k}.
    //
    GEN to_ext = cgetg(ncols + 1, t_MAT);
    for (int i = 1; i < ncols + 1; i++) {
        gel(to_ext, i) = cgetg(nrows + 1, t_COL);
    }

    for (int i = 1; i < ncols + 1; i++) {
        int dentry = degpol(isopow);
        GEN col = gel(to_ext, i);
        for (int j = 0; j < dentry + 1; j++) {
            gel(col, j + 1) = gcopy((GEN)isopow[2 + j]);
        }
        for (int j = dentry + 1; j < nrows; j++) {
            gel(col, j + 1) = stoi(0);
        }
        isopow = FPXQ_MUL(isopow, isofqx, Q, p);
    }
    to_ext = RgM_to_FpM(to_ext, p);

    GEN b;
    while (1) {
        //
        // We want to construct F_{p^k} as F_{p^(k/2)}(a) where a^2 = b
        // with b is a non-square in F_{p^(k/2)}.
        //
        b = random_FpX(degP, vno, p);
        if (!FPXQ_ISSQUARE(b, P, p)) {
            break;
        }
    }
    //
    // Embed b in F_{p^k} and get its square root in F_{p^k} to define
    // F_{p^(k/2)}(a).
    //
    GEN eb = FPM_FPXQ_MUL(to_ext, b, P, p);
    GEN a  = FpXQ_sqrtn(eb, stoi(2), Q, p, NULL);

    //
    // Computes pdata->to_iso, the change of basis matrix from F_{p^(k/2)}(a)
    // to the isomorphic field F_{p^k}.
    //
    GEN to_iso = cgetg(degQ + 1, t_MAT);
    for (int i = 1; i < degQ + 1; i++) {
        gel(to_iso, i) = cgetg(degQ + 1, t_COL);
    }
    isopow    = Rg_to_FpXQ(gen_1, Q, p);
    GEN trans = a;
    for (int i = 1; i < degP + 1; i++) {
        //
        // This vector is in F_{p^(k/2)}
        //
        GEN col = gel(to_iso, i);
        int dentry = degpol(isopow);
        for (int j = 0; j < dentry + 1; j++) {
            gel(col, j + 1) = gcopy((GEN)isopow[2 + j]);
        }
        for (int j = dentry + 1; j < degQ; j++) {
            gel(col, j + 1) = stoi(0);
        }
        //
        // This vector is in F_{p^(k/2)}(a)
        //
        col = gel(to_iso, i + degP);
        dentry = degpol(trans);
        for (int j = 0; j < dentry + 1; j++) {
            gel(col, j + 1) = gcopy((GEN)trans[2 + j]);
        }
        for (int j = dentry + 1; j < degQ; j++) {
            gel(col, j + 1) = stoi(0);
        }
        isopow = FPXQ_MUL(isopow, isofqx, Q, p);
        trans  = FPXQ_MUL(isopow, a, Q, p);
    }
    pdata->to_iso = gclone(RgM_to_FpM(to_iso, p));

    //
    // Deduces pdata->to_iso_inv, the change of basis matrix from F_{p^k}
    // to the isomorphic field F_{p^(k/2)}(a).
    //
    pdata->to_iso_inv = gclone(FpM_inv(pdata->to_iso, p));

    int ecode = ERROR;
    if (pdata->to_iso_inv != NULL) {
        ecode = SUCCESS;
    }
    avma = av;
    OUT();

    return ecode;
}
//-----------------------------------------------------------------------------
static void
_set_r_kato(pairing_data_t* const pdata, unsigned int t, unsigned short nc)
{
    float frobcost = pdata->frobcost;
    unsigned int   mincost = nc * t + nc - 1; 
    unsigned short bestr = 1;
    unsigned short maxr  = 10;
    if (nc < maxr) {
        maxr = nc;
    }
    
    for (unsigned short r = 2; r <= maxr; r++) {
        unsigned short c = (nc / r);
        if (nc % r) {
            c++;
        }
        unsigned int cost = 0.5 * r * ((1 << r) - 1) + r - 1;        
        cost += c * t * (1 - 1/((float)(1 << r)));
        cost += (unsigned int)(frobcost * ((c - 1) * ((1 << r) - 1) + r - 1));
        pariprintf("r = %i -- cost = %lu\n", r, cost);
        if (cost < mincost) {
            mincost = cost;
            bestr = r;
        }
    }
    pdata->r = bestr;
    return;
}
//-----------------------------------------------------------------------------
static void
_set_l_avanzi(pairing_data_t* const pdata, unsigned int m, unsigned int u)
{
    float frobcost = pdata->frobcost;
    float sqrcost  = pdata->sqrcost;
    
    unsigned int lmin    = 2;
    unsigned int lmax    = 14;
    unsigned int bestl   = lmin;
    unsigned int mincost = 1 << 31;
    
    for (unsigned int l = lmin; l <= lmax; l++) {
        int K = u / l;
        if (u % l) {
            K++;
        }
        unsigned int cost = (1 << (l - 1));
        cost += (unsigned int)(m * K * (1 - 1/((float)(1 << l))));
        cost += (unsigned int)(sqrcost  * ((1 << (l-1)) + (K - 1) * l -1));
        cost += (unsigned int)(frobcost * ((m - 1) * K));
        pariprintf("l = %i -- cost = %lu\n", l, cost);
        if (cost < mincost) {
            mincost = cost;
            bestl   = l;
        }
    }
    pdata->r = bestl;
    return;
}
//-----------------------------------------------------------------------------
